/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/asng/asinfo.c                                            *
 * Created:     2006-08-19 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2006-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "asinfo.h"

#include <lib/applesingle.h>
#include <lib/buffer.h>
#include <lib/getopt.h>
#include <lib/mactime.h>
#include <lib/fourcc.h>


static int par_list = 0;
static int par_inf_mf = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'i', 1, "input", "string", "Set an input file name" },
	{ 'I', 0, "info", NULL, "Dump the mac file info [no]" },
	{ 'l', 0, "list", NULL, "List applesingle entries [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"asinfo: print AppleSingle header information",
		"usage: asinfo [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"asinfo version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2006-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int as_info_list (applesingle_t *as, FILE *fp)
{
	unsigned      i;
	unsigned long id, ofs, size;
	const char    *name;

	i = 0;
	while (1) {
		id = as_get_index_id (as, i);

		if (id == 0) {
			break;
		}

		switch (id) {
		case AS_ENTRY_DATA:
			name = "data fork";
			break;
		case AS_ENTRY_RSRC:
			name = "resource fork";
			break;
		case AS_ENTRY_NAME:
			name = "file name";
			break;
		case AS_ENTRY_COMMENT:
			name = "comment";
			break;
		case AS_ENTRY_ICON:
			name = "icon (b&w)";
			break;
		case AS_ENTRY_CICON:
			name = "icon (color)";
			break;
		case AS_ENTRY_TIME:
			name = "file times";
			break;
		case AS_ENTRY_FINDER_INFO:
			name = "finder info";
			break;
		case AS_ENTRY_FILE_INFO:
			name = "file flags";
			break;
		default:
			name = "unknown";
			break;
		}

		ofs = as_get_index_offset (as, i);
		size = as_get_index_size (as, i);

		fprintf (fp, "ENTRY %u: ID=%02lX OFS=%08lX SIZE=%08lX (%s)\n",
			i, id, ofs, size, name
		);

		i += 1;
	}

	return (0);
}

static
int as_info_fp (FILE *fp, FILE *out, const char *fname)
{
	applesingle_t as;
	macfile_t     mf;

	static unsigned cnt = 0;

	as_init (&as, fp);

	if (as_read_header (&as)) {
		fprintf (stderr, "error reading applesingle header\n");
		return (1);
	}

	if (par_list) {
		as_info_list (&as, out);
	}
	else if (par_inf_mf) {
		mf_init (&mf);
		as_get_macfile (&as, &mf);
		mf_save (&mf, out);
	}
	else {
		if (cnt > 0) {
			fputs ("\n", out);
		}

		cnt += 1;

		fprintf (out, "%s:\n", fname);

		mf_init (&mf);
		as_get_macfile (&as, &mf);
		mf_print_info (&mf, out);

		fprintf (out, "as magic:           %08lX\n", as_get_magic (&as));
		fprintf (out, "as version:         %08lX\n", as_get_version (&as));
		fprintf (out, "as entries:         %u\n", as_get_entry_count (&as));
	}

	fflush (out);

	return (0);
}

static
int as_info (const char *fname, FILE *out)
{
	int  r;
	FILE *fp;

	fp = fopen (fname, "rb");
	if (fp == NULL) {
		fprintf (stderr, "error opening file (%s)\n", fname);
		return (1);
	}

	r = as_info_fp (fp, out, fname);

	fclose (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r;
	int  ret;
	char **optarg;

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'i':
			ret |= as_info (optarg[0], stdout);
			break;

		case 'I':
			par_inf_mf = 1;
			break;

		case 'l':
			par_list = 1;
			break;

		case 0:
			ret |= as_info (optarg[0], stdout);
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (ret) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
