/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/binhex/hqxinfo.c                                         *
 * Created:     2005-04-07 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/crc.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/hqx.h>
#include <lib/unixfile.h>


static const char *argv0 = NULL;

static int par_inf_all = 1;
static int par_inf_mf = 0;

static int par_short = 0;
static int par_inf_fname = 0;
static int par_inf_hqx_version = 0;
static int par_inf_crc = 0;
static int par_inf_type = 0;
static int par_inf_creator = 0;
static int par_inf_finder = 0;
static int par_inf_dsize = 0;
static int par_inf_rsize = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "all", NULL, "Print all information [default]" },
	{ 'c', 0, "creator", NULL, "Print the creator" },
	{ 'f', 0, "file-name", NULL, "Print the file name" },
	{ 'i', 1, "input", "string", "Set an input file name" },
	{ 'I', 0, "info", NULL, "Dump the mac file info [no]" },
	{ 'l', 0, "finder-flags", NULL, "Print the finder flags" },
	{ 's', 0, "short", NULL, "Short output format [no]" },
	{ 't', 0, "type", NULL, "Print the type" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"hqxinfo: print BinHex header information",
		"usage: hqxinfo [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"hqxinfo version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2005-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int hqx_info_fp (FILE *fp, FILE *out)
{
	hqx_t           hqx;
	char            buf[256];
	unsigned long   v;
	static unsigned cnt = 0;

	if (cnt > 0) {
		fputs ("\n", out);
	}

	cnt += 1;

	if (hqx_init_read (&hqx, fp)) {
		fprintf (stderr, "%s: error reading hqx header\n", argv0);
		return (1);
	}

	if (hqx_read_header (&hqx)) {
		fprintf (stderr, "%s: error reading hqx header\n", argv0);
		return (1);
	}

	if (hqx_get_crc (&hqx) != hqx_calc_crc (&hqx)) {
		fprintf (stderr, "%s: header crc error\n", argv0);
		return (1);
	}

	if (par_inf_mf) {
		macfile_t     mf;
		unsigned long mtime;

		mf_init (&mf);
		hqx_get_macfile (&hqx, &mf);

		mtime = uf_get_mtime_fp (fp);

		mf_set_ctime (&mf, mtime);
		mf_set_mtime (&mf, mtime);

		mf_save (&mf, out);
	}

	if (par_inf_fname || par_inf_all) {
		char *name;

		name = hqx_get_name_cstr (&hqx);

		if (par_short) {
			fprintf (out, "%s\n", name);
		}
		else {
			fprintf (out, "file name:          '%s'\n", name);
		}

		free (name);
	}

	if (par_inf_hqx_version || par_inf_all) {
		v = hqx_get_version (&hqx);

		if (par_short) {
			fprintf (out, "%lu\n", v);
		}
		else {
			fprintf (out, "binhex version:     %lu\n", v);
		}
	}

	if (par_inf_crc || par_inf_all) {
		v = hqx_get_crc (&hqx);
		if (par_short) {
			fprintf (out, "%04lX\n", v);
		}
		else {
			fprintf (out, "CRC:                %04lX\n", v);
		}
	}

	if (par_inf_type || par_inf_all) {
		v = hqx_get_type (&hqx);
		str_set_fourcc (buf, 256, v);
		if (par_short) {
			fprintf (out, "%s\n", buf);
		}
		else {
			fprintf (out,
				"type:               \"%s\" (0x%08lX)\n",
				buf, v
			);
		}
	}

	if (par_inf_creator || par_inf_all) {
		v = hqx_get_creator (&hqx);
		str_set_fourcc (buf, 256, v);
		if (par_short) {
			fprintf (out, "%s\n", buf);
		}
		else {
			fprintf (out,
				"creator:            \"%s\" (0x%08lX)\n",
				buf, v
			);
		}
	}

	if (par_inf_finder || par_inf_all) {
		v = hqx_get_finder_flags (&hqx);
		if (par_short) {
			fprintf (out, "%04lX\n", v);
		}
		else {
			fprintf (out, "finder flags:       0x%04lX\n", v);
		}
	}

	if (par_inf_dsize || par_inf_all) {
		v = hqx_get_data_size (&hqx);
		if (par_short) {
			fprintf (out, "%lu\n", v);
		}
		else {
			fprintf (out, "data fork size:     %lu\n", v);
		}
	}

	if (par_inf_rsize || par_inf_all) {
		v = hqx_get_rsrc_size (&hqx);
		if (par_short) {
			fprintf (out, "%lu\n", v);
		}
		else {
			fprintf (out, "resource fork size: %lu\n", v);
		}
	}

	fflush (out);

	return (0);
}

static
int hqx_info (const char *fname, FILE *out)
{
	int  r;
	FILE *fp;

	fp = fopen (fname, "r");

	if (fp == NULL) {
		fprintf (stderr, "error opening file (%s)\n", fname);
		return (1);
	}

	r = hqx_info_fp (fp, out);

	fclose (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r;
	int  ret;
	char **optarg;

	argv0 = argv[0];

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'a':
			par_inf_all = 1;
			break;

		case 'c':
			par_inf_creator = 1;
			par_inf_all = 0;
			break;

		case 'f':
			par_inf_fname = 1;
			par_inf_all = 0;
			break;

		case 'i':
			ret |= hqx_info (optarg[0], stdout);
			break;

		case 'I':
			par_inf_mf = 1;
			par_inf_all = 0;
			break;

		case 'l':
			par_inf_finder = 1;
			par_inf_all = 0;
			break;

		case 's':
			par_short = 1;
			break;

		case 't':
			par_inf_type = 1;
			par_inf_all = 0;
			break;

		case 0:
			ret |= hqx_info (optarg[0], stdout);
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (ret) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
