/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/convert/dmdftodc42.c                                     *
 * Created:     2011-06-22 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/dc42.h>
#include <lib/file.h>
#include <lib/getopt.h>


#define DMDF_CLUSTER_MAX    40
#define DMDF_CLUSTER_BLOCKS 40
#define DMDF_CLUSTER_SIZE   (DMDF_CLUSTER_BLOCKS * (512 + 12))


typedef struct {
	unsigned long  size;

	unsigned long  offset;

	unsigned short cluster_size[DMDF_CLUSTER_MAX];

	unsigned long  data_cnt;
	unsigned long  data_max;

	unsigned long  tags_cnt;
	unsigned long  tags_max;

	unsigned char  *data;
	unsigned char  *tags;
} dmdf_buf_t;


static const char *argv0 = NULL;

static char par_verbose = 0;
static char par_write_data = 0;
static char par_write_tags = 0;

static const char *par_image_name = "Untitled";


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'd', 0, "data", NULL, "Write only the image data [no]" },
	{ 'i', 1, "input", "string", "Set the input file [stdin]" },
	{ 'n', 1, "image-name", "string", "Set the image name [Untitled]" },
	{ 'o', 1, "output", "string", "Set the output file [stdout]" },
	{ 't', 0, "tags", NULL, "Write only the tags [no]" },
	{ 'v', 0, "verbose", NULL, "Verbose operation [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"dmdftodc42: convert DMDF (DiskMaker) to DC42 (DiskCopy 4.2)",
		"usage: dmdftodc42 [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"dmdftodc42 version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2011 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}


static
int dmdf_read_uint16 (FILE *fp, unsigned *val)
{
	int c;

	c = fgetc (fp);

	if (c == EOF) {
		return (1);
	}

	*val = c & 0xff;

	c = fgetc (fp);

	if (c == EOF) {
		return (1);
	}

	*val = (*val << 8) | (c & 0xff);

	return (0);
}

static
dmdf_buf_t *dmdf_buf_new (unsigned long size)
{
	dmdf_buf_t *buf;

	buf = malloc (sizeof (dmdf_buf_t));

	if (buf == NULL) {
		return (NULL);
	}

	buf->size = size;

	buf->offset = 0;

	buf->data_cnt = 0;
	buf->data_max = size;

	buf->tags_cnt = 0;
	buf->tags_max = 12 * (size / 512);

	buf->data = malloc (buf->data_max);

	if (buf->data == NULL) {
		free (buf);
		return (NULL);
	}

	buf->tags = malloc (buf->tags_max);

	if (buf->tags == NULL) {
		free (buf->data);
		free (buf);
		return (NULL);
	}

	return (buf);
}

static
void dmdf_buf_free (dmdf_buf_t *buf)
{
	free (buf->tags);
	free (buf->data);
	free (buf);
}

static
int dmdf_write (FILE *fp, const void *buf, unsigned long cnt)
{
	unsigned long       n;
	const unsigned char *tmp;

	tmp = buf;

	while (cnt > 0) {
		n = (cnt < 16384) ? cnt : 16384;

		if (fwrite (tmp, 1, n, fp) != n) {
			return (1);
		}

		tmp += n;
		cnt -= n;
	}

	return (0);
}

static
int dmdf_buf_write_data (dmdf_buf_t *buf, FILE *fp)
{
	return (dmdf_write (fp, buf->data, buf->data_cnt));
}

static
int dmdf_buf_write_tags (dmdf_buf_t *buf, FILE *fp)
{
	return (dmdf_write (fp, buf->tags, buf->tags_cnt));
}

static
int dmdf_buf_write_dc42 (dmdf_buf_t *buf, FILE *fp)
{
	unsigned char hdr[128];
	unsigned long dchk, tchk;

	if ((12 * buf->data_cnt) != (512 * buf->tags_cnt)) {
		return (1);
	}

	dchk = dc42_calc_checksum (buf->data, buf->data_cnt, 0);

	if (buf->tags_cnt >= 12) {
		tchk = dc42_calc_checksum (buf->tags + 12, buf->tags_cnt - 12, 0);
	}
	else {
		tchk = 0;
	}

	dc42_init (hdr);
	dc42_set_name_cstr (hdr, par_image_name);
	dc42_set_data_size (hdr, buf->data_cnt);
	dc42_set_tag_size (hdr, buf->tags_cnt);
	dc42_set_data_checksum (hdr, dchk);
	dc42_set_tag_checksum (hdr, tchk);

	if (dc42_write (hdr, fp)) {
		return (1);
	}

	if (dmdf_write (fp, buf->data, buf->data_cnt)) {
		return (1);
	}

	if (dmdf_write (fp, buf->tags, buf->tags_cnt)) {
		return (1);
	}

	return (0);
}

static
int dmdf_buf_add_uint16 (dmdf_buf_t *buf, unsigned val)
{
	unsigned long tmp;

	tmp = buf->offset % DMDF_CLUSTER_SIZE;

	if (tmp < (DMDF_CLUSTER_BLOCKS * 512)) {
		if ((buf->data_cnt + 2) > buf->data_max) {
			return (1);
		}

		buf->data[buf->data_cnt++] = (val >> 8) & 0xff;
		buf->data[buf->data_cnt++] = val & 0xff;
	}
	else {
		if ((buf->tags_cnt + 2) > buf->tags_max) {
			return (1);
		}

		buf->tags[buf->tags_cnt++] = (val >> 8) & 0xff;
		buf->tags[buf->tags_cnt++] = val & 0xff;
	}

	buf->offset += 2;

	return (0);
}

static
int dmdf_decode_buf (dmdf_buf_t *buf, FILE *inp)
{
	unsigned cnt, val;
	unsigned cl;

	while (1) {
		cl = buf->offset / DMDF_CLUSTER_SIZE;

		if ((cl < DMDF_CLUSTER_MAX) && (buf->cluster_size[cl] == 0xffff)) {
			cnt = DMDF_CLUSTER_SIZE / 2;

			while (cnt > 0) {
				if (dmdf_read_uint16 (inp, &val)) {
					return (1);
				}

				if (dmdf_buf_add_uint16 (buf, val)) {
					return (1);
				}

				cnt -= 1;
			}

			continue;
		}

		if (dmdf_read_uint16 (inp, &cnt)) {
			break;
		}

		if (cnt & 0x8000) {
			cnt = (~cnt + 1) & 0xffff;

			if (dmdf_read_uint16 (inp, &val)) {
				return (1);
			}

			while (cnt > 0) {
				if (dmdf_buf_add_uint16 (buf, val)) {
					return (1);
				}

				cnt -= 1;
			}
		}
		else {
			while (cnt > 0) {
				if (dmdf_read_uint16 (inp, &val)) {
					return (1);
				}

				if (dmdf_buf_add_uint16 (buf, val)) {
					return (1);
				}

				cnt -= 1;
			}
		}
	}

	return (0);
}

static
int dmdf_decode (FILE *inp, FILE *out)
{
	unsigned   i;
	unsigned   val, size;
	dmdf_buf_t *buf;

	if (dmdf_read_uint16 (inp, &val)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	if (val != 0x0001) {
		fprintf (stderr, "%s: bad magic? (0x%04x)\n", argv0, val);
		return (1);
	}

	if (dmdf_read_uint16 (inp, &size)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	if ((size != 400) && (size != 800)) {
		fprintf (stderr, "%s: bad size (%u)\n", argv0, size);
		return (1);
	}

	buf = dmdf_buf_new (1024UL * size);

	if (buf == NULL) {
		return (1);
	}

	for (i = 0; i < 40; i++) {
		if (dmdf_read_uint16 (inp, &val)) {
			fprintf (stderr, "%s: read error\n", argv0);
			return (1);
		}

		buf->cluster_size[i] = val;
	}

	if (dmdf_decode_buf (buf, inp)) {
		fprintf (stderr, "%s: decode error\n", argv0);
		return (1);
	}

	if (par_verbose) {
		fprintf (stderr, "data size: %lu\n", buf->data_cnt);
		fprintf (stderr, "tags size: %lu\n", buf->tags_cnt);
	}

	if (par_write_data) {
		if (dmdf_buf_write_data (buf, out)) {
			return (1);
		}
	}

	if (par_write_tags) {
		if (dmdf_buf_write_tags (buf, out)) {
			return (1);
		}
	}

	if ((par_write_data == 0) && (par_write_tags == 0)) {
		if (dmdf_buf_write_dc42 (buf, out)) {
			return (1);
		}
	}

	dmdf_buf_free (buf);

	return (0);
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;
	char *finp, *fout;
	FILE *inp, *out;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'd':
			par_write_data = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'n':
			par_image_name = optarg[0];
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 't':
			par_write_tags = 1;
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr,
					"%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = mac_open_inp (finp, "rb", 1);

	if (finp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n",
			argv0, finp
		);
		return (EXIT_FAILURE);
	}

	out = mac_open_out (fout, "wb", 1);

	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n",
			argv0, fout
		);
		return (EXIT_FAILURE);
	}

	r = dmdf_decode (inp, out);

	mac_close (out);
	mac_close (inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
