/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/convert/mbptotar.c                                       *
 * Created:     2007-11-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2007-2011 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/mactime.h>
#include <lib/strings.h>
#include <lib/tar.h>


static const char *arg0 = NULL;


static int        par_verbose = 0;

static const char *par_folder_file = ".folder";


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'f', 1, "folder-file", "string", "Set the folder file name [.folder]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'v', 0, "verbose", NULL, "Print file names to stderr [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mbptotar: convert MacBinary II+ to tar",
		"usage: mbptotar [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mbptotar version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2007-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int tar_set_mb (unsigned char *tar, const unsigned char *mb)
{
	unsigned long val;

	tar_init (tar);

	val = mb_get_mtime (mb);
	val = mac_time_to_unix (val);
	tar_set_mtime (tar, val);

	tar_set_user (tar, 0, "root");
	tar_set_group (tar, 0, "root");
	tar_set_device (tar, 0, 0);

	return (0);
}

static
int convert_file (FILE *inp, FILE *out, const unsigned char *mb, char *prefix)
{
	char          *fname;
	unsigned char tar[512];
	unsigned long n;
	unsigned long size, ofs;
	unsigned long val;

	if (mb_get_version (mb) == 0) {
		fprintf (stderr, "%s: macbinary header error\n", arg0);
		return (1);
	}

	if (tar_set_mb (tar, mb)) {
		return (1);
	}

	n = strlen (prefix);

	fname = mb_get_name_cstr (mb);

	if ((n + strlen (fname) + 5) >= 256) {
		fprintf (stderr, "%s: name too long (%s)\n", arg0, fname);
		free (fname);
		return (1);
	}

	strcat (prefix + n, fname);
	strcat (prefix + n, ".bin");

	free (fname);

	if (par_verbose) {
		fprintf (stderr, "%s\n", prefix);
	}

	if (tar_set_fname (tar, prefix)) {
		fprintf (stderr, "%s: bad file name (%s)\n", arg0, prefix);
		return (1);
	}

	prefix[n] = 0;

	val = mb_get_data_size (mb);
	size = (val + 127) & ~0x7fUL;

	val = mb_get_rsrc_size (mb);
	size += (val + 127) & ~0x7fUL;

	val = mb_get_getinfo_size (mb);
	size += (val + 127) & ~0x7fUL;

	tar_set_size (tar, size + 128);

	tar_set_mode (tar, 0644);
	tar_set_type (tar, '0');

	tar_set_checksum (tar);

	if (fwrite (tar, 1, 512, out) != 512) {
		fprintf (stderr, "%s: write error\n", arg0);
		return (1);
	}

	if (fwrite (mb, 1, 128, out) != 128) {
		fprintf (stderr, "%s: write error\n", arg0);
		return (1);
	}

	ofs = 128;

	while (size > 0) {
		n = (size < 512) ? size : 512;

		if (fread (tar, 1, n, inp) != n) {
			fprintf (stderr, "%s: read error\n", arg0);
			return (1);
		}

		if (fwrite (tar, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", arg0);
			return (1);
		}

		size -= n;
		ofs += n;
	}

	if (ofs & 511) {
		n = 512 - (ofs & 511);

		memset (tar, 0, n);

		if (fwrite (tar, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", arg0);
			return (1);
		}
	}

	return (0);
}

static
int convert_folder_start (FILE *inp, FILE *out, const unsigned char *mb, char *prefix)
{
	char          *fname;
	unsigned char tar[512];
	unsigned long n;

	if (tar_set_mb (tar, mb)) {
		return (1);
	}

	n = strlen (prefix);

	fname = mb_get_name_cstr (mb);

	if ((n + strlen (fname) + 2) >= 256) {
		fprintf (stderr, "%s: folder name too long (%s)\n", arg0, fname);
		free (fname);
		return (1);
	}

	strcpy (prefix + n, fname);
	strcat (prefix + n, "/");

	free (fname);

	if (tar_set_fname (tar, prefix)) {
		fprintf (stderr, "%s: bad file name (%s)\n", arg0, prefix);
		return (1);
	}

	tar_set_size (tar, 0);

	tar_set_mode (tar, 0755);
	tar_set_type (tar, '5');

	tar_set_checksum (tar);

	if (fwrite (tar, 1, 512, out) != 512) {
		fprintf (stderr, "%s: write error\n", arg0);
		return (1);
	}

	if (par_folder_file != NULL) {
		if (tar_set_mb (tar, mb)) {
			return (1);
		}

		fname = mac_combine_path (prefix, par_folder_file);

		if (tar_set_fname (tar, fname)) {
			fprintf (stderr, "%s: bad file name (%s)\n", arg0, fname);
			free (fname);
			return (1);
		}

		free (fname);

		tar_set_size (tar, 128);

		tar_set_mode (tar, 0644);
		tar_set_type (tar, '0');

		tar_set_checksum (tar);

		if (fwrite (tar, 1, 512, out) != 512) {
			fprintf (stderr, "%s: write error\n", arg0);
			return (1);
		}

		memset (tar, 0, 512);
		memcpy (tar, mb, 128);

		if (fwrite (tar, 1, 512, out) != 512) {
			fprintf (stderr, "%s: write error\n", arg0);
			return (1);
		}

	}

	return (0);
}

static
int convert_folder_end (FILE *inp, FILE *out, char *prefix)
{
	unsigned i, j;

	i = 0;
	j = 0;

	while (prefix[i] != 0) {
		if ((prefix[i] == '/') && (prefix[i + 1] != 0)) {
			j = i + 1;
		}

		i += 1;
	}

	prefix[j] = 0;

	return (0);
}

static
int mbp_to_tar (FILE *inp, FILE *out)
{
	int           r;
	unsigned char mb[128];
	char          prefix[512];
	unsigned long creator;

	strcpy (prefix, "");

	r = 0;

	while (1) {
		if (fread (mb, 1, 128, inp) != 128) {
			if (ferror (inp)) {
				return(1);
			}

			break;
		}

		creator = mb_get_creator (mb);

		if (mb[0] == 0) {
			if (convert_file (inp, out, mb, prefix)) {
				return (1);
			}
		}
		else if (mb[0] == 1) {
			if (creator == 0xffffffff) {
				r = convert_folder_start (inp, out, mb, prefix);
			}
			else if (creator == 0xfffffffe) {
				r = convert_folder_end (inp, out, prefix);
			}
			else {
				fprintf (stderr,
					"%s: macbinary header error\n",
					arg0
				);
				return (1);
			}
		}
		else {
			fprintf (stderr, "%s: macbinary header error\n",
				arg0
			);
			return (1);
		}
	}

	memset (prefix, 0, 512);

	if (fwrite (prefix, 1, 512, out) != 512) {
		return (1);
	}

	if (fwrite (prefix, 1, 512, out) != 512) {
		return (1);
	}

	return (r);
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;
	char *finp, *fout;
	FILE *inp, *out;

	arg0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'f':
			par_folder_file = optarg[0];

			if (par_folder_file[0] == 0) {
				par_folder_file = NULL;
			}
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr,
					"%s: too many files (%s)\n",
					arg0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, "rb");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", arg0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, "wb");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", arg0, fout);
			return (EXIT_FAILURE);
		}
	}

	r = mbp_to_tar (inp, out);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
