/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/hfs/hfsfix.c                                             *
 * Created:     2011-07-05 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2012 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include "hfsfix.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/hfs.h>
#include <lib/mactime.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


static const char     *arg0 = NULL;

static unsigned       par_part = 0;

static unsigned long  par_offset = 0;
static unsigned long  par_blocks = 0;

static char           par_update_alt_mdb = 0;
static char           par_restore_alt_mdb = 0;

static const char     *par_vol_name = NULL;

static const char     *par_ctime = NULL;
static const char     *par_mtime = NULL;
static const char     *par_btime = NULL;

static char           par_vol_attr_ok = 0;
static unsigned short par_vol_attr = 0;

static char           par_write_count_ok = 0;
static unsigned long  par_write_count = 0;

static char           par_zero_boot = 0;
static char           par_zero_unused = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "update-alt-mdb", NULL, "Update the alternate MDB [no]" },
	{ 'A', 0, "restore-alt-mdb", NULL, "Restore the alternate MDB [no]" },
	{ 'B', 1, "btime", "string", "Set the backup time" },
	{ 'C', 1, "ctime", "string", "Set the creation time" },
	{ 'f', 0, "file", NULL, "Set an HFS file name" },
	{ 'M', 1, "mtime", "string", "Set the modification time" },
	{ 'n', 1, "blocks", "int", "Set the number of blocks in the volume [auto]" },
	{ 'o', 1, "offset", "int", "Set the starting offset [0]" },
	{ 'p', 1, "partition", "int", "Set the partition number [0]" },
	{ 't', 1, "volume-attributes", "int", "Set the volume attributes" },
	{ 'v', 1, "volume-name", "string", "Set the volume name" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'w', 1, "write-count", "int", "Set the write count" },
	{ 'z', 0, "zero-unused", NULL, "Clear unused space" },
	{ 'Z', 0, "zero-boot", NULL, "Clear the boot block" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"hfsfix: fix HFS data",
		"usage: hfsfix [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"hfsfix version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2011-2012 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}


static
void hfs_zero_name (hfs_vol_t *hfs)
{
	unsigned i;

	i = hfs->mdb[36];

	while (i < 27) {
		hfs->mdb[37 + i++] = 0;
	}
}

static
int hfs_zero_blocks (hfs_vol_t *hfs)
{
	unsigned      i;
	unsigned long j;
	unsigned      free_cnt, used_cnt;
	unsigned long idx, cnt;
	unsigned char buf[512];

	memset (buf, 0, 512);

	free_cnt = 0;
	used_cnt = 0;

	cnt = hfs->ablk_size / 512;
	idx = hfs->ablk_first;

	for (i = 0; i < hfs->ablk_cnt; i++) {
		if (hfs->bitmap[i >> 3] & (0x80 >> (i & 7))) {
			used_cnt += 1;
		}
		else {
			for (j = 0; j < cnt; j++) {
				if (hfs_write_block (hfs, buf, idx + j, 1)) {
					fprintf (stderr, "%s: write error\n",
						arg0
					);
					return (1);
				}
			}

			free_cnt += 1;
		}

		idx += cnt;
	}

	if (hfs->blocks > 2) {
		while (idx < (hfs->blocks - 2)) {
			if (hfs_write_block (hfs, buf, idx, 1)) {
				fprintf (stderr, "%s: write error\n", arg0);
				return (1);
			}

			idx += 1;
		}

		if (hfs_write_block (hfs, buf, hfs->blocks - 1, 1)) {
			fprintf (stderr, "%s: write error\n", arg0);
			return (1);
		}
	}

	return (0);
}

static
int hfs_zero_unused (hfs_vol_t *hfs)
{
	hfs_zero_name (hfs);

	if (hfs_zero_blocks (hfs)) {
		return (1);
	}

	return (0);
}

static
int hfs_fix_hfs (hfs_vol_t *hfs, unsigned long mtime)
{
	if (par_restore_alt_mdb) {
		hfs_restore_alt_mdb (hfs);
	}

	if (par_vol_name != NULL) {
		if (hfs_set_vol_name (hfs->mdb, par_vol_name)) {
			fprintf (stderr, "%s: bad volume name (%s)\n",
				arg0, par_vol_name
			);
			return (1);
		}
	}

	if (par_ctime != NULL) {
		if (strcmp (par_ctime, "file") == 0) {
			hfs_set_ctime (hfs->mdb, mtime);
		}
		else if (strcmp (par_ctime, "mtime") == 0) {
			hfs_set_ctime (hfs->mdb, hfs_get_mtime (hfs->mdb));
		}
		else {
			hfs_set_ctime (hfs->mdb, mac_time_from_string (par_ctime));
		}
	}

	if (par_mtime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			hfs_set_mtime (hfs->mdb, mtime);
		}
		else if (strcmp (par_mtime, "ctime") == 0) {
			hfs_set_mtime (hfs->mdb, hfs_get_ctime (hfs->mdb));
		}
		else {
			hfs_set_mtime (hfs->mdb, mac_time_from_string (par_mtime));
		}
	}

	if (par_btime != NULL) {
		if (strcmp (par_btime, "file") == 0) {
			hfs_set_btime (hfs->mdb, mtime);
		}
		else {
			hfs_set_btime (hfs->mdb, mac_time_from_string (par_btime));
		}
	}

	if (par_vol_attr_ok) {
		hfs_set_volume_attributes (hfs->mdb, par_vol_attr);
	}

	if (par_write_count_ok) {
		hfs_set_write_count (hfs->mdb, par_write_count);
	}

	if (par_zero_boot) {
		memset (hfs->boot, 0, 1024);
	}

	if (par_zero_unused) {
		hfs_zero_unused (hfs);
	}

	if (par_update_alt_mdb) {
		hfs_update_alt_mdb (hfs);
	}

	return (0);
}

static
int hfs_fix_fp (FILE *fp, unsigned long offset, unsigned long blocks, unsigned long *mtime)
{
	int       r;
	hfs_vol_t hfs;

	if (hfs_open_auto (&hfs, fp, offset, blocks, par_part)) {
		fprintf (stderr, "%s: opening HFS volume failed\n", arg0);
		return (1);
	}

	r = hfs_fix_hfs (&hfs, *mtime);

	*mtime = hfs_get_mtime (hfs.mdb);

	if (hfs_close (&hfs, 1)) {
		fprintf (stderr, "%s: write error\n", arg0);
		r = 1;
	}

	return (r);
}

static
int hfs_fix (const char *fname, unsigned long offset)
{
	int           r;
	unsigned long blocks, mtime;
	FILE          *fp;

	if (uf_stat (fname, &blocks, &mtime)) {
		fprintf (stderr, "%s: stat failed (%s)\n", arg0, fname);
		return (1);
	}

	if (par_blocks > 0) {
		blocks = par_blocks;
	}
	else {
		if (offset >= blocks) {
			fprintf (stderr, "%s: bad offset (%lu > %lu)\n",
				arg0, offset, blocks
			);
			return (1);
		}

		blocks = (blocks - offset) / 512;
	}

	fp = mac_open_inp (fname, "r+b", 0);

	if (fp == NULL) {
		fprintf (stderr, "%s: can't open file (%s)\n", arg0, fname);
		return (1);
	}

	r = hfs_fix_fp (fp, offset, blocks, &mtime);

	mac_close (fp);

	if (r == 0) {
		uf_set_mtime (fname, mtime);
	}

	return (r);
}

int main (int argc, char *argv[])
{
	int  r, ret;
	char **optarg;

	arg0 = argv[0];

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'a':
			par_update_alt_mdb = 1;
			break;

		case 'A':
			par_restore_alt_mdb = 1;
			break;

		case 'B':
			par_btime = optarg[0];
			break;

		case 'C':
			par_ctime = optarg[0];
			break;

		case 'f':
			ret |= hfs_fix (optarg[0], par_offset);
			break;

		case 'M':
			par_mtime = optarg[0];
			break;

		case 'n':
			par_blocks = strtoul (optarg[0], NULL, 0);
			break;

		case 'o':
			par_offset = strtoul (optarg[0], NULL, 0);
			break;

		case 'p':
			par_part = strtoul (optarg[0], NULL, 0);
			break;

		case 't':
			par_vol_attr = strtoul (optarg[0], NULL, 0);
			par_vol_attr_ok = 1;
			break;

		case 'v':
			par_vol_name = optarg[0];
			break;

		case 'w':
			par_write_count = strtoul (optarg[0], NULL, 0);
			par_write_count_ok = 1;
			break;

		case 'z':
			par_zero_unused = 1;
			break;

		case 'Z':
			par_zero_boot = 1;
			break;

		case 0:
			ret |= hfs_fix (optarg[0], par_offset);
			break;

		default:
			return (1);
		}
	}

	if (ret) {
		return (1);
	}

	return (0);
}
