/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/hfs/hfsinfo.c                                            *
 * Created:     2011-07-04 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2016 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include "hfsinfo.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/hfs.h>
#include <lib/mactime.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


#define TAG "%-24s "


static const char    *arg0 = NULL;

static unsigned      par_part = 0;

static unsigned long par_offset = 0;
static unsigned long par_blocks = 0;

static char          par_info_short = 0;
static char          par_info_mdb = 0;
static char          par_info_boot = 0;

static char          par_info_name = 0;
static char          par_info_ctime = 0;
static char          par_info_mtime = 0;

static char          par_cat = 0;

static char          par_first = 1;

static char          par_alt_mdb = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "alternate-mdb", NULL, "Use the alternate MDB [no]" },
	{ 'b', 0, "info-boot", NULL, "Print boot block info [no]" },
	{ 'c', 0, "cat", NULL, "Dump the volume to standard output [no]" },
	{ 'C', 0, "ctime", NULL, "Print the volume creation time [no]" },
	{ 'f', 1, "offset", "int", "Set the starting offset [0]" },
	{ 'i', 1, "input", "string", "Set an input file name [none]" },
	{ 'm', 0, "info-mdb", NULL, "Print MDB info [no]" },
	{ 'M', 0, "mtime", NULL, "Print the volume modification time [no]" },
	{ 'n', 1, "blocks", "int", "Set the number of blocks in the volume [auto]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'p', 1, "partition", "int", "Set the partition number [0]" },
	{ 's', 0, "info-short", NULL, "Print short info [yes]" },
	{ 'v', 0, "volume-name", NULL, "Print the volume name [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"hfsinfo: print information about HFS volumes",
		"usage: hfsinfo [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"hfsinfo version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2011-2016 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}


static
int hfs_cat (hfs_vol_t *hfs, FILE *fp)
{
	unsigned long i, n, max;
	unsigned char buf[8192];

	if (hfs->blocks == 0) {
		return (1);
	}

	max = 16;

	i = 0;
	while (i < hfs->blocks) {
		n = hfs->blocks - i;

		if (n > max) {
			n = max;
		}

		if (hfs_read_block (hfs, buf, i, n)) {
			if (max > 1) {
				max = 1;
				continue;
			}

			fprintf (stderr, "%s: read error at block %lu\n", arg0, i);

			memset (buf, 0, 512 * n);
		}

		if (fwrite (buf, 512, n, fp) != n) {
			return (1);
		}

		i += n;
	}

	return (0);
}

static
void hfs_print_hex16 (FILE *fp, const char *tag, unsigned val)
{
	fprintf (fp, TAG "0x%04x\n", tag, val);
}

static
void hfs_print_uint (FILE *fp, const char *tag, unsigned long val)
{
	fprintf (fp, TAG "%lu\n", tag, val);
}

static
void hfs_print_time (FILE *fp, const char *tag, unsigned long val)
{
	unsigned y, m, d, hrs, min, sec;

	mac_time_to_ymd (val, &y, &m, &d);
	mac_time_to_hms (val, &hrs, &min, &sec);

	fprintf (fp, TAG "%04u-%02u-%02u %02u:%02u:%02u (%lu)\n",
		tag, y, m + 1, d + 1, hrs, min, sec, val
	);
}

static
void hfs_print_time_short (FILE *fp, unsigned long val)
{
	unsigned y, m, d, hrs, min, sec;

	mac_time_to_ymd (val, &y, &m, &d);
	mac_time_to_hms (val, &hrs, &min, &sec);

	fprintf (fp, "%04u-%02u-%02u %02u:%02u:%02u\n",
		y, m + 1, d + 1, hrs, min, sec
	);
}

static
void hfs_print_pstr (FILE *fp, const char *tag, const unsigned char *str, unsigned max)
{
	char *tmp;

	if (str[0] >= max) {
		return;
	}

	tmp = mac_pstring_decode (str, "");

	if (tmp == NULL) {
		return;
	}

	fprintf (fp, TAG "\"%s\"\n", tag, tmp);

	free (tmp);
}


static
int hfs_print_info_boot (FILE *fp, hfs_vol_t *hfs, const unsigned char *boot)
{
	if (get_uint16_be (hfs->boot, 0) != HFS_BOOT_ID) {
		return (1);
	}

	hfs_print_hex16 (fp, "Boot block version:", get_uint16_be (boot, 6));
	hfs_print_pstr (fp, "System:", boot + 10, 16);
	hfs_print_pstr (fp, "Shell:", boot + 26, 16);
	hfs_print_pstr (fp, "Debugger:", boot + 42, 16);
	hfs_print_pstr (fp, "Debugger 2:", boot + 58, 16);
	hfs_print_pstr (fp, "Startup screen:", boot + 74, 16);
	hfs_print_pstr (fp, "Startup program:", boot + 90, 16);
	hfs_print_pstr (fp, "Scrap file:", boot + 106, 16);

	return (0);
}

static
int hfs_print_info_mdb (FILE *fp, hfs_vol_t *hfs, const unsigned char *mdb)
{
	if (hfs_get_mdb_id (mdb) != HFS_MDB_ID) {
		return (1);
	}

	hfs_print_pstr (fp, "Volume name:", mdb + 36, 28);
	fprintf (fp, TAG "%lu K (%lu)\n", "Volume size:", hfs->blocks / 2, 512 * hfs->blocks);
	hfs_print_time (fp, "Volume ctime:", hfs_get_ctime (mdb));
	hfs_print_time (fp, "Volume mtime:", hfs_get_mtime (mdb));
	hfs_print_time (fp, "Volume btime:", hfs_get_btime (mdb));
	hfs_print_hex16 (fp, "MDB signature:", hfs_get_mdb_id (mdb));
	hfs_print_hex16 (fp, "Volume attributes:", hfs_get_volume_attributes (mdb));
	hfs_print_uint (fp, "Root directory files:", get_uint16_be (mdb, 12));
	hfs_print_uint (fp, "Root directory dirs:", get_uint16_be (mdb, 82));
	hfs_print_uint (fp, "First bitmap block:", get_uint16_be (mdb, 14));
	hfs_print_uint (fp, "Allocation blocks:", hfs_get_ablk_count (mdb));
	hfs_print_uint (fp, "Allocation block size:", hfs_get_ablk_size (mdb));
	hfs_print_uint (fp, "Clump size:", get_uint32_be (mdb, 24));
	hfs_print_uint (fp, "First allocation block:", hfs_get_first_ablk (mdb));
	hfs_print_uint (fp, "Free allocation blocks:", hfs_get_free_ablk (mdb));
	hfs_print_uint (fp, "Backup sequence:", get_uint16_be (mdb, 68));
	hfs_print_uint (fp, "Write count:", hfs_get_write_count (mdb));
	hfs_print_uint (fp, "Clump size for extents:", get_uint32_be (mdb, 74));
	hfs_print_uint (fp, "Clump size for catalog:", get_uint32_be (mdb, 78));
	hfs_print_uint (fp, "Files:", get_uint32_be (mdb, 84));
	hfs_print_uint (fp, "Directories:", get_uint32_be (mdb, 88));
	hfs_print_uint (fp, "Extents file size:", get_uint32_be (mdb, 130));
	hfs_print_uint (fp, "Catalog file size:", get_uint32_be (mdb, 146));

	return (0);
}

static
int hfs_print_info_short (FILE *fp, hfs_vol_t *hfs, const unsigned char *mdb)
{
	if (hfs_get_mdb_id (mdb) != HFS_MDB_ID) {
		return (1);
	}

	hfs_print_pstr (fp, "Volume name:", mdb + 36, 28);
	fprintf (fp, TAG "%lu K (%lu)\n", "Volume size:", hfs->blocks / 2, 512 * hfs->blocks);
	hfs_print_time (fp, "Volume ctime:", hfs_get_ctime (mdb));
	hfs_print_time (fp, "Volume mtime:", hfs_get_mtime (mdb));
	hfs_print_time (fp, "Volume btime:", hfs_get_btime (mdb));
	hfs_print_hex16 (fp, "Volume attributes:", hfs_get_volume_attributes (mdb));
	fprintf (fp, TAG "%u * %lu (%u free)\n", "Allocation blocks:",
		hfs_get_ablk_count (mdb), hfs_get_ablk_size (mdb), hfs_get_free_ablk (mdb)
	);
	hfs_print_uint (fp, "Write count:", hfs_get_write_count (mdb));
	fprintf (fp, TAG "%lu files, %lu directories\n", "Entries:",
		get_uint32_be (mdb, 84), get_uint32_be (mdb, 88)
	);

	return (0);
}

static
int hfs_print_info_name (FILE *fp, hfs_vol_t *hfs, const unsigned char *mdb)
{
	char *str;

	str = hfs_get_vol_name (mdb);

	if (str == NULL) {
		return (1);
	}

	fprintf (fp, "%s\n", str);

	free (str);

	return (0);
}

static
int hfs_info_fp (hfs_vol_t *hfs, FILE *out, const char *fname)
{
	int                 done;
	const unsigned char *mdb;

	if (par_alt_mdb) {
		mdb = hfs->alt_mdb;
	}
	else {
		mdb = hfs->mdb;
	}

	if (par_info_name) {
		return (hfs_print_info_name (out, hfs, mdb));
	}
	else if (par_info_ctime) {
		hfs_print_time_short (out, hfs_get_ctime (mdb));
		return (0);
	}
	else if (par_info_mtime) {
		hfs_print_time_short (out, hfs_get_mtime (mdb));
		return (0);
	}

	done = 0;

	if (par_first) {
		par_first = 0;
	}
	else {
		fputs ("\n", out);
	}

	fprintf (out, TAG "%s\n", "File name:", fname);

	if (par_info_boot) {
		hfs_print_info_boot (out, hfs, hfs->boot);
		done = 1;
	}

	if (par_info_mdb) {
		hfs_print_info_mdb (out, hfs, mdb);
		done = 1;
	}

	if (done == 0) {
		hfs_print_info_short (out, hfs, par_alt_mdb ? hfs->alt_mdb : hfs->mdb);
	}

	return (0);
}

static
int hfs_info (const char *finp, const char *fout, unsigned long offset)
{
	int           r;
	unsigned long blocks, mtime;
	FILE          *inp, *out;
	hfs_vol_t     hfs;

	if (uf_stat (finp, &blocks, &mtime)) {
		fprintf (stderr, "%s: stat failed (%s)\n", arg0, finp);
		return (1);
	}

	if (par_blocks > 0) {
		blocks = par_blocks;
	}
	else {
		if (offset >= blocks) {
			fprintf (stderr, "%s: bad offset (%lu > %lu)\n",
				arg0, offset, blocks
			);
			return (1);
		}

		blocks = (blocks - offset) / 512;
	}

	inp = mac_open_inp (finp, "rb", 0);

	if (inp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n",
			arg0, finp
		);
		return (1);
	}

	if (hfs_open_auto (&hfs, inp, offset, blocks, par_part)) {
		fprintf (stderr, "%s: opening HFS volume failed (%s)\n",
			arg0, finp
		);
		mac_close (inp);
		return (1);
	}

	out = mac_open_out (fout, par_cat ? "wb" : "w", 1);

	if (out == NULL) {
		fprintf (stderr, "%s: can't open output file (%s)\n",
			arg0, fout
		);
		r = 1;
	}
	else if (par_cat) {
		r = hfs_cat (&hfs, out);
	}
	else {
		r = hfs_info_fp (&hfs, out, finp);
	}

	hfs_close (&hfs, 0);

	mac_close (out);
	mac_close (inp);

	return (r);
}

int main (int argc, char *argv[])
{
	int        r, ret;
	char       **optarg;
	const char *out;

	arg0 = argv[0];

	out = NULL;

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'a':
			par_alt_mdb = 1;
			break;

		case 'b':
			par_info_boot = 1;
			break;

		case 'c':
			par_cat = 1;
			break;

		case 'C':
			par_info_ctime = 1;
			break;

		case 'f':
			par_offset = strtoul (optarg[0], NULL, 0);
			break;

		case 'i':
			ret |= hfs_info (optarg[0], out, par_offset);
			break;

		case 'm':
			par_info_mdb = 1;
			break;

		case 'M':
			par_info_mtime = 1;
			break;

		case 'n':
			par_blocks = strtoul (optarg[0], NULL, 0);
			break;

		case 'o':
			out = optarg[0];
			break;

		case 'p':
			par_part = strtoul (optarg[0], NULL, 0);
			break;

		case 's':
			par_info_short = 1;
			break;

		case 'v':
			par_info_name = 1;
			break;

		case 0:
			ret |= hfs_info (optarg[0], out, par_offset);
			break;

		default:
			return (1);
		}
	}

	if (ret) {
		return (1);
	}

	return (0);
}
