/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/hfs.c                                                *
 * Created:     2011-07-05 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2012 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/hfs.h>
#include <lib/strings.h>


int hfs_open (hfs_vol_t *hfs, FILE *fp, unsigned long offset, unsigned long blocks)
{
	hfs->fp = fp;

	hfs->offset = offset;
	hfs->blocks = blocks;

	hfs->readonly = 0;

	if (hfs_read_block (hfs, hfs->boot, 0, 2)) {
		return (1);
	}

	if (hfs_read_block (hfs, hfs->mdb, 2, 1)) {
		return (1);
	}

	if (get_uint16_be (hfs->mdb, 0) != HFS_MDB_ID) {
		return (1);
	}

	hfs->ablk_first = get_uint16_be (hfs->mdb, 28);
	hfs->ablk_cnt = get_uint16_be (hfs->mdb, 18);
	hfs->ablk_size = get_uint32_be (hfs->mdb, 20);

	hfs->bitmap_start = get_uint16_be (hfs->mdb, 14);
	hfs->bitmap_size = ((unsigned long) hfs->ablk_cnt + 4095) / 4096;

	if (hfs->bitmap_size > 16) {
		return (1);
	}

	if (hfs_read_block (hfs, hfs->bitmap, hfs->bitmap_start, hfs->bitmap_size)) {
		return (1);
	}

	if (hfs->blocks > 2) {
		if (hfs_read_block (hfs, hfs->alt_mdb, hfs->blocks - 2, 1)) {
			memset (hfs->alt_mdb, 0, 512);
		}
	}
	else {
		memset (hfs->alt_mdb, 0, 512);
	}

	return (0);
}

static
int hfs_open_part (hfs_vol_t *hfs, FILE *fp, unsigned char *buf, unsigned long devblk)
{
	unsigned long ptidx, ptblk;

	if (get_uint16_be (buf, 0) != HFS_PMAP_ID) {
		return (1);
	}

	if (strcmp ((char *) buf + 48, "Apple_HFS") != 0) {
		return (1);
	}

	ptidx = get_uint32_be (buf, 8);
	ptblk = get_uint32_be (buf, 12);

	if (devblk > 0) {
		if (ptidx >= devblk) {
			return (1);
		}

		if ((ptidx + ptblk) > devblk) {
			return (1);
		}
	}

	return (hfs_open (hfs, fp, 512 * ptidx, ptblk));
}

int hfs_open_pm (hfs_vol_t *hfs, FILE *fp, unsigned part)
{
	unsigned long i;
	unsigned      blksize;
	unsigned long devblk, pmblk;
	unsigned char buf[512];

	if (mac_read (fp, buf, 0, 512)) {
		return (1);
	}

	if (get_uint16_be (buf, 0) != HFS_DEVICE_ID) {
		return (1);
	}

	blksize = get_uint16_be (buf, 2);
	devblk = get_uint32_be (buf, 4);

	if (mac_read (fp, buf, 512, 512)) {
		return (1);
	}

	if (get_uint16_be (buf, 0) != HFS_PMAP_ID) {
		return (1);
	}

	pmblk = get_uint32_be (buf, 4);

	if (devblk > 0) {
		if (pmblk >= devblk) {
			return (1);
		}
	}

	if (part > 0) {
		if (part > pmblk) {
			return (1);
		}

		if (mac_read (fp, buf, 512 * part, 512)) {
			return (1);
		}

		return (hfs_open_part (hfs, fp, buf, (blksize / 512) * devblk));
	}

	for (i = 0; i < pmblk; i++) {
		if (mac_read (fp, buf, 512 + 512 * i, 512)) {
			return (1);
		}

		if (hfs_open_part (hfs, fp, buf, (blksize / 512) * devblk) == 0) {
			return (0);
		}
	}

	return (1);
}

int hfs_open_dc42 (hfs_vol_t *hfs, FILE *fp)
{
	int           r;
	unsigned long datasize;
	unsigned char buf[128];

	if (mac_read (fp, buf, 0, 84)) {
		return (1);
	}

	if (buf[0] > 63) {
		return (1);
	}

	if (get_uint16_be (buf, 82) != 0x0100) {
		return (1);
	}

	datasize = get_uint32_be (buf, 64);

	if ((datasize < 512) || (datasize & 511)) {
		return (1);
	}

	r = hfs_open (hfs, fp, 84, datasize / 512);

	hfs->readonly = 1;

	return (r);
}

int hfs_open_auto (hfs_vol_t *hfs, FILE *fp, unsigned long offset, unsigned long blocks, unsigned part)
{
	if (part > 0) {
		return (hfs_open_pm (hfs, fp, part));
	}

	if (hfs_open (hfs, fp, offset, blocks) == 0) {
		return (0);
	}

	if (hfs_open_dc42 (hfs, fp) == 0) {
		return (0);
	}

	if (hfs_open_pm (hfs, fp, part) == 0) {
		return (0);
	}

	return (1);
}

int hfs_close (hfs_vol_t *hfs, int write)
{
	if (write == 0) {
		return (0);
	}

	if (hfs_write_block (hfs, hfs->boot, 0, 2)) {
		return (1);
	}

	if (hfs_write_block (hfs, hfs->mdb, 2, 1)) {
		return (1);
	}

	if (hfs_write_block (hfs, hfs->bitmap, hfs->bitmap_start, hfs->bitmap_size)) {
		return (1);
	}

	if (hfs->blocks > 2) {
		if (hfs_write_block (hfs, hfs->alt_mdb, hfs->blocks - 2, 1)) {
			return (1);
		}
	}

	return (0);
}

int hfs_read_block (hfs_vol_t *hfs, void *buf, unsigned long idx, unsigned long cnt)
{
	if ((hfs->blocks > 0) && ((idx + cnt) > hfs->blocks)) {
		return (1);
	}

	if (mac_read (hfs->fp, buf, hfs->offset + 512 * idx, 512 * cnt)) {
		return (1);
	}

	return (0);
}

int hfs_write_block (hfs_vol_t *hfs, const void *buf, unsigned long idx, unsigned long cnt)
{
	if (hfs->readonly) {
		return (1);
	}

	if ((hfs->blocks > 0) && ((idx + cnt) > hfs->blocks)) {
		return (1);
	}

	if (mac_write (hfs->fp, buf, hfs->offset + 512 * idx, 512 * cnt)) {
		return (1);
	}

	return (0);
}


unsigned hfs_get_mdb_id (const void *mdb)
{
	return (get_uint16_be (mdb, 0));
}

void hfs_set_mdb_id (void *mdb, unsigned val)
{
	set_uint16_be (mdb, 0, val);
}


char *hfs_get_vol_name (const void *mdb)
{
	const unsigned char *str;

	str = (const unsigned char *) mdb + 36;

	if (str[0] > 27) {
		return (NULL);
	}

	return (mac_pstring_decode (str, ""));
}

int hfs_set_vol_name (void *mdb, const char *val)
{
	unsigned char *str;

	str = (unsigned char *) mdb + 36;

	memset (str, 0, 28);

	return (mac_pstring_encode_buf (str, 36, val));
}

unsigned long hfs_get_ctime (const void *mdb)
{
	return (get_uint32_be (mdb, 2));
}

void hfs_set_ctime (void *mdb, unsigned long val)
{
	set_uint32_be (mdb, 2, val);
}


unsigned long hfs_get_mtime (const void *mdb)
{
	return (get_uint32_be (mdb, 6));
}

void hfs_set_mtime (void *mdb, unsigned long val)
{
	set_uint32_be (mdb, 6, val);
}


unsigned long hfs_get_btime (const void *mdb)
{
	return (get_uint32_be (mdb, 64));
}

void hfs_set_btime (void *mdb, unsigned long val)
{
	set_uint32_be (mdb, 64, val);
}


unsigned hfs_get_volume_attributes (const void *mdb)
{
	return (get_uint16_be (mdb, 10));
}

void hfs_set_volume_attributes (void *mdb, unsigned val)
{
	set_uint16_be (mdb, 10, val);
}


unsigned hfs_get_ablk_count (const void *mdb)
{
	return (get_uint16_be (mdb, 18));
}

void hfs_set_ablk_count (void *mdb, unsigned val)
{
	set_uint16_be (mdb, 18, val);
}


unsigned long hfs_get_ablk_size (const void *mdb)
{
	return (get_uint32_be (mdb, 20));
}

void hfs_set_ablk_size (void *mdb, unsigned long val)
{
	set_uint32_be (mdb, 20, val);
}


unsigned hfs_get_first_ablk (const void *mdb)
{
	return (get_uint16_be (mdb, 28));
}

void hfs_set_first_ablk (void *mdb, unsigned val)
{
	set_uint16_be (mdb, 28, val);
}


unsigned hfs_get_free_ablk (const void *mdb)
{
	return (get_uint16_be (mdb, 34));
}

void hfs_set_free_ablk (void *mdb, unsigned val)
{
	set_uint16_be (mdb, 34, val);
}


unsigned long hfs_get_write_count (const void *mdb)
{
	return (get_uint32_be (mdb, 70));
}

void hfs_set_write_count (void *mdb, unsigned long val)
{
	set_uint32_be (mdb, 70, val);
}


void hfs_get_finder_info (const void *mdb, void *buf)
{
	unsigned            i;
	const unsigned char *src;
	unsigned char       *dst;

	src = (const unsigned char *) mdb + 92;
	dst = buf;

	for (i = 0; i < 16; i++) {
		dst[i] = src[i];
	}
}

void hfs_set_finder_info (void *mdb, const void *buf)
{
	unsigned            i;
	const unsigned char *src;
	unsigned char       *dst;

	src = buf;
	dst = (unsigned char *) mdb + 92;

	for (i = 0; i < 16; i++) {
		dst[i] = src[i];
	}
}


void hfs_update_alt_mdb (hfs_vol_t *hfs)
{
	memcpy (hfs->alt_mdb, hfs->mdb, 512);
}

void hfs_restore_alt_mdb (hfs_vol_t *hfs)
{
	if (hfs->blocks > 2) {
		memcpy (hfs->mdb, hfs->alt_mdb, 512);
	}
}
