/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/hqx.c                                                *
 * Created:     2005-04-07 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <lib/crc.h>
#include <lib/hqx.h>
#include <lib/strings.h>


static unsigned char hqx_dectab[128] = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* 00 */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* 10 */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, /* 20 */
	0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0xff, 0xff,
	0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0xff, /* 30 */
	0x14, 0x15, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, /* 40 */
	0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0xff,
	0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0xff, /* 50 */
	0x2c, 0x2d, 0x2e, 0x2f, 0xff, 0xff, 0xff, 0xff,
	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0xff, /* 60 */
	0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0xff, 0xff,
	0x3d, 0x3e, 0x3f, 0xff, 0xff, 0xff, 0xff, 0xff, /* 70 */
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
};

static unsigned char hqx_enctab[64] = {
	0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28,
	0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x30, 0x31, 0x32,
	0x33, 0x34, 0x35, 0x36, 0x38, 0x39, 0x40, 0x41,
	0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49,
	0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x50, 0x51, 0x52,
	0x53, 0x54, 0x55, 0x56, 0x58, 0x59, 0x5a, 0x5b,
	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x68,
	0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x70, 0x71, 0x72
};


unsigned char hqx_get_uint8 (const void *buf, unsigned i)
{
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	return (tmp[0] & 0xff);
}

unsigned short hqx_get_uint16 (const void *buf, unsigned i)
{
	unsigned short      r;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	r = tmp[0] & 0xff;
	r = (r << 8) | (tmp[1] & 0xff);

	return (r);
}

unsigned long hqx_get_uint32 (const void *buf, unsigned i)
{
	unsigned long       r;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	r = tmp[0] & 0xff;
	r = (r << 8) | (tmp[1] & 0xff);
	r = (r << 8) | (tmp[2] & 0xff);
	r = (r << 8) | (tmp[3] & 0xff);

	return (r);
}

void hqx_set_uint8 (void *buf, unsigned i, unsigned char v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = v & 0xff;
}

void hqx_set_uint16 (void *buf, unsigned i, unsigned short v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (v >> 8) & 0xff;
	tmp[1] = v & 0xff;
}

void hqx_set_uint32 (void *buf, unsigned i, unsigned long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (v >> 24) & 0xff;
	tmp[1] = (v >> 16) & 0xff;
	tmp[2] = (v >> 8) & 0xff;
	tmp[3] = v & 0xff;
}


static
void hqx_init (hqx_t *hqx, FILE *fp)
{
	unsigned i;

	hqx->fp = fp;

	hqx->eof = 0;
	hqx->rle = 0;

	hqx->column = 0;

	hqx->bufidx = 0;
	hqx->bufcnt = 0;

	hqx->runcnt = 0;
	hqx->runchr = 256;

	hqx->fname[0] = 0;

	for (i = 0; i < 32; i++) {
		hqx->header[i] = 0;
	}
}

int hqx_init_read (hqx_t *hqx, FILE *fp)
{
	int c;
	int first;

	hqx_init (hqx, fp);

	first = 1;

	while (1) {
		c = fgetc (hqx->fp);

		if (c == EOF) {
			return (1);
		}

		if (first && (c == ':')) {
			return (0);
		}

		if ((c == 0x0d) || (c == 0x0a)) {
			first = 1;
		}
		else {
			first = 0;
		}
	}

	return (1);
}

static
int hqx_get_char (hqx_t *hqx, unsigned char *val)
{
	int c;

	if (hqx->eof) {
		return (1);
	}

	while (1) {
		c = fgetc (hqx->fp);

		if ((c == EOF) || (c == ':')) {
			hqx->eof = 1;
			return (1);
		}

		if ((c > 0) && (c < 128)) {
			*val = hqx_dectab[c];
			if (*val < 0x40) {
				return (0);
			}
		}
	}
}

int hqx_get_byte_raw (hqx_t *hqx, unsigned char *val)
{
	unsigned      i;
	unsigned char buf[4];

	if (hqx->bufidx < hqx->bufcnt) {
		*val = hqx->buf[hqx->bufidx];
		hqx->bufidx += 1;
		return (0);
	}

	if (hqx->eof) {
		return (1);
	}

	i = 0;
	while (i < 4) {
		if (hqx_get_char (hqx, &buf[i])) {
			break;
		}

		buf[i] &= 0x3f;

		i += 1;
	}

	if (i == 0) {
		return (1);
	}

	hqx->bufcnt = (i < 4) ? i : 3;

	while (i < 4) {
		buf[i++] = 0;
	}

	hqx->buf[0] = ((buf[0] << 2) | (buf[1] >> 4)) & 0xff;
	hqx->buf[1] = ((buf[1] << 4) | (buf[2] >> 2)) & 0xff;
	hqx->buf[2] = ((buf[2] << 6) | buf[3]) & 0xff;

	*val = hqx->buf[0];

	hqx->bufidx = 1;

	return (0);
}

int hqx_get_byte (hqx_t *hqx, unsigned char *val)
{
	unsigned char c;

	if (hqx->runcnt > 0) {
		hqx->runcnt -= 1;
		*val = hqx->runchr;
		return (0);
	}

	while (1) {
		if (hqx_get_byte_raw (hqx, &c)) {
			return (1);
		}

		if (c != 0x90) {
			*val = c;
			hqx->runchr = c;
			return (0);
		}

		if (hqx_get_byte_raw (hqx, &c)) {
			return (1);
		}

		if (c == 0) {
			*val = 0x90;
			hqx->runchr = 0x90;
			return (0);
		}

		if (c > 1) {
			*val = hqx->runchr;
			hqx->runcnt = c - 2;
			return (0);
		}
	}

	return (1);
}

int hqx_read (hqx_t *hqx, void *buf, unsigned cnt)
{
	unsigned      i;
	unsigned char *tmp;

	tmp = buf;

	for (i = 0; i < cnt; i++) {
		if (hqx_get_byte (hqx, &tmp[i])) {
			return (1);
		}
	}

	return (0);
}


void hqx_init_write (hqx_t *hqx, FILE *fp, int rle)
{
	hqx_init (hqx, fp);

	hqx->rle = (rle != 0);

	fprintf (fp, "(This file must be converted with BinHex 4.0)\n:");

	hqx->column = 1;
}

static
void hqx_set_char (hqx_t *hqx, unsigned char c)
{
	if (hqx->column < 64) {
		hqx->column += 1;
	}
	else {
		hqx->column = 1;
		fputc ('\n', hqx->fp);
	}

	fputc (hqx_enctab[c & 0x3f], hqx->fp);
}

void hqx_set_byte_raw (hqx_t *hqx, unsigned char c)
{
	unsigned      i;
	unsigned char buf[4];

	hqx->buf[hqx->bufidx++] = c;

	if (hqx->bufidx < 3) {
		return;
	}

	buf[0] = (hqx->buf[0] >> 2) & 0x3f;
	buf[1] = ((hqx->buf[0] & 0x03) << 4) | ((hqx->buf[1] >> 4) & 0x0f);
	buf[2] = ((hqx->buf[1] & 0x0f) << 2) | ((hqx->buf[2] >> 6) & 0x03);
	buf[3] = hqx->buf[2] & 0x3f;

	for (i = 0; i < 4; i++) {
		hqx_set_char (hqx, buf[i]);
	}

	hqx->bufidx = 0;
}

static
void hqx_set_byte_esc (hqx_t *hqx, unsigned char c)
{
	hqx_set_byte_raw (hqx, c);

	if (c == 0x90) {
		hqx_set_byte_raw (hqx, 0x00);
	}
}

static
void hqx_set_run (hqx_t *hqx, unsigned cnt, unsigned char c)
{
	unsigned i;

	if ((cnt < 3) && (c != 0x90)) {
		for (i = 0; i < cnt; i++) {
			hqx_set_byte_esc (hqx, c);
		}
	}
	else {
		hqx_set_byte_esc (hqx, c);

		if (cnt > 1) {
			hqx_set_byte_raw (hqx, 0x90);
			hqx_set_byte_raw (hqx, cnt);
		}
	}
}

void hqx_set_byte (hqx_t *hqx, unsigned char c)
{
	if (hqx->rle == 0) {
		hqx_set_byte_esc (hqx, c);
		return;
	}

	if ((c == hqx->runchr) && (hqx->runcnt < 255)) {
		hqx->runcnt += 1;
		return;
	}

	if (hqx->runcnt > 0) {
		hqx_set_run (hqx, hqx->runcnt, hqx->runchr);
	}

	hqx->runcnt = 1;
	hqx->runchr = c;
}

void hqx_write (hqx_t *hqx, const void *buf, unsigned cnt)
{
	unsigned            i;
	const unsigned char *tmp;

	tmp = buf;

	for (i = 0; i < cnt; i++) {
		hqx_set_byte (hqx, tmp[i]);
	}
}

void hqx_flush (hqx_t *hqx)
{
	if (hqx->rle) {
		if (hqx->runcnt > 0) {
			hqx_set_run (hqx, hqx->runcnt, hqx->runchr);
		}
	}

	while (hqx->bufidx > 0) {
		hqx_set_byte_raw (hqx, 0);
	}

	fputs (":\n", hqx->fp);

	fflush (hqx->fp);
}


int hqx_read_header (hqx_t *hqx)
{
	unsigned char cnt;

	if (hqx_get_byte (hqx, &cnt)) {
		return (1);
	}

	hqx->fname[0] = cnt;

	if (hqx_read (hqx, hqx->fname + 1, cnt)) {
		return (1);
	}

	if (hqx_read (hqx, hqx->header, 21)) {
		return (1);
	}

	return (0);
}

int hqx_write_header (hqx_t *hqx)
{
	hqx_set_byte (hqx, hqx->fname[0]);
	hqx_write (hqx, hqx->fname + 1, hqx->fname[0]);
	hqx_write (hqx, hqx->header, 21);

	return (0);
}

unsigned short hqx_calc_crc (const hqx_t *hqx)
{
	crc16_t crc;

	crc16_init (&crc, CRC_CRC16_CCITT, 0);

	crc16_calc (&crc, hqx->fname, hqx->fname[0] + 1);
	crc16_calc (&crc, hqx->header, 19);
	crc16_done (&crc);

	return (crc16_get_crc (&crc));
}

int hqx_check_crc (const hqx_t *hqx)
{
	return (hqx_calc_crc (hqx) != hqx_get_crc (hqx));
}

const unsigned char *hqx_get_name_pstr (const hqx_t *hqx)
{
	return (hqx->fname);
}

void hqx_set_name_pstr (hqx_t *hqx, const void *str)
{
	const unsigned char *tmp = str;

	memcpy (hqx->fname, tmp, tmp[0] + 1);
}

char *hqx_get_name_cstr (const hqx_t *hqx)
{
	return (mac_pstring_decode (hqx->fname, "/"));
}

int hqx_set_name_cstr (hqx_t *hqx, const char *str)
{
	return (mac_pstring_encode_buf (hqx->fname, 256, str));
}

unsigned hqx_get_version (const hqx_t *hqx)
{
	return (hqx_get_uint8 (hqx->header, 0));
}

void hqx_set_version (hqx_t *hqx, unsigned val)
{
	hqx_set_uint8 (hqx->header, 0, val);
}

unsigned long hqx_get_type (const hqx_t *hqx)
{
	return (hqx_get_uint32 (hqx->header, 1));
}

void hqx_set_type (hqx_t *hqx, unsigned long val)
{
	hqx_set_uint32 (hqx->header, 1, val);
}

unsigned long hqx_get_creator (const hqx_t *hqx)
{
	return (hqx_get_uint32 (hqx->header, 5));
}

void hqx_set_creator (hqx_t *hqx, unsigned long val)
{
	hqx_set_uint32 (hqx->header, 5, val);
}

unsigned short hqx_get_finder_flags (const hqx_t *hqx)
{
	return (hqx_get_uint16 (hqx->header, 9));
}

void hqx_set_finder_flags (hqx_t *hqx, unsigned short val)
{
	hqx_set_uint16 (hqx->header, 9, val);
}

unsigned long hqx_get_data_size (const hqx_t *hqx)
{
	return (hqx_get_uint32 (hqx->header, 11));
}

void hqx_set_data_size (hqx_t *hqx, unsigned long val)
{
	hqx_set_uint32 (hqx->header, 11, val);
}

unsigned long hqx_get_rsrc_size (const hqx_t *hqx)
{
	return (hqx_get_uint32 (hqx->header, 15));
}

void hqx_set_rsrc_size (hqx_t *hqx, unsigned long val)
{
	hqx_set_uint32 (hqx->header, 15, val);
}

unsigned short hqx_get_crc (const hqx_t *hqx)
{
	return (hqx_get_uint16 (hqx->header, 19));
}

void hqx_set_crc (hqx_t *hqx, unsigned short val)
{
	hqx_set_uint16 (hqx->header, 19, val);
}

void hqx_get_macfile (const hqx_t *hqx, macfile_t *mf)
{
	mf_set_name_pstr (mf, hqx_get_name_pstr (hqx));

	mf_set_data_size (mf, hqx_get_data_size (hqx));
	mf_set_rsrc_size (mf, hqx_get_rsrc_size (hqx));

	mf_set_type (mf, hqx_get_type (hqx));
	mf_set_creator (mf, hqx_get_creator (hqx));
	mf_set_finder_flags (mf, hqx_get_finder_flags (hqx));
}

void hqx_set_macfile (hqx_t *hqx, const macfile_t *mf)
{
	hqx_set_name_pstr (hqx, mf_get_name_pstr (mf));

	hqx_set_data_size (hqx, mf_get_data_size (mf));
	hqx_set_rsrc_size (hqx, mf_get_rsrc_size (mf));

	hqx_set_type (hqx, mf_get_type (mf));
	hqx_set_creator (hqx, mf_get_creator (mf));
	hqx_set_finder_flags (hqx, mf_get_finder_flags (mf));
}
