/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/macbinary.c                                          *
 * Created:     2004-02-01 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2011 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/macbinary.h>
#include <lib/macfile.h>
#include <lib/strings.h>


void mb_init (void *mb)
{
	unsigned      i;
	unsigned char *buf;

	buf = (unsigned char *) mb;

	for (i = 0; i < 128; i++) {
		buf[i] = 0;
	}
}

int mb_read (void *mb, FILE *fp)
{
	if (fread (mb, 1, 128, fp) != 128) {
		return (1);
	}

	return (0);
}

int mb_write (const void *mb, FILE *fp)
{
	if (fwrite (mb, 1, 128, fp) != 128) {
		return (1);
	}

	return (0);
}

unsigned short mb_calc_crc (const void *mb)
{
	return (crc16_crc (mb, 124, CRC_CRC16_CCITT, 0));
}

unsigned mb_get_version (const void *mb)
{
	const unsigned char *buf;

	buf = (const unsigned char *) mb;

	if ((buf[0] != 0) && (buf[0] != 1)) {
		return (0);
	}

	if (buf[74] != 0x00) {
		return (0);
	}

	if ((buf[1] < 1) || (buf[1] > 63)) {
		return (0);
	}

	if (get_uint16_be (mb, 124) == crc16_crc (mb, 124, CRC_CRC16_CCITT, 0)) {
		return (2);
	}

	if (buf[82] != 0x00) {
		return (0);
	}

	return (1);
}

unsigned mb_get_id (const void *mb)
{
	return (get_uint8_be (mb, 0));
}

void mb_set_id (void *mb, unsigned id)
{
	set_uint8_be (mb, 0, id);
}

const unsigned char *mb_get_name_pstr (const void *mb)
{
	const unsigned char *tmp = mb;

	return (tmp + 1);
}

int mb_set_name_pstr (void *mb, const void *str)
{
	unsigned char       *d;
	const unsigned char *s;

	d = mb;
	s = str;

	if (s[0] > 63) {
		return (1);
	}

	memset (d + 1, 0, 64);
	memcpy (d + 1, s, s[0] + 1);

	return (0);
}

char *mb_get_name_cstr (const void *mb)
{
	const unsigned char *tmp = mb;

	return (mac_pstring_decode (tmp + 1, "/"));
}

int mb_set_name_cstr (void *mb, const char *str)
{
	unsigned char *tmp = mb;

	memset (tmp + 1, 0, 64);

	return (mac_pstring_encode_buf (tmp + 1, 64, str));
}

unsigned long mb_get_type (const void *mb)
{
	return (get_uint32_be (mb, 65));
}

void mb_set_type (void *mb, unsigned long v)
{
	set_uint32_be (mb, 65, v);
}

unsigned long mb_get_creator (const void *mb)
{
	return (get_uint32_be (mb, 69));
}

void mb_set_creator (void *mb, unsigned long v)
{
	set_uint32_be (mb, 69, v);
}

unsigned short mb_get_finder_flags (const void *mb)
{
	unsigned short ret;

	ret = get_uint8_be (mb, 0x49);
	ret = (ret << 8) | get_uint8_be (mb, 0x65);

	return (ret);
}

void mb_set_finder_flags (void *mb, unsigned short v)
{
	set_uint8_be (mb, 0x65, v & 0xff);
	set_uint8_be (mb, 0x49, v >> 8);
}

unsigned short mb_get_position_y (const void *mb)
{
	return (get_uint16_be (mb, 0x4b));
}

void mb_set_position_y (void *mb, unsigned short v)
{
	set_uint16_be (mb, 0x4b, v);
}

unsigned short mb_get_position_x (const void *mb)
{
	return (get_uint16_be (mb, 0x4d));
}

void mb_set_position_x (void *mb, unsigned short v)
{
	set_uint16_be (mb, 0x4d, v);
}

unsigned short mb_get_folder_id (const void *mb)
{
	return (get_uint16_be (mb, 0x4f));
}

void mb_set_folder_id (void *mb, unsigned short v)
{
	set_uint16_be (mb, 0x4f, v);
}

unsigned short mb_get_folder_view (const void *mb)
{
	return (get_uint16_be (mb, 0x4f));
}

void mb_set_folder_view (void *mb, unsigned short v)
{
	set_uint16_be (mb, 0x4f, v);
}

unsigned char mb_get_file_flags (const void *mb)
{
	return (get_uint8_be (mb, 0x51));
}

void mb_set_file_flags (void *mb, unsigned char v)
{
	set_uint8_be (mb, 0x51, v);
}

unsigned long mb_get_data_size (const void *mb)
{
	return (get_uint32_be (mb, 83));
}

void mb_set_data_size (void *mb, unsigned long v)
{
	set_uint32_be (mb, 83, v);
}

unsigned long mb_get_rsrc_size (const void *mb)
{
	return (get_uint32_be (mb, 87));
}

void mb_set_rsrc_size (void *mb, unsigned long v)
{
	set_uint32_be (mb, 87, v);
}

unsigned long mb_get_ctime (const void *mb)
{
	return (get_uint32_be (mb, 91));
}

void mb_set_ctime (void *mb, unsigned long v)
{
	set_uint32_be (mb, 91, v);
}

unsigned long mb_get_mtime (const void *mb)
{
	return (get_uint32_be (mb, 95));
}

void mb_set_mtime (void *mb, unsigned long v)
{
	set_uint32_be (mb, 95, v);
}

unsigned mb_get_getinfo_size (const void *mb)
{
	return (get_uint16_be (mb, 0x63));
}

void mb_set_getinfo_size (void *mb, unsigned v)
{
	set_uint16_be (mb, 0x63, v);
}

unsigned short mb_get_scroll_y (const void *mb)
{
	return (get_uint16_be (mb, 0x6d));
}

void mb_set_scroll_y (void *mb, unsigned short v)
{
	set_uint16_be (mb, 0x6d, v);
}

unsigned short mb_get_scroll_x (const void *mb)
{
	return (get_uint16_be (mb, 0x6f));
}

void mb_set_scroll_x (void *mb, unsigned short v)
{
	set_uint16_be (mb, 0x6f, v);
}

unsigned mb_get_secondary_size (const void *mb)
{
	return (get_uint16_be (mb, 0x78));
}

void mb_set_secondary_size (void *mb, unsigned v)
{
	set_uint16_be (mb, 0x78, v);
}

unsigned mb_get_vers_read (const void *mb)
{
	return (get_uint8_be (mb, 0x7b));
}

void mb_set_vers_read (void *mb, unsigned v)
{
	set_uint8_be (mb, 0x7b, v);
}

unsigned mb_get_vers_write (const void *mb)
{
	return (get_uint8_be (mb, 0x7a));
}

void mb_set_vers_write (void *mb, unsigned v)
{
	set_uint8_be (mb, 0x7a, v);
}

unsigned short mb_get_crc (const void *mb)
{
	return (get_uint16_be (mb, 124));
}

void mb_set_crc (void *mb, unsigned short v)
{
	set_uint16_be (mb, 124, v);
}

void mb_get_macfile (const void *mb, macfile_t *mf)
{
	mf_set_name_pstr (mf, mb_get_name_pstr (mb));

	mf_set_data_size (mf, mb_get_data_size (mb));
	mf_set_rsrc_size (mf, mb_get_rsrc_size (mb));

	mf_set_type (mf, mb_get_type (mb));
	mf_set_creator (mf, mb_get_creator (mb));
	mf_set_finder_flags (mf, mb_get_finder_flags (mb));
	mf_set_file_flags (mf, mb_get_file_flags (mb));
	mf_set_position_x (mf, mb_get_position_x (mb));
	mf_set_position_y (mf, mb_get_position_y (mb));
	mf_set_folder_id (mf, mb_get_folder_id (mb));

	if (mb_get_id (mb) == 1) {
		mf_set_data_size (mf, 0);
		mf_set_rsrc_size (mf, 0);
		mf_set_type (mf, 0);
		mf_set_creator (mf, 0);
		mf_set_scroll_x (mf, 0);
		mf_set_scroll_y (mf, 0);
	}

	mf_set_mtime (mf, mb_get_mtime (mb));
	mf_set_ctime (mf, mb_get_ctime (mb));
}

void mb_set_macfile (void *mb, const macfile_t *mf)
{
	mb_set_name_pstr (mb, mf_get_name_pstr (mf));

	mb_set_data_size (mb, mf_get_data_size (mf));
	mb_set_rsrc_size (mb, mf_get_rsrc_size (mf));

	mb_set_type (mb, mf_get_type (mf));
	mb_set_creator (mb, mf_get_creator (mf));
	mb_set_finder_flags (mb, mf_get_finder_flags (mf));
	mb_set_file_flags (mb, mf_get_file_flags (mf) & 0xff);
	mb_set_position_x (mb, mf_get_position_x (mf));
	mb_set_position_y (mb, mf_get_position_y (mf));
	mb_set_folder_id (mb, mf_get_folder_id (mf));

	mb_set_mtime (mb, mf_get_mtime (mf));
	mb_set_ctime (mb, mf_get_ctime (mf));
}
