/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/mactime.c                                            *
 * Created:     2004-02-07 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mactime.h"


void mac_time_to_hms (unsigned long t, unsigned *h, unsigned *m, unsigned *s)
{
	t %= (24UL * 60UL * 60UL);

	*h = t / (60U * 60U);
	*m = (t / 60U) % 60U;
	*s = t % 60;
}

unsigned long mac_hms_to_time (unsigned h, unsigned m, unsigned s)
{
	if ((h > 23) || (m > 59) || (s > 59)) {
		return (0);
	}

	return (60UL * (60UL * h + m) + s);
}

void mac_time_to_ymd (unsigned long t, unsigned *y, unsigned *m, unsigned *d)
{
	unsigned i, n;
	unsigned *month;

	static unsigned months[2][12] = {
		{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
		{ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
	};

	t /= (24UL * 60UL * 60UL);

	n = t / (365U * 4U + 1U);
	t = t % (365U * 4U + 1U);

	*y = 1904U + 4U * n;

	if (t < 366U) {
		month = months[1];
	}
	else {
		*y += 1 + (t - 366U) / 365U;
		t = (t - 366U) % 365U;
		month = months[0];
	}

	n = 0;
	for (i = 0; i < 12; i++) {
		if ((n + month[i]) > t) {
			break;
		}

		n += month[i];
	}

	*m = i;
	*d = t - n;
}

unsigned long mac_ymd_to_time (unsigned y, unsigned m, unsigned d)
{
	unsigned long t;
	unsigned      i;
	unsigned      *month;

	static unsigned months[2][12] = {
		{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
		{ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }
	};

	if ((y < 1904U) || (m > 11)) {
		return (0);
	}

	y -= 1904U;

	t = (y / 4UL) * (4UL * 365UL + 1UL);
	y = y % 4U;

	t += 365UL * y + (y > 0);

	month = (y == 0) ? months[1] : months[0];

	for (i = 0; i < m; i++) {
		t += month[i];
	}

	if (d >= month[m]) {
		return (0);
	}

	t += d;

	t *= (24UL * 60UL * 60UL);

	return (t);
}

time_t mac_time_to_unix (unsigned long t)
{
	unsigned  y, m, d, hrs, min, sec;
	time_t    ret;
	struct tm tm;

	mac_time_to_ymd (t, &y, &m, &d);
	mac_time_to_hms (t, &hrs, &min, &sec);

	tm.tm_sec = sec;
	tm.tm_min = min;
	tm.tm_hour = hrs;
	tm.tm_mday = d + 1;
	tm.tm_mon = m;
	tm.tm_year = y - 1900;
	tm.tm_isdst = -1;

	ret = mktime (&tm);

	return (ret);
}

unsigned long mac_time_from_unix (time_t tm)
{
	unsigned long ret;
	struct tm     *t;

	t = localtime (&tm);

	ret = mac_ymd_to_time (1900 + t->tm_year, t->tm_mon, t->tm_mday - 1);
	ret += mac_hms_to_time (t->tm_hour, t->tm_min, t->tm_sec);

	return (ret);
}

unsigned long mac_time_from_utc (time_t tm)
{
	unsigned long ret;
	struct tm     *t;

	t = gmtime (&tm);

	ret = mac_ymd_to_time (1900 + t->tm_year, t->tm_mon, t->tm_mday - 1);
	ret += mac_hms_to_time (t->tm_hour, t->tm_min, t->tm_sec);

	return (ret);
}

static
int mac_is_digit (char c)
{
	return ((c >= '0') && (c <= '9'));
}

static
void mac_get_uint (const char **str, unsigned *val)
{
	const char *s;
	unsigned   v;

	s = *str;
	v = 0;

	while (mac_is_digit (*s) == 0) {
		if (*s == 0) {
			*str = s;
			*val = 0;
			return;
		}

		s += 1;
	}

	while ((*s >= '0') && (*s <= '9')) {
		v = 10 * v + (*s - '0');
		s += 1;
	}

	*str = s;
	*val = v;
}

unsigned long mac_time_from_string (const char *str)
{
	unsigned      i;
	unsigned      val[6];
	unsigned long ret;

	for (i = 0; i < 6; i++) {
		mac_get_uint (&str, &val[i]);
	}

	if (val[1] > 0) {
		val[1] -= 1;
	}

	if (val[2] > 0) {
		val[2] -= 1;
	}

	ret = mac_ymd_to_time (val[0], val[1], val[2]);
	ret += mac_hms_to_time (val[3], val[4], val[5]);

	return (ret);
}
