/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/mar.c                                                *
 * Created:     2011-06-25 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2015 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/macfile.h>
#include <lib/mar.h>
#include <lib/strings.h>


void mar_init (void *mar)
{
	unsigned      i;
	unsigned char *buf;

	buf = (unsigned char *) mar;

	for (i = 0; i < 128; i++) {
		buf[i] = 0;
	}

	set_uint32_be (buf, 0, MAR_MAGIC);
}

int mar_read (void *mar, FILE *fp)
{
	if (fread (mar, 1, 128, fp) != 128) {
		return (1);
	}

	return (0);
}

int mar_write (const void *mar, FILE *fp)
{
	if (fwrite (mar, 1, 128, fp) != 128) {
		return (1);
	}

	return (0);
}

unsigned long mar_calc_crc (const void *buf, unsigned long cnt, unsigned long crc)
{
	unsigned            i;
	const unsigned char *src = buf;

	while (cnt > 0) {
		crc ^= (unsigned long) *(src++) << 24;

		for (i = 0; i < 8; i++) {
			if (crc & 0x80000000) {
				crc = (crc << 1) ^ MAR_CRC_POLY;
			}
			else {
				crc = crc << 1;
			}
		}

		cnt -= 1;
	}

	return (crc & 0xffffffff);
}

int mar_is_mar (const void *mar)
{
	if (get_uint32_be (mar, 0) != MAR_MAGIC) {
		return (0);
	}

	if (get_uint32_be (mar, 124) != mar_calc_crc (mar, 124, 0)) {
		return (0);
	}

	return (1);
}

unsigned long mar_get_id (const void *mar)
{
	return (get_uint32_be (mar, 4));
}

void mar_set_id (void *mar, unsigned long v)
{
	set_uint32_be (mar, 4, v);
}

const unsigned char *mar_get_name_pstr (const void *mar)
{
	return ((const unsigned char *) mar + 8);
}

int mar_set_name_pstr (void *mar, const void *str)
{
	unsigned char       *d;
	const unsigned char *s;

	d = mar;
	s = str;

	if (s[0] > 63) {
		return (1);
	}

	memset (d + 8, 0, 64);
	memcpy (d + 8, s, s[0] + 1);

	return (0);
}

char *mar_get_name_cstr (const void *mar)
{
	const unsigned char *tmp = mar;

	return (mac_pstring_decode (tmp + 8, "/"));
}

int mar_set_name_cstr (void *mar, const char *str)
{
	unsigned char *tmp = mar;

	memset (tmp + 8, 0, 64);

	return (mac_pstring_encode_buf (tmp + 8, 64, str));
}


unsigned long mar_get_data_size (const void *mar)
{
	return (get_uint32_be (mar, 72));
}

void mar_set_data_size (void *mar, unsigned long v)
{
	set_uint32_be (mar, 72, v);
}


unsigned long mar_get_rsrc_size (const void *mar)
{
	return (get_uint32_be (mar, 76));
}

void mar_set_rsrc_size (void *mar, unsigned long v)
{
	set_uint32_be (mar, 76, v);
}


unsigned mar_get_comment_size (const void *mar)
{
	return (get_uint32_be (mar, 80));
}

void mar_set_comment_size (void *mar, unsigned v)
{
	set_uint32_be (mar, 80, v);
}


unsigned long mar_get_file_flags (const void *mar)
{
	return (get_uint32_be (mar, 84));
}

void mar_set_file_flags (void *mar, unsigned long v)
{
	set_uint32_be (mar, 84, v);
}


unsigned long mar_get_ctime (const void *mar)
{
	return (get_uint32_be (mar, 88));
}

void mar_set_ctime (void *mar, unsigned long v)
{
	set_uint32_be (mar, 88, v);
}


unsigned long mar_get_mtime (const void *mar)
{
	return (get_uint32_be (mar, 92));
}

void mar_set_mtime (void *mar, unsigned long v)
{
	set_uint32_be (mar, 92, v);
}


void mar_get_finfo (const void *mar, void *buf)
{
	memcpy (buf, (const unsigned char *) mar + 96, 16);
}

void mar_set_finfo (void *mar, const void *buf)
{
	memcpy ((unsigned char *) mar + 96, buf, 16);
}


unsigned long mar_get_type (const void *mar)
{
	return (get_uint32_be (mar, 96));
}

void mar_set_type (void *mar, unsigned long v)
{
	set_uint32_be (mar, 96, v);
}


unsigned long mar_get_creator (const void *mar)
{
	return (get_uint32_be (mar, 100));
}

void mar_set_creator (void *mar, unsigned long v)
{
	set_uint32_be (mar, 100, v);
}


unsigned short mar_get_window_y1 (const void *mar)
{
	return (get_uint16_be (mar, 96));
}

void mar_set_window_y1 (void *mar, unsigned short v)
{
	set_uint16_be (mar, 96, v);
}


unsigned short mar_get_window_x1 (const void *mar)
{
	return (get_uint16_be (mar, 98));
}

void mar_set_window_x1 (void *mar, unsigned short v)
{
	set_uint16_be (mar, 98, v);
}


unsigned short mar_get_window_y2 (const void *mar)
{
	return (get_uint16_be (mar, 100));
}

void mar_set_window_y2 (void *mar, unsigned short v)
{
	set_uint16_be (mar, 100, v);
}


unsigned short mar_get_window_x2 (const void *mar)
{
	return (get_uint16_be (mar, 102));
}

void mar_set_window_x2 (void *mar, unsigned short v)
{
	set_uint16_be (mar, 102, v);
}


unsigned short mar_get_finder_flags (const void *mar)
{
	return (get_uint16_be (mar, 104));
}

void mar_set_finder_flags (void *mar, unsigned short v)
{
	set_uint16_be (mar, 104, v);
}


unsigned short mar_get_position_y (const void *mar)
{
	return (get_uint16_be (mar, 106));
}

void mar_set_position_y (void *mar, unsigned short v)
{
	set_uint16_be (mar, 106, v);
}


unsigned short mar_get_position_x (const void *mar)
{
	return (get_uint16_be (mar, 108));
}

void mar_set_position_x (void *mar, unsigned short v)
{
	set_uint16_be (mar, 108, v);
}


unsigned short mar_get_folder_id (const void *mar)
{
	return (get_uint16_be (mar, 110));
}

void mar_set_folder_id (void *mar, unsigned short v)
{
	set_uint16_be (mar, 110, v);
}


unsigned short mar_get_window_view (const void *mar)
{
	return (get_uint16_be (mar, 110));
}

void mar_set_window_view (void *mar, unsigned short v)
{
	set_uint16_be (mar, 110, v);
}


unsigned short mar_get_scroll_y (const void *mar)
{
	return (get_uint16_be (mar, 112));
}

void mar_set_scroll_y (void *mar, unsigned short v)
{
	set_uint16_be (mar, 112, v);
}


unsigned short mar_get_scroll_x (const void *mar)
{
	return (get_uint16_be (mar, 114));
}

void mar_set_scroll_x (void *mar, unsigned short v)
{
	set_uint16_be (mar, 114, v);
}


unsigned short mar_get_crc (const void *mar)
{
	return (get_uint16_be (mar, 124));
}

void mar_set_crc (void *mar)
{
	set_uint32_be (mar, 124, mar_calc_crc (mar, 124, 0));
}


void mar_get_macfile (const void *mar, macfile_t *mf)
{
	mf_set_name_pstr (mf, mar_get_name_pstr (mar));
	mf_set_data_size (mf, mar_get_data_size (mar));
	mf_set_rsrc_size (mf, mar_get_rsrc_size (mar));
	mf_set_comment_size (mf, mar_get_comment_size (mar));
	mf_set_type (mf, mar_get_type (mar));
	mf_set_creator (mf, mar_get_creator (mar));
	mf_set_finder_flags (mf, mar_get_finder_flags (mar));
	mf_set_file_flags (mf, mar_get_file_flags (mar));
	mf_set_position_x (mf, mar_get_position_x (mar));
	mf_set_position_y (mf, mar_get_position_y (mar));
	mf_set_folder_id (mf, mar_get_folder_id (mar));
	mf_set_scroll_x (mf, mar_get_scroll_x (mar));
	mf_set_scroll_y (mf, mar_get_scroll_y (mar));
	mf_set_mtime (mf, mar_get_mtime (mar));
	mf_set_ctime (mf, mar_get_ctime (mar));
}

void mar_set_macfile (void *mar, const macfile_t *mf)
{
	if (mf_get_file_flags (mf) & MF_FILE_ISDIR) {
		mar_set_id (mar, MAR_FOLDER_START);
	}
	else {
		mar_set_id (mar, MAR_FILE);
	}

	mar_set_name_pstr (mar, mf_get_name_pstr (mf));
	mar_set_data_size (mar, mf_get_data_size (mf));
	mar_set_rsrc_size (mar, mf_get_rsrc_size (mf));
	mar_set_comment_size (mar, mf_get_comment_size (mf));
	mar_set_type (mar, mf_get_type (mf));
	mar_set_creator (mar, mf_get_creator (mf));
	mar_set_finder_flags (mar, mf_get_finder_flags (mf));
	mar_set_file_flags (mar, mf_get_file_flags (mf));
	mar_set_position_x (mar, mf_get_position_x (mf));
	mar_set_position_y (mar, mf_get_position_y (mf));
	mar_set_folder_id (mar, mf_get_folder_id (mf));
	mar_set_scroll_x (mar, mf_get_scroll_x (mf));
	mar_set_scroll_y (mar, mf_get_scroll_y (mf));
	mar_set_mtime (mar, mf_get_mtime (mf));
	mar_set_ctime (mar, mf_get_ctime (mf));
}
