/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/mfs.c                                                *
 * Created:     2012-04-21 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2012 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/mfs.h>
#include <lib/strings.h>


static
int mfs_read_bmap (mfs_vol_t *mfs)
{
	unsigned      i, n;
	unsigned      v;
	unsigned      si, bn;
	unsigned      blk;
	unsigned char buf[512];

	n = mfs_get_ablk_count (mfs);

	if (n > 4096) {
		n = 4096;
	}

	blk = 2;

	if (mfs_read_block (mfs, buf, blk++, 1)) {
		return (1);
	}

	v = 0;
	si = 64;
	bn = 0;

	for (i = 0; i < n; i++) {
		while (bn < 12) {
			if (si >= 512) {
				if (mfs_read_block (mfs, buf, blk++, 1)) {
					return (1);
				}

				si = 0;
			}

			v = (v << 8) | buf[si++];

			bn += 8;
		}

		mfs->bmap[i] = (v >> (bn - 12)) & 0xfff;

		bn -= 12;
	}

	for (i = n; i < 4096; i++) {
		mfs->bmap[i] = 0;
	}

	return (0);
}

int mfs_open (mfs_vol_t *mfs, FILE *fp, unsigned long offset, unsigned long blocks)
{
	unsigned char buf[512];

	mfs->fp = fp;

	mfs->offset = offset;
	mfs->blocks = blocks;

	mfs->readonly = 0;
	mfs->vinf_mod = 0;
	mfs->boot_mod = 0;
	mfs->volinfo_mod = 0;
	mfs->altinfo_mod = 0;

	if (mfs_read_block (mfs, mfs->boot, 0, 2)) {
		return (1);
	}

	if (mfs_read_block (mfs, buf, 2, 1)) {
		return (1);
	}

	memcpy (mfs->volinfo, buf, 64);

	if (get_uint16_be (mfs->volinfo, 0) != MFS_ID) {
		return (1);
	}

	memset (mfs->altinfo, 0, 64);

	if (mfs->blocks > 2) {
		if (mfs_read_block (mfs, buf, mfs->blocks - 2, 1) == 0) {
			memcpy (mfs->altinfo, buf, 64);
		}
	}

	mfs->vinf = mfs->volinfo;

	if (mfs_read_bmap (mfs)) {
		return (1);
	}

	return (0);
}

int mfs_open_dc42 (mfs_vol_t *mfs, FILE *fp)
{
	int           r;
	unsigned long datasize;
	unsigned char buf[128];

	if (mac_read (fp, buf, 0, 84)) {
		return (1);
	}

	if (buf[0] > 63) {
		return (1);
	}

	if (get_uint16_be (buf, 82) != 0x0100) {
		return (1);
	}

	datasize = get_uint32_be (buf, 64);

	if ((datasize < 512) || (datasize & 511)) {
		return (1);
	}

	r = mfs_open (mfs, fp, 84, datasize / 512);

	mfs->readonly = 1;

	return (r);
}

int mfs_open_auto (mfs_vol_t *mfs, FILE *fp, unsigned long offset, unsigned long blocks)
{
	if (mfs_open (mfs, fp, offset, blocks) == 0) {
		return (0);
	}

	if (mfs_open_dc42 (mfs, fp) == 0) {
		return (0);
	}

	return (1);
}

int mfs_close (mfs_vol_t *mfs, int write)
{
	unsigned char buf[512];

	if (write == 0) {
		return (0);
	}

	mfs_set_volinfo (mfs, 0);

	if (mfs->boot_mod) {
		if (mfs_write_block (mfs, mfs->boot, 0, 2)) {
			return (1);
		}
	}

	if (mfs->volinfo_mod) {
		if (mfs_read_block (mfs, buf, 2, 1)) {
			return (1);
		}

		memcpy (buf, mfs->volinfo, 64);

		if (mfs_write_block (mfs, buf, 2, 1)) {
			return (1);
		}
	}

	if (mfs->altinfo_mod && (mfs->blocks > 2)) {
		if (mfs_read_block (mfs, buf, mfs->blocks - 2, 1)) {
			return (1);
		}

		memcpy (buf, mfs->altinfo, 64);

		if (mfs_write_block (mfs, buf, mfs->blocks - 2, 1)) {
			return (1);
		}
	}

	return (0);
}

int mfs_read_block (mfs_vol_t *mfs, void *buf, unsigned long idx, unsigned long cnt)
{
	if ((mfs->blocks > 0) && ((idx + cnt) > mfs->blocks)) {
		return (1);
	}

	if (mac_read (mfs->fp, buf, mfs->offset + 512 * idx, 512 * cnt)) {
		return (1);
	}

	return (0);
}

int mfs_write_block (mfs_vol_t *mfs, const void *buf, unsigned long idx, unsigned long cnt)
{
	if (mfs->readonly) {
		return (1);
	}

	if ((mfs->blocks > 0) && ((idx + cnt) > mfs->blocks)) {
		return (1);
	}

	if (mac_write (mfs->fp, buf, mfs->offset + 512 * idx, 512 * cnt)) {
		return (1);
	}

	return (0);
}


void mfs_set_volinfo (mfs_vol_t *mfs, int alt)
{
	if (mfs->vinf == mfs->volinfo) {
		mfs->volinfo_mod |= mfs->vinf_mod;
	}
	else {
		mfs->altinfo_mod |= mfs->vinf_mod;
	}

	mfs->vinf_mod = 0;
	mfs->vinf = alt ? mfs->altinfo : mfs->volinfo;
}

unsigned mfs_get_id (const mfs_vol_t *mfs)
{
	return (get_uint16_be (mfs->vinf, 0));
}

void mfs_set_id (mfs_vol_t *mfs, unsigned val)
{
	mfs->vinf_mod = 1;
	set_uint16_be (mfs->vinf, 0, val);
}


char *mfs_get_vol_name (const mfs_vol_t *mfs)
{
	const unsigned char *str;

	str = mfs->vinf + 36;

	if (str[0] > 27) {
		return (NULL);
	}

	return (mac_pstring_decode (str, ""));
}

int mfs_set_vol_name (mfs_vol_t *mfs, const char *val)
{
	unsigned char *str;

	mfs->vinf_mod = 1;
	str = mfs->vinf + 36;

	memset (str, 0, 28);

	return (mac_pstring_encode_buf (str, 28, val));
}

unsigned long mfs_get_ctime (const mfs_vol_t *mfs)
{
	return (get_uint32_be (mfs->vinf, 2));
}

void mfs_set_ctime (mfs_vol_t *mfs, unsigned long val)
{
	mfs->vinf_mod = 1;
	set_uint32_be (mfs->vinf, 2, val);
}


unsigned long mfs_get_mtime (const mfs_vol_t *mfs)
{
	return (get_uint32_be (mfs->vinf, 6));
}

void mfs_set_mtime (mfs_vol_t *mfs, unsigned long val)
{
	mfs->vinf_mod = 1;
	set_uint32_be (mfs->vinf, 6, val);
}


unsigned mfs_get_volume_attributes (const mfs_vol_t *mfs)
{
	return (get_uint16_be (mfs->vinf, 10));
}

void mfs_set_volume_attributes (mfs_vol_t *mfs, unsigned val)
{
	mfs->vinf_mod = 1;
	set_uint16_be (mfs->vinf, 10, val);
}


unsigned mfs_get_file_count (const mfs_vol_t *mfs)
{
	return (get_uint16_be (mfs->vinf, 12));
}

void mfs_set_file_count (mfs_vol_t *mfs, unsigned val)
{
	mfs->vinf_mod = 1;
	set_uint16_be (mfs->vinf, 12, val);
}


unsigned mfs_get_directory_start (const mfs_vol_t *mfs)
{
	return (get_uint16_be (mfs->vinf, 14));
}

void mfs_set_directory_start (mfs_vol_t *mfs, unsigned val)
{
	mfs->vinf_mod = 1;
	set_uint16_be (mfs->vinf, 14, val);
}


unsigned mfs_get_directory_blocks (const mfs_vol_t *mfs)
{
	return (get_uint16_be (mfs->vinf, 16));
}

void mfs_set_directory_blocks (mfs_vol_t *mfs, unsigned val)
{
	mfs->vinf_mod = 1;
	set_uint16_be (mfs->vinf, 16, val);
}


unsigned mfs_get_ablk_count (const mfs_vol_t *mfs)
{
	return (get_uint16_be (mfs->vinf, 18));
}

void mfs_set_ablk_count (mfs_vol_t *mfs, unsigned val)
{
	mfs->vinf_mod = 1;
	set_uint16_be (mfs->vinf, 18, val);
}


unsigned long mfs_get_ablk_size (const mfs_vol_t *mfs)
{
	return (get_uint32_be (mfs->vinf, 20));
}

void mfs_set_ablk_size (mfs_vol_t *mfs, unsigned long val)
{
	mfs->vinf_mod = 1;
	set_uint32_be (mfs->vinf, 20, val);
}


unsigned long mfs_get_clump_size (const mfs_vol_t *mfs)
{
	return (get_uint32_be (mfs->vinf, 24));
}

void mfs_set_clump_size (mfs_vol_t *mfs, unsigned long val)
{
	mfs->vinf_mod = 1;
	set_uint32_be (mfs->vinf, 24, val);
}


unsigned mfs_get_ablk_start (const mfs_vol_t *mfs)
{
	return (get_uint16_be (mfs->vinf, 28));
}

void mfs_set_ablk_start (mfs_vol_t *mfs, unsigned val)
{
	mfs->vinf_mod = 1;
	set_uint16_be (mfs->vinf, 28, val);
}


unsigned long mfs_get_next_file (const mfs_vol_t *mfs)
{
	return (get_uint32_be (mfs->vinf, 30));
}

void mfs_set_next_file (mfs_vol_t *mfs, unsigned long val)
{
	mfs->vinf_mod = 1;
	set_uint32_be (mfs->vinf, 30, val);
}


unsigned mfs_get_free_ablk (const mfs_vol_t *mfs)
{
	return (get_uint16_be (mfs->vinf, 34));
}

void mfs_set_free_ablk (mfs_vol_t *mfs, unsigned val)
{
	mfs->vinf_mod = 1;
	set_uint16_be (mfs->vinf, 34, val);
}

void mfs_update_alt_volinfo (mfs_vol_t *mfs)
{
	mfs->altinfo_mod = 1;
	memcpy (mfs->altinfo, mfs->volinfo, 64);
}

void mfs_restore_alt_volinfo (mfs_vol_t *mfs)
{
	if (mfs->blocks > 2) {
		mfs->volinfo_mod = 1;
		memcpy (mfs->volinfo, mfs->altinfo, 64);
	}
}
