/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macbin/macbin.c                                          *
 * Created:     2001-11-20 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2001-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>

#include "config.h"
#include "macbin.h"

#include <lib/crc.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/mactime.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


static const char    *arg0 = NULL;

static unsigned      par_op = OP_NONE;

static int           par_macbin_id_ok = 0;
static unsigned      par_macbin_id = 0;

static int           par_type_ok = 0;
static unsigned long par_type = MAC_FOURCC_UNKN;

static int           par_creator_ok = 0;
static unsigned long par_creator = MAC_FOURCC_UNKN;

static int           par_macname_ok = 0;
static char          *par_macname = NULL;

static int           par_finder_flags_ok = 0;
static unsigned      par_finder_flags = 0;

static int           par_file_flags_ok = 0;
static unsigned      par_file_flags = 0;

static const char    *par_ctime = NULL;
static const char    *par_mtime = NULL;

static int           par_position_x_ok = 0;
static unsigned      par_position_x = 0;

static int           par_position_y_ok = 0;
static unsigned      par_position_y = 0;

static char          *par_fn_bin = NULL;
static FILE          *par_fp_bin = NULL;

static char          *par_fn_dat = NULL;
static FILE          *par_fp_dat = NULL;

static char          *par_fn_rsc = NULL;
static FILE          *par_fp_rsc = NULL;

static char          *par_fn_inf = NULL;
static FILE          *par_fp_inf = NULL;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'b', 1, "binary", "string", "Set the binary file name" },
	{ 'c', 1, "creator", "fourcc", "Set the creator [????]" },
	{ 'C', 1, "ctime", "string", "Set the creation time [0]" },
	{ 'd', 0, "decode", NULL, "Decode a macbinary file" },
	{ 'e', 0, "encode", NULL, "Encode a macbinary file" },
	{ 'f', 1, "data", "string", "Set the data fork file name [none]" },
	{ 'I', 1, "macbinary-id", "int", "Set the macbinary ID [0]" },
	{ 'l', 1, "finder-flags", "int", "Set the finder flags [0]" },
	{ 'L', 1, "file-flags", "int", "Set the file flags [0]" },
	{ 'm', 1, "mac-name", "string", "Set the mac file name [same as data file]" },
	{ 'M', 1, "mtime", "string", "Set the modification time [0]" },
	{ 'n', 1, "finf", "string", "Set the finf file name [none]" },
	{ 'r', 1, "rsrc", "string", "Set the resource fork file name [none]" },
	{ 't', 1, "type", "fourcc", "Set the type [????]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'x', 1, "position-x", "int", "Set the x position [0]" },
	{ 'y', 1, "position-y", "int", "Set the y position [0]" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"macbin: encode and decode MacBinary II files",
		"usage: macbin [options] [binary [data [resource [finf]]]]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"macbin version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2001-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int macbin_decode_mb (unsigned char *mb)
{
	unsigned long i, n;
	unsigned long datn, rscn;
	unsigned char buf[4096];

	if (par_fp_bin == NULL) {
		return (1);
	}

	if (mb_get_version (mb) == 0) {
		fprintf (stderr, "not a macbinary file\n");
		return (1);
	}

	datn = mb_get_data_size (mb);
	rscn = mb_get_rsrc_size (mb);

	n = datn;
	while (n > 0) {
		i = (n < 4096) ? n : 4096;

		if (fread (buf, 1, i, par_fp_bin) != i) {
			return (1);
		}

		if (par_fp_dat != NULL) {
			if (fwrite (buf, 1, i, par_fp_dat) != i) {
				return (1);
			}
		}

		n -= i;
	}

	n = (128 - (datn & 127)) & 127;
	if (n > 0) {
		if (fread (buf, 1, n, par_fp_bin) != n) {
			return (1);
		}
	}

	n = rscn;
	while (n > 0) {
		i = (n < 4096) ? n : 4096;

		if (fread (buf, 1, i, par_fp_bin) != i) {
			return (1);
		}

		if (par_fp_rsc != NULL) {
			if (fwrite (buf, 1, i, par_fp_rsc) != i) {
				return (1);
			}
		}

		n -= i;
	}

	if (par_fp_inf != NULL) {
		macfile_t mf;

		mf_init (&mf);
		mb_get_macfile (mb, &mf);
		mf_save (&mf, par_fp_inf);
	}

	return (0);
}

static
int macbin_decode (void)
{
	int           r;
	unsigned long tm;
	unsigned char mb[128];

	par_fp_bin = mac_open_inp (par_fn_bin, "rb", 0);
	if (par_fp_bin == NULL) {
		fprintf (stderr, "error opening macbinary file\n");
		return (1);
	}

	if (mb_read (mb, par_fp_bin)) {
		return (1);
	}

	par_fp_dat = mac_open_out (par_fn_dat, "wb", 0);
	if ((par_fn_dat != NULL) && (par_fp_dat == NULL)) {
		fprintf (stderr, "error opening data fork file\n");
		return (1);
	}

	par_fp_rsc = mac_open_out (par_fn_rsc, "wb", 0);
	if ((par_fn_rsc != NULL) && (par_fp_rsc == NULL)) {
		fprintf (stderr, "error opening resource fork file\n");
		return (1);
	}

	par_fp_inf = mac_open_out (par_fn_inf, "w", 0);
	if ((par_fn_inf != NULL) && (par_fp_inf == NULL)) {
		fprintf (stderr, "error opening finf file\n");
		return (1);
	}

	r = macbin_decode_mb (mb);

	if (r) {
		return (1);
	}

	tm = mb_get_mtime (mb);

	mac_close (par_fp_dat);
	mac_close (par_fp_rsc);

	par_fp_dat = NULL;
	par_fp_rsc = NULL;

	uf_set_mtime (par_fn_dat, tm);
	uf_set_mtime (par_fn_rsc, tm);

	return (0);
}


static
int macbin_encode_mb (unsigned char *mb)
{
	unsigned long i, n;
	unsigned long datn, rscn;
	unsigned char buf[4096];

	if (par_fp_bin == NULL) {
		return (1);
	}

	datn = mb_get_data_size (mb);
	rscn = mb_get_rsrc_size (mb);

	if ((datn > 0) && (par_fp_dat == NULL)) {
		return (1);
	}

	if ((rscn > 0) && (par_fp_rsc == NULL)) {
		return (1);
	}

	if (mb_write (mb, par_fp_bin)) {
		fprintf (stderr, "error writing macbinary header\n");
		return (1);
	}

	n = datn;
	while (n > 0) {
		i = (n < 4096) ? n : 4096;

		if (fread (buf, 1, i, par_fp_dat) != i) {
			return (1);
		}

		if (fwrite (buf, 1, i, par_fp_bin) != i) {
			return (1);
		}

		n -= i;
	}

	n = (128 - (datn & 127)) & 127;
	if (n > 0) {
		for (i = 0; i < n; i++) {
			buf[i] = 0;
		}

		if (fwrite (buf, 1, n, par_fp_bin) != n) {
			return (1);
		}
	}

	n = rscn;
	while (n > 0) {
		i = (n < 4096) ? n : 4096;

		if (fread (buf, 1, i, par_fp_rsc) != i) {
			return (1);
		}

		if (fwrite (buf, 1, i, par_fp_bin) != i) {
			return (1);
		}

		n -= i;
	}

	n = (128 - (rscn & 127)) & 127;
	if (n > 0) {
		for (i = 0; i < n; i++) {
			buf[i] = 0;
		}

		if (fwrite (buf, 1, n, par_fp_bin) != n) {
			return (1);
		}
	}

	return (0);
}

static
int macbin_encode (void)
{
	int           r;
	unsigned long datn, rscn;
	unsigned long rsct, datt;
	unsigned char mb[128];

	datn = 0;
	rscn = 0;

	datt = 0;
	rsct = 0;

	par_fp_bin = mac_open_out (par_fn_bin, "wb", 0);
	if (par_fp_bin == NULL) {
		fprintf (stderr, "error opening macbinary file\n");
		return (1);
	}

	par_fp_dat = mac_open_inp (par_fn_dat, "rb", 0);
	if ((par_fn_dat != NULL) && (par_fp_dat == NULL)) {
		fprintf (stderr, "error opening data fork file\n");
		return (1);
	}
	if (uf_stat (par_fn_dat, &datn, &datt)) {
		datn = 0;
		datt = 0;
	}

	par_fp_rsc = mac_open_inp (par_fn_rsc, "rb", 0);
	if ((par_fn_rsc != NULL) && (par_fp_rsc == NULL)) {
		fprintf (stderr, "error opening resource fork file\n");
		return (1);
	}
	if (uf_stat (par_fn_rsc, &rscn, &rsct)) {
		rscn = 0;
		rsct = 0;
	}

	par_fp_inf = mac_open_inp (par_fn_inf, "r", 0);
	if ((par_fn_inf != NULL) && (par_fp_inf == NULL)) {
		fprintf (stderr, "error opening finf file\n");
		return (1);
	}

	if (datt == 0) {
		datt = rsct;
	}

	mb_init (mb);

	if (par_fp_inf == NULL) {
		const char    *cstr;
		unsigned char *pstr;

		if (par_fn_dat == NULL) {
			cstr = "Noname";
		}
		else {
			cstr = mac_get_last_component (par_fn_dat, '/');
		}

		pstr = mac_pstring_encode (cstr);

		if (pstr != NULL) {
			mb_set_name_pstr (mb, pstr);
			free (pstr);
		}

		mb_set_type (mb, par_type);
		mb_set_creator (mb, par_creator);
		mb_set_finder_flags (mb, par_finder_flags);
		mb_set_file_flags (mb, par_file_flags);
		mb_set_ctime (mb, datt);
		mb_set_mtime (mb, datt);
		mb_set_position_x (mb, par_position_x);
		mb_set_position_y (mb, par_position_y);
	}
	else {
		macfile_t mf;

		if (mf_load (&mf, par_fp_inf)) {
			fprintf (stderr, "error loading finf file\n");
			return (1);
		}

		mb_set_macfile (mb, &mf);
	}

	if (par_macbin_id_ok) {
		mb_set_id (mb, par_macbin_id);
	}

	if (par_macname_ok) {
		if (mb_set_name_cstr (mb, par_macname)) {
			fprintf (stderr, "bad mac file name\n");
			return (1);
		}
	}

	if (par_type_ok) {
		mb_set_type (mb, par_type);
	}

	if (par_creator_ok) {
		mb_set_creator (mb, par_creator);
	}

	if (par_finder_flags_ok) {
		mb_set_finder_flags (mb, par_finder_flags);
	}

	if (par_file_flags_ok) {
		mb_set_file_flags (mb, par_file_flags);
	}

	if (par_ctime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mb_set_ctime (mb, datt);
		}
		else {
			mb_set_ctime (mb, mac_time_from_string (par_ctime));
		}
	}

	if (par_mtime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mb_set_mtime (mb, datt);
		}
		else {
			mb_set_mtime (mb, mac_time_from_string (par_mtime));
		}
	}

	if (par_position_x_ok) {
		mb_set_position_x (mb, par_position_x);
	}

	if (par_position_y_ok) {
		mb_set_position_y (mb, par_position_y);
	}

	mb_set_data_size (mb, datn);
	mb_set_rsrc_size (mb, rscn);

	mb_set_getinfo_size (mb, 0);
	mb_set_secondary_size (mb, 0);

	mb_set_vers_read (mb, 0x81);
	mb_set_vers_write (mb, 0x81);

	mb_set_crc (mb, crc16_crc (mb, 124, CRC_CRC16_CCITT, 0));

	r = macbin_encode_mb (mb);

	return (r);
}

int main (int argc, char **argv)
{
	int  r;
	char **optarg;

	arg0 = argv[0];

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'b':
			par_fn_bin = optarg[0];
			break;

		case 'c':
			if (str_get_fourcc (optarg[0], &par_creator)) {
				fprintf (stderr,
					"%s: bad creator fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_creator_ok = 1;
			break;

		case 'C':
			par_ctime = optarg[0];
			break;

		case 'd':
			par_op = OP_DECODE;
			break;

		case 'e':
			par_op = OP_ENCODE;
			break;

		case 'f':
			par_fn_dat = optarg[0];
			break;

		case 'I':
			par_macbin_id = (unsigned) strtoul (optarg[0], NULL, 0);
			par_macbin_id_ok = 1;
			break;

		case 'l':
			par_finder_flags |= mf_finder_flag_from_string (optarg[0]);
			par_finder_flags_ok = 1;
			break;

		case 'L':
			par_file_flags |= mf_file_flag_from_string (optarg[0]);
			par_file_flags_ok = 1;
			break;

		case 'm':
			par_macname = optarg[0];
			par_macname_ok = 1;
			break;

		case 'M':
			par_mtime = optarg[0];
			break;

		case 'n':
			par_fn_inf = optarg[0];
			break;

		case 'r':
			par_fn_rsc = optarg[0];
			break;

		case 't':
			if (str_get_fourcc (optarg[0], &par_type)) {
				fprintf (stderr, "%s: bad type fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_type_ok = 1;
			break;

		case 'x':
			par_position_x = (unsigned) strtoul (optarg[0], NULL, 0);
			par_position_x_ok = 1;
			break;

		case 'y':
			par_position_y = (unsigned) strtoul (optarg[0], NULL, 0);
			par_position_y_ok = 1;
			break;

		case 0:
			if (par_fn_bin == NULL) {
				par_fn_bin = optarg[0];
			}
			else if (par_fn_dat == NULL) {
				par_fn_dat = optarg[0];
			}
			else if (par_fn_rsc == NULL) {
				par_fn_rsc = optarg[0];
			}
			else if (par_fn_inf == NULL) {
				par_fn_inf = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_op == OP_NONE) {
		par_op = OP_DECODE;
	}

	if (par_op == OP_DECODE) {
		r = macbin_decode();
	}
	else if (par_op == OP_ENCODE) {
		r = macbin_encode();
	}
	else {
		r = 1;
	}

	mac_close (par_fp_inf);
	mac_close (par_fp_rsc);
	mac_close (par_fp_dat);
	mac_close (par_fp_bin);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (0);
}
