/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macbin/mbfix.c                                           *
 * Created:     2004-02-27 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>

#include "mbfix.h"

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/macfile.h>
#include <lib/mactime.h>
#include <lib/unixfile.h>


static int           par_fix_name = 0;

static int           par_force = 0;

static int           par_type_ok = 0;
static unsigned long par_type = 0;

static int           par_creator_ok = 0;
static unsigned long par_creator = 0;

static int           par_macname_ok = 0;
static char          *par_macname = NULL;

static const char    *par_ctime = NULL;
static const char    *par_mtime = NULL;

static int           par_finder_flags_ok = 0;
static unsigned      par_finder_flags_msk = 0;
static unsigned      par_finder_flags_val = 0;

static int           par_file_flags_ok = 0;
static unsigned long par_file_flags_msk = 0;
static unsigned long par_file_flags_val = 0;

static int           par_position_x_ok = 0;
static unsigned      par_position_x = 0;

static int           par_position_y_ok = 0;
static unsigned      par_position_y = 0;

static int           par_macbinary_id_ok = 0;
static unsigned      par_macbinary_id = 0;

static int           par_zero_mb2 = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'c', 1, "creator", "fourcc", "Set the creator" },
	{ 'C', 1, "ctime", "string", "Set creation time" },
	{ 'f', 0, "fix-name", NULL, "Fix the mac file name" },
	{ 'F', 0, "force", NULL, "Don't check header integrity [no]" },
	{ 'i', 1, "input", "string", "Set a file name" },
	{ 'I', 1, "macbinary-id", "int", "Set the macbinary id" },
	{ 'l', 1, "finder-flags", "int", "Set the finder flags" },
	{ 'L', 1, "file-flags", "int", "Set the file flags" },
	{ 'm', 1, "mac-name", "string", "Set the mac file name" },
	{ 'M', 1, "mtime", "string", "Set modification time" },
	{ 't', 1, "type", "fourcc", "Set the type" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'x', 1, "position-x", "int", "Set x position" },
	{ 'y', 1, "position-y", "int", "Set y position" },
	{ 'z', 0, "zero-mb2", NULL, "Clear unused header fields [no]" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mbfix: set MacBinary II header fields",
		"usage: mbfix [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mbfix version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2004-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
char *str_base (const char *fname, const char *ext)
{
	char     *ret;
	unsigned n, m;

	n = strlen (fname);
	m = strlen (ext);

	ret = malloc (n + 1);
	if (ret == NULL) {
		return (NULL);
	}

	strcpy (ret, fname);

	if (n <= m) {
		return (ret);
	}

	if (strcmp (ret + n - m, ext) == 0) {
		ret[n - m] = 0;
	}

	return (ret);
}

static
void mb_zero_unused (unsigned char *mb)
{
	unsigned i;

	if (mb[0] != 1) {
		mb[0] = 0;
	}

	i = 2 + mb[1];
	while (i < 64) {
		mb[i++] = 0;
	}

	set_uint8_be (mb, 0x4a, 0);
	set_uint16_be (mb, 0x4f, 0);
	set_uint8_be (mb, 0x52, 0);

	if (mb[0] == 1) {
		mb_set_data_size (mb, 0);
		mb_set_rsrc_size (mb, 0);

		mb_set_vers_read (mb, 0x82);
		mb_set_vers_write (mb, 0x82);
	}
	else {
		mb_set_vers_read (mb, 0x81);
		mb_set_vers_write (mb, 0x81);
	}

	for (i = 0x66; i < 0x78; i++) {
		mb[i] = 0;
	}

	set_uint16_be (mb, 0x7e, 0);
}

static
int mbfix_mb (unsigned char *mb, FILE *fp, const char *fname)
{
	unsigned long ftime;
	unsigned long v;

	if (mb_get_version (mb) == 0) {
		if (par_force == 0) {
			fprintf (stderr, "%s: not a macbinary file\n", fname);
			return (1);
		}

		mb[0] = 0;
		mb[74] = 0;

		if (mb[1] < 1) {
			mb[1] = 1;
		}
		else if (mb[1] > 63) {
			mb[1] = 63;
		}
	}

	if (par_fix_name) {
		char *name;

		name = str_base (fname, ".bin");

		if (name != NULL) {
			if (mb_set_name_cstr (mb, name)) {
				fprintf (stderr, "%s: bad file name\n", fname);
				free (name);
				return (1);
			}

			free (name);
		}
	}
	else if (par_macname_ok) {
		if (mb_set_name_cstr (mb, par_macname)) {
			fprintf (stderr, "%s: bad file name\n", fname);
			return (1);
		}
	}

	if (par_type_ok) {
		mb_set_type (mb, par_type);
	}

	if (par_creator_ok) {
		mb_set_creator (mb, par_creator);
	}

	ftime = uf_get_mtime (fname);

	if (par_ctime != NULL) {
		if (strcmp (par_ctime, "file") == 0) {
			mb_set_ctime (mb, ftime);
		}
		else {
			mb_set_ctime (mb, mac_time_from_string (par_ctime));
		}
	}

	if (par_mtime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mb_set_mtime (mb, ftime);
		}
		else {
			mb_set_mtime (mb, mac_time_from_string (par_mtime));
		}
	}

	if (par_finder_flags_ok) {
		v = mb_get_finder_flags (mb);

		v &= ~par_finder_flags_msk;
		v |= par_finder_flags_val & par_finder_flags_msk;

		mb_set_finder_flags (mb, v);
	}

	if (par_file_flags_ok) {
		v = mb_get_file_flags (mb);

		v &= ~par_file_flags_msk;
		v |= par_file_flags_val & par_file_flags_msk;

		mb_set_file_flags (mb, v);
	}

	if (par_position_x_ok) {
		mb_set_position_x (mb, par_position_x);
	}

	if (par_position_y_ok) {
		mb_set_position_y (mb, par_position_y);
	}

	if (par_macbinary_id_ok) {
		mb_set_id (mb, par_macbinary_id);
	}

	if (mb_get_vers_read (mb) < 0x81) {
		mb_set_vers_read (mb, 0x81);
	}

	if (mb_get_vers_write (mb) < 0x81) {
		mb_set_vers_write (mb, 0x81);
	}

	if (par_zero_mb2) {
		mb_zero_unused (mb);
	}

	mb_set_crc (mb, crc16_crc (mb, 124, CRC_CRC16_CCITT, 0));

	return (0);
}

static
int mbfix_fp (unsigned char *mb, FILE *fp, const char *fname)
{
	if (mb_read (mb, fp)) {
		if (par_force) {
			mb_init (mb);
			mb_set_name_cstr (mb, "Noname");
		}
		else {
			fprintf (stderr,
				"%s: error reading macbinary header\n", fname
			);
			return (1);
		}
	}

	if (mbfix_mb (mb, fp, fname)) {
		return (1);
	}

	if (fseek (fp, 0, SEEK_SET)) {
		fprintf (stderr, "%s: seek error\n", fname);
		return (1);
	}

	if (mb_write (mb, fp)) {
		fprintf (stderr, "%s: write error\n", fname);
		return (1);
	}

	return (0);
}

static
int mbfix (const char *fname)
{
	int           r;
	FILE          *fp;
	unsigned char mb[128];
	unsigned long mtime;

	fp = fopen (fname, "r+b");

	if ((fp == NULL) && par_force) {
		fp = fopen (fname, "w+b");
	}

	if (fp == NULL) {
		fprintf (stderr, "%s: open failed\n", fname);
		return (1);
	}

	r = mbfix_fp (mb, fp, fname);

	fclose (fp);

	if (r) {
		return (1);
	}

	mtime = mb_get_mtime (mb);

	if (uf_set_mtime (fname, mtime)) {
		fprintf (stderr, "%s: can't set file time\n", fname);
		return (1);
	}

	return (0);
}

static
int set_finder_flags (const char *str)
{
	unsigned v;

	if (str[0] == '+') {
		v = mf_finder_flag_from_string (str + 1);
		par_finder_flags_msk |= v;
		par_finder_flags_val |= v;
	}
	else if (str[0] == '-') {
		v = mf_finder_flag_from_string (str + 1);
		par_finder_flags_msk |= v;
		par_finder_flags_val &= ~v;
	}
	else {
		v = mf_finder_flag_from_string (str);
		par_finder_flags_msk = 0xffff;
		par_finder_flags_val = v;
	}

	par_finder_flags_ok = 1;

	return (0);
}

static
int set_file_flags (const char *str)
{
	unsigned long v;

	if (str[0] == '+') {
		v = mf_file_flag_from_string (str + 1);
		par_file_flags_msk |= v;
		par_file_flags_val |= v;
	}
	else if (str[0] == '-') {
		v = mf_file_flag_from_string (str + 1);
		par_file_flags_msk |= v;
		par_file_flags_val &= ~v;
	}
	else {
		v = mf_file_flag_from_string (str);
		par_file_flags_msk = 0xffff;
		par_file_flags_val = v;
	}

	par_file_flags_ok = 1;

	return (0);
}

int main (int argc, char *argv[])
{
	int  r;
	int  ret;
	char **optarg;

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'c':
			if (str_get_fourcc (optarg[0], &par_creator)) {
				fprintf (stderr,
					"%s: bad creator fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_creator_ok = 1;
			break;

		case 'C':
			par_ctime = optarg[0];
			break;

		case 'f':
			par_fix_name = 1;
			break;

		case 'F':
			par_force = 1;
			break;

		case 'i':
			ret |= mbfix (optarg[0]);
			break;

		case 'I':
			par_macbinary_id = (unsigned) strtoul (optarg[0], NULL, 0);
			par_macbinary_id_ok = 1;
			break;

		case 'l':
			set_finder_flags (optarg[0]);
			break;

		case 'L':
			set_file_flags (optarg[0]);
			break;

		case 'm':
			par_macname = optarg[0];
			par_macname_ok = 1;
			break;

		case 'M':
			par_mtime = optarg[0];
			break;

		case 't':
			if (str_get_fourcc (optarg[0], &par_type)) {
				fprintf (stderr, "%s: bad type fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_type_ok = 1;
			break;

		case 'x':
			par_position_x = (unsigned) strtoul (optarg[0], NULL, 0);
			par_position_x_ok = 1;
			break;

		case 'y':
			par_position_y = (unsigned) strtoul (optarg[0], NULL, 0);
			par_position_y_ok = 1;
			break;

		case 'z':
			par_zero_mb2 = 1;
			break;

		case 0:
			ret |= mbfix (optarg[0]);
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (ret) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
