/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macbin/mbinfo.c                                          *
 * Created:     2004-03-03 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2001-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "macbin.h"

#include <lib/crc.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/macfile.h>
#include <lib/mactime.h>


static int par_inf_all = 1;
static int par_inf_mf = 0;

static int par_short = 0;
static int par_inf_fname = 0;
static int par_inf_macbin_version = 0;
static int par_inf_macbin_id = 0;
static int par_inf_crc = 0;
static int par_inf_type = 0;
static int par_inf_creator = 0;
static int par_inf_finder = 0;
static int par_inf_file_flags = 0;
static int par_inf_position_x = 0;
static int par_inf_position_y = 0;
static int par_inf_folder_id = 0;
static int par_inf_dsize = 0;
static int par_inf_rsize = 0;
static int par_inf_ctime = 0;
static int par_inf_mtime = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "all", NULL, "Print all information [default]" },
	{ 'c', 0, "creator", NULL, "Print the creator" },
	{ 'C', 0, "ctime", NULL, "Print the creation time stamp" },
	{ 'f', 0, "file-name", NULL, "Print the file name" },
	{ 'i', 1, "input", "string", "Set an input file name" },
	{ 'I', 0, "info", NULL, "Dump the mac file info [no]" },
	{ 'l', 0, "finder-flags", NULL, "Print the finder flags" },
	{ 'L', 0, "file-flags", NULL, "Print the file flags" },
	{ 'M', 0, "mtime", NULL, "Print the modification time stamp" },
	{ 's', 0, "short", NULL, "Short output format [no]" },
	{ 't', 0, "type", NULL, "Print the type" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'x', 0, "position-x", NULL, "Print the x position" },
	{ 'y', 0, "position-y", NULL, "Print the y position" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mbinfo: print MacBinary header information",
		"usage: mbinfo [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mbinfo version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2001-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
void mb_print_finder_flags (FILE *fp, unsigned ff)
{
	int        first;
	unsigned   mask;
	const char *flag;

	first = 1;

	mask = 0x8000;

	while (ff != 0) {
		if (ff & mask) {
			flag = mf_finder_flag_to_string (mask);
			ff &= ~mask;

			if (first == 0) {
				fputs (", ", fp);
			}
			else {
				first = 0;
			}

			fputs (flag, fp);
		}

		mask = mask >> 1;
	}
}

static
void mb_print_file_flags (FILE *fp, unsigned long ff)
{
	int           first;
	unsigned long mask;
	const char    *flag;

	first = 1;

	mask = 0x80000000;

	while (ff != 0) {
		if (ff & mask) {
			flag = mf_file_flag_to_string (mask);
			ff &= ~mask;

			if (first == 0) {
				fputs (", ", fp);
			}
			else {
				first = 0;
			}

			fputs (flag, fp);
		}

		mask = mask >> 1;
	}
}

static
int mb_info_fp (FILE *fp, FILE *out)
{
	unsigned char   mb[128];
	char            buf[256];
	unsigned        vers;
	unsigned        y, m, d, hrs, min, sec;
	unsigned long   v;
	static unsigned cnt = 0;

	if (cnt > 0) {
		fputs ("\n", out);
	}

	cnt += 1;

	if (mb_read (mb, fp)) {
		fprintf (stderr, "error reading macbinary header\n");
		return (1);
	}

	vers = mb_get_version (mb);
	if (vers == 0) {
		fprintf (stderr, "not a macbinary file\n");
		return (1);
	}

	if (par_inf_mf) {
		macfile_t mf;

		mf_init (&mf);
		mb_get_macfile (mb, &mf);
		mf_save (&mf, out);
	}

	if (par_inf_fname || par_inf_all) {
		char *name;

		name = mb_get_name_cstr (mb);

		if (par_short) {
			fprintf (out, "%s\n", name);
		}
		else {
			fprintf (out, "file name:          '%s'\n", name);
		}

		free (name);
	}

	if (par_inf_macbin_version || par_inf_all) {
		if (par_short) {
			fprintf (out, "%u\n", vers);
		}
		else {
			fprintf (out, "macbinary version:  %u\n", vers);
		}
	}

	if (par_inf_macbin_id || par_inf_all) {
		v = mb_get_id (mb);
		if (par_short) {
			fprintf (out, "%lu\n", v);
		}
		else {
			fprintf (out, "macbinary id:       %lu\n", v);
		}
	}

	if (par_inf_crc || par_inf_all) {
		v = mb_get_crc (mb);
		if (par_short) {
			fprintf (out, "%04lX\n", v);
		}
		else {
			fprintf (out, "crc:                %04lX\n", v);
		}
	}

	if (par_inf_type || par_inf_all) {
		v = mb_get_type (mb);
		str_set_fourcc (buf, 256, v);
		if (par_short) {
			fprintf (out, "%s\n", buf);
		}
		else {
			fprintf (out,
				"type:               \"%s\" (0x%08lX)\n",
				buf, v
			);
		}
	}

	if (par_inf_creator || par_inf_all) {
		v = mb_get_creator (mb);
		str_set_fourcc (buf, 256, v);
		if (par_short) {
			fprintf (out, "%s\n", buf);
		}
		else {
			fprintf (out,
				"creator:            \"%s\" (0x%08lX)\n",
				buf, v
			);
		}
	}

	if (par_inf_finder || par_inf_all) {
		v = mb_get_finder_flags (mb);
		if (par_short) {
			fprintf (out, "%04lX\n", v);
		}
		else {
			fprintf (out, "finder flags:       0x%04lX", v);

			if (v != 0) {
				fputs (" (", out);
				mb_print_finder_flags (out, v);
				fputs (")", out);
			}

			fputs ("\n", out);
		}
	}

	if (par_inf_file_flags || par_inf_all) {
		v = mb_get_file_flags (mb);
		if (par_short) {
			fprintf (out, "%02lX\n", v);
		}
		else {
			fprintf (out, "file flags:         0x%02lX", v);

			if (v != 0) {
				fputs (" (", out);
				mb_print_file_flags (out, v);
				fputs (")", out);
			}

			fputs ("\n", out);
		}
	}

	if (par_inf_position_x || par_inf_all) {
		v = mb_get_position_x (mb);
		if (par_short) {
			fprintf (out, "%04lX\n", v);
		}
		else {
			fprintf (out, "position x:         0x%04lX\n", v);
		}
	}

	if (par_inf_position_y || par_inf_all) {
		v = mb_get_position_y (mb);
		if (par_short) {
			fprintf (out, "%04lX\n", v);
		}
		else {
			fprintf (out, "position y:         0x%04lX\n", v);
		}
	}

	if (par_inf_folder_id || par_inf_all) {
		v = mb_get_folder_id (mb);
		if (par_short) {
			fprintf (out, "%04lX\n", v);
		}
		else {
			fprintf (out, "folder id:          0x%04lX\n", v);
		}
	}

	if (par_inf_dsize || par_inf_all) {
		v = mb_get_data_size (mb);
		if (par_short) {
			fprintf (out, "%lu\n", v);
		}
		else {
			fprintf (out, "data fork size:     %lu\n", v);
		}
	}

	if (par_inf_rsize || par_inf_all) {
		v = mb_get_rsrc_size (mb);
		if (par_short) {
			fprintf (out, "%lu\n", v);
		}
		else {
			fprintf (out, "resource fork size: %lu\n", v);
		}
	}

	if (par_inf_ctime || par_inf_all) {
		v = mb_get_ctime (mb);
		mac_time_to_ymd (v, &y, &m, &d);
		mac_time_to_hms (v, &hrs, &min, &sec);
		if (par_short) {
			fprintf (out, "%04u-%02u-%02u %02u:%02u:%02u\n",
				y, m + 1, d + 1, hrs, min, sec
			);
		}
		else {
			fprintf (out,
				"creation time:      "
				"%04u-%02u-%02u %02u:%02u:%02u (%lu)\n",
				y, m + 1, d + 1, hrs, min, sec, v
			);
		}
	}

	if (par_inf_mtime || par_inf_all) {
		v = mb_get_mtime (mb);
		mac_time_to_ymd (v, &y, &m, &d);
		mac_time_to_hms (v, &hrs, &min, &sec);
		if (par_short) {
			fprintf (out, "%04u-%02u-%02u %02u:%02u:%02u\n",
				y, m + 1, d + 1, hrs, min, sec
			);
		}
		else {
			fprintf (out,
				"modification time:  "
				"%04u-%02u-%02u %02u:%02u:%02u (%lu)\n",
				y, m + 1, d + 1, hrs, min, sec, v
			);
		}
	}

	fflush (out);

	return (0);
}

static
int mb_info (const char *fname, FILE *out)
{
	int  r;
	FILE *fp;

	fp = fopen (fname, "rb");
	if (fp == NULL) {
		fprintf (stderr, "error opening file (%s)\n", fname);
		return (1);
	}

	r = mb_info_fp (fp, out);

	fclose (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r;
	int  ret;
	char **optarg;

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'a':
			par_inf_all = 1;
			break;

		case 'c':
			par_inf_creator = 1;
			par_inf_all = 0;
			break;

		case 'C':
			par_inf_ctime = 1;
			par_inf_all = 0;
			break;

		case 'f':
			par_inf_fname = 1;
			par_inf_all = 0;
			break;

		case 'F':
			par_inf_folder_id = 1;
			par_inf_all = 0;
			break;

		case 'i':
			ret |= mb_info (optarg[0], stdout);
			break;

		case 'I':
			par_inf_mf = 1;
			par_inf_all = 0;
			break;

		case 'l':
			par_inf_finder = 1;
			par_inf_all = 0;
			break;

		case 'L':
			par_inf_file_flags = 1;
			par_inf_all = 0;
			break;

		case 'M':
			par_inf_mtime = 1;
			par_inf_all = 0;
			break;

		case 's':
			par_short = 1;
			break;

		case 't':
			par_inf_type = 1;
			par_inf_all = 0;
			break;

		case 'x':
			par_inf_position_x = 1;
			par_inf_all = 0;
			break;

		case 'y':
			par_inf_position_y = 1;
			par_inf_all = 0;
			break;

		case 0:
			ret |= mb_info (optarg[0], stdout);
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (ret) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
