/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc.c                                              *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011-2013 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdio.h>
#include <string.h>

#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/unixfile.h>


#define MACC_BUF_MAX 4096


const char           *arg0 = NULL;

static const char    *par_mac_name = NULL;

static int           par_rle = 0;

static const char    *par_data_name1 = NULL;
static const char    *par_data_name2 = NULL;

static unsigned char par_buf[MACC_BUF_MAX];


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'd', 1, "data", "string", "Set the data file name for appledouble [none]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'I', 1, "input-type", "string", "Set the input file type [auto]" },
	{ 'm', 1, "mac-name", "string", "Set the macintosh file name [none]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'O', 1, "output-type", "string", "Set the output file type [auto]" },
	{ 'r', 0, "rle", NULL, "Enable RLE compression for hqx [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"macc: convert macintosh files",
		"usage: macc [options] [input [output]]",
		opts
	);

	fputs (
		"\nFile types are:\n"
		"  ad, as, bin, hqx, mac, mar, mb\n",
		stdout
	);

	fflush (stdout);
}

void print_version (void)
{
	fputs (
		"macc version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2011-2013 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
unsigned macc_get_type (const char *fname, const char *type)
{
	unsigned   i;
	const char *ext;

	if (type != NULL) {
		if (strcmp (type, "ad") == 0) {
			return (MACC_AS);
		}
		else if (strcmp (type, "as") == 0) {
			return (MACC_AS);
		}
		else if (strcmp (type, "bin") == 0) {
			return (MACC_MB);
		}
		else if (strcmp (type, "hex") == 0) {
			return (MACC_HEX);
		}
		else if (strcmp (type, "hcx") == 0) {
			return (MACC_HCX);
		}
		else if (strcmp (type, "hqx") == 0) {
			return (MACC_HQX);
		}
		else if (strcmp (type, "mac") == 0) {
			return (MACC_MAR);
		}
		else if (strcmp (type, "mar") == 0) {
			return (MACC_MAR);
		}
		else if (strcmp (type, "mb") == 0) {
			return (MACC_MB);
		}

		return (MACC_NONE);
	}

	if (fname == NULL) {
		return (MACC_NONE);
	}

	if (strcmp (fname, "-") == 0) {
		return (MACC_NONE);
	}

	ext = fname;

	i = 0;
	while (fname[i] != 0) {
		if (fname[i] == '.') {
			ext = fname + i + 1;
		}

		i += 1;
	}

	if (strcasecmp (ext, "ad") == 0) {
		return (MACC_AS);
	}
	else if (strcasecmp (ext, "adf") == 0) {
		return (MACC_AS);
	}
	else if (strcasecmp (ext, "as") == 0) {
		return (MACC_AS);
	}
	else if (strcasecmp (ext, "asf") == 0) {
		return (MACC_AS);
	}
	else if (strcasecmp (ext, "bin") == 0) {
		return (MACC_MB);
	}
	else if (strcasecmp (ext, "hex") == 0) {
		return (MACC_HEX);
	}
	else if (strcasecmp (ext, "hcx") == 0) {
		return (MACC_HCX);
	}
	else if (strcasecmp (ext, "hqx") == 0) {
		return (MACC_HQX);
	}
	else if (strcasecmp (ext, "mac") == 0) {
		return (MACC_MAR);
	}
	else if (strcasecmp (ext, "mar") == 0) {
		return (MACC_MAR);
	}
	else if (strcasecmp (ext, "mb") == 0) {
		return (MACC_MB);
	}

	return (MACC_NONE);
}


static
macc_inp_t *macc_inp_open (const char *inp, const char *type)
{
	switch (macc_get_type (inp, type)) {
	case MACC_AS:
		return (macc_inp_as_open (inp, par_data_name1));

	case MACC_HEX:
	case MACC_HCX:
		return (macc_inp_hcx_open (inp));

	case MACC_HQX:
		return (macc_inp_hqx_open (inp));

	case MACC_MAR:
		return (macc_inp_mar_open (inp));

	case MACC_MB:
		return (macc_inp_mb_open (inp));

	default:
		fprintf (stderr, "%s: unknown input file type\n", arg0);
		return (NULL);
	}
}

static
macc_out_t *macc_out_open (const char *out, const char *type)
{
	const char *str;

	switch (macc_get_type (out, type)) {
	case MACC_AS:
		if (par_data_name2 != NULL) {
			str = par_data_name2;
		}
		else {
			str = par_data_name1;
		}

		return (macc_out_as_open (out, str));

	case MACC_HEX:
		return (macc_out_hcx_open (out, 0));

	case MACC_HCX:
		return (macc_out_hcx_open (out, 1));

	case MACC_HQX:
		return (macc_out_hqx_open (out, par_rle));

	case MACC_MAR:
		return (macc_out_mar_open (out));

	case MACC_MB:
		return (macc_out_mb_open (out));

	default:
		fprintf (stderr, "%s: unknown output file type\n", arg0);
		return (NULL);
	}
}

static
int macc_copy_data (macc_inp_t *minp, macc_out_t *mout, unsigned long cnt)
{
	unsigned long n;

	if (minp->read_data_start (minp)) {
		return (1);
	}

	if (mout->write_data_start (mout)) {
		return (1);
	}

	while (cnt > 0) {
		n = (cnt < MACC_BUF_MAX) ? cnt : MACC_BUF_MAX;

		if (minp->read_data (minp, par_buf, n)) {
			return (1);
		}

		if (mout->write_data (mout, par_buf, n)) {
			return (1);
		}

		cnt -= n;
	}

	if (mout->write_data_done (mout)) {
		return (1);
	}

	if (minp->read_data_done (minp)) {
		return (1);
	}

	return (0);
}

static
int macc_copy_rsrc (macc_inp_t *minp, macc_out_t *mout, unsigned long cnt)
{
	unsigned long n;

	if (minp->read_rsrc_start (minp)) {
		return (1);
	}

	if (mout->write_rsrc_start (mout)) {
		return (1);
	}

	while (cnt > 0) {
		n = (cnt < MACC_BUF_MAX) ? cnt : MACC_BUF_MAX;

		if (minp->read_rsrc (minp, par_buf, n)) {
			return (1);
		}

		if (mout->write_rsrc (mout, par_buf, n)) {
			return (1);
		}

		cnt -= n;
	}

	if (mout->write_rsrc_done (mout)) {
		return (1);
	}

	if (minp->read_rsrc_done (minp)) {
		return (1);
	}

	return (0);
}

static
int macc_copy_comm (macc_inp_t *minp, macc_out_t *mout, unsigned long cnt)
{
	unsigned long n;

	if (minp->read_comm_start (minp)) {
		return (1);
	}

	if (mout->write_comm_start (mout)) {
		return (1);
	}

	while (cnt > 0) {
		n = (cnt < MACC_BUF_MAX) ? cnt : MACC_BUF_MAX;

		if (minp->read_comm (minp, par_buf, n)) {
			return (1);
		}

		if (mout->write_comm (mout, par_buf, n)) {
			return (1);
		}

		cnt -= n;
	}

	if (mout->write_comm_done (mout)) {
		return (1);
	}

	if (minp->read_comm_done (minp)) {
		return (1);
	}

	return (0);
}

static
int macc_convert_files (macc_inp_t *minp, macc_out_t *mout, unsigned long *mtime)
{
	macfile_t mf;

	mf_init (&mf);

	if (minp->read_header (minp, &mf)) {
		return (1);
	}

	if (par_mac_name != NULL) {
		if (mf_set_name_cstr (&mf, par_mac_name)) {
			fprintf (stderr, "%s: bad mac name (%s)\n",
				arg0, par_mac_name
			);
			return (1);
		}
	}

	if (mout->write_header (mout, &mf)) {
		return (1);
	}

	if (macc_copy_data (minp, mout, mf_get_data_size (&mf))) {
		return (1);
	}

	if (macc_copy_rsrc (minp, mout, mf_get_rsrc_size (&mf))) {
		return (1);
	}

	if (macc_copy_comm (minp, mout, mf_get_comment_size (&mf))) {
		return (1);
	}

	*mtime = mf_get_mtime (&mf);

	return (0);
}

static
int macc_convert (const char *inp, const char *out, const char *itype, const char *otype)
{
	int           r;
	unsigned long mtime;
	macc_inp_t    *minp;
	macc_out_t    *mout;

	minp = macc_inp_open (inp, itype);

	if (minp == NULL) {
		fprintf (stderr, "%s: error opening input file (%s)\n",
			arg0, (inp != NULL) ? inp : "<stdin>"
		);
		return (1);
	}

	mout = macc_out_open (out, otype);

	if (mout == NULL) {
		fprintf (stderr, "%s: error opening output file (%s)\n",
			arg0, (out != NULL) ? out : "<stdout>"
		);
		minp->close (minp);
		return (1);
	}

	r = macc_convert_files (minp, mout, &mtime);

	mout->close (mout);
	minp->close (minp);

	if ((out != NULL) && (strcmp (out, "-") != 0)) {
		uf_set_mtime (out, mtime);
	}

	return (r);
}

int main (int argc, char **argv)
{
	int        r;
	char       **optarg;
	const char *inp, *out;
	const char *itype, *otype;

	arg0 = argv[0];

	inp = NULL;
	out = NULL;

	itype = NULL;
	otype = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'i':
			inp = optarg[0];
			break;

		case 'd':
			if (par_data_name1 == NULL) {
				par_data_name1 = optarg[0];
			}
			else if (par_data_name2 == NULL) {
				par_data_name2 = optarg[0];
			}
			else {
				fprintf (stderr,
					"%s: too many data file names (%s)\n",
					arg0, optarg[0]
				);
				return (1);
			}
			break;

		case 'I':
			itype = optarg[0];
			break;

		case 'm':
			par_mac_name = optarg[0];
			break;

		case 'o':
			out = optarg[0];
			break;

		case 'O':
			otype = optarg[0];
			break;

		case 'r':
			par_rle = 1;
			break;

		case 0:
			if (inp == NULL) {
				inp = optarg[0];
			}
			else if (out == NULL) {
				out = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					arg0, optarg[0]
				);
				return (1);
			}
			break;

		default:
			return (1);
		}
	}

	if (macc_convert (inp, out, itype, otype)) {
		return (1);
	}

	return (0);
}
