/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_inp_mar.c                                      *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/file.h>
#include <lib/macfile.h>
#include <lib/mar.h>


typedef struct {
	macc_inp_t    inp;

	unsigned char mar[128];

	unsigned long data_size;
	unsigned long rsrc_size;
	unsigned long comm_size;

	crc32_t       crc;
} macc_inp_mar_t;


static
void macc_mar_close (macc_inp_t *minp)
{
	mac_close (minp->fp);
	free (minp->ext);
}

static
int macc_mar_read_header (macc_inp_t *minp, macfile_t *mf)
{
	unsigned long  id;
	macc_inp_mar_t *mar;

	mar = minp->ext;

	if (mar_read (mar->mar, minp->fp)) {
		fprintf (stderr, "%s: mar header read error\n", arg0);
		return (1);
	}

	if (mar_is_mar (mar->mar) == 0) {
		fprintf (stderr, "%s: mar header error\n", arg0);
		return (1);
	}

	id = mar_get_id (mar->mar);

	if ((id != MAR_FILE) && (id != MAR_FOLDER_START)) {
		fprintf (stderr, "%s: mar id unknown (0x%08lx)\n", arg0, id);
		return (1);
	}

	mf_init (mf);

	mar_get_macfile (mar->mar, mf);

	mar->data_size = mar_get_data_size (mar->mar);
	mar->rsrc_size = mar_get_rsrc_size (mar->mar);
	mar->comm_size = mar_get_comment_size (mar->mar);

	return (0);
}

static
int macc_mar_read_block_start (macc_inp_t *minp)
{
	macc_inp_mar_t *mar;

	mar = minp->ext;

	crc32_reset (&mar->crc, 0);

	return (0);
}

static
int macc_mar_read_block (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	macc_inp_mar_t *mar;

	mar = minp->ext;

	if (fread (buf, 1, cnt, minp->fp) != cnt) {
		fprintf (stderr, "%s: mar read error\n", arg0);
		return (1);
	}

	crc32_calc (&mar->crc, buf, cnt);

	return (0);
}

static
int macc_mar_read_block_done (macc_inp_t *minp, unsigned long size)
{
	unsigned char  crc_buf[4];
	macc_inp_mar_t *mar;

	if (size == 0) {
		return (0);
	}

	mar = minp->ext;

	crc32_done (&mar->crc);

	if (fread (crc_buf, 1, 4, minp->fp) != 4) {
		fprintf (stderr, "%s: mar read error\n", arg0);
		return (1);
	}

	if (get_uint32_be (crc_buf, 0) != crc32_get_crc (&mar->crc)) {
		fprintf (stderr, "%s: mar crc error\n", arg0);
		return (1);
	}

	size = (128 - ((size + 4) & 127)) & 127;

	while (size > 0) {
		fgetc (minp->fp);
		size -= 1;
	}

	return (0);
}

static
int macc_mar_read_data_done (macc_inp_t *minp)
{
	macc_inp_mar_t *mar;

	mar = minp->ext;

	return (macc_mar_read_block_done (minp, mar->data_size));
}

static
int macc_mar_read_rsrc_done (macc_inp_t *minp)
{
	macc_inp_mar_t *mar;

	mar = minp->ext;

	return (macc_mar_read_block_done (minp, mar->rsrc_size));
}

static
int macc_mar_read_comm_done (macc_inp_t *minp)
{
	macc_inp_mar_t *mar;

	mar = minp->ext;

	return (macc_mar_read_block_done (minp, mar->comm_size));
}

macc_inp_t *macc_inp_mar_open (const char *fname)
{
	macc_inp_t     *minp;
	macc_inp_mar_t *mar;

	mar = malloc (sizeof (macc_inp_mar_t));

	if (mar == NULL) {
		return (NULL);
	}

	minp = &mar->inp;

	minp->ext = mar;

	minp->close = macc_mar_close;

	minp->read_header = macc_mar_read_header;

	minp->read_data_start = macc_mar_read_block_start;
	minp->read_data = macc_mar_read_block;
	minp->read_data_done = macc_mar_read_data_done;

	minp->read_rsrc_start = macc_mar_read_block_start;
	minp->read_rsrc = macc_mar_read_block;
	minp->read_rsrc_done = macc_mar_read_rsrc_done;

	minp->read_comm_start = macc_mar_read_block_start;
	minp->read_comm = macc_mar_read_block;
	minp->read_comm_done = macc_mar_read_comm_done;

	minp->fp = mac_open_inp (fname, "rb", 1);

	if (minp->fp == NULL) {
		free (mar);
		return (NULL);
	}

	crc32_init (&mar->crc, MAR_CRC_POLY, 0);

	return (minp);
}
