/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_inp_mb.c                                       *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/macfile.h>
#include <lib/macbinary.h>


typedef struct {
	macc_inp_t    inp;

	unsigned char mb[128];

	unsigned long data_size;
	unsigned long rsrc_size;
	unsigned long comm_size;
} macc_inp_mb_t;


static
void macc_mb_close (macc_inp_t *minp)
{
	mac_close (minp->fp);
	free (minp->ext);
}

static
int macc_mb_read_header (macc_inp_t *minp, macfile_t *mf)
{
	unsigned      vers;
	macc_inp_mb_t *mb;

	mb = minp->ext;

	if (mb_read (mb->mb, minp->fp)) {
		fprintf (stderr, "%s: mb header read error\n", arg0);
		return (1);
	}

	vers = mb_get_version (mb->mb);

	if (vers == 2) {
		;
	}
	else if (vers == 1) {
		fprintf (stderr, "%s: warning: macbinary version 1\n", arg0);
	}
	else {
		fprintf (stderr, "%s: mb header error\n", arg0);
		return (1);
	}

	mf_init (mf);

	mb_get_macfile (mb->mb, mf);

	mb->data_size = mb_get_data_size (mb->mb);
	mb->rsrc_size = mb_get_rsrc_size (mb->mb);
	mb->comm_size = mb_get_getinfo_size (mb->mb);

	return (0);
}

static
int macc_mb_read_block_start (macc_inp_t *minp)
{
	return (0);
}

static
int macc_mb_read_block (macc_inp_t *minp, void *buf, unsigned long cnt)
{
	if (fread (buf, 1, cnt, minp->fp) != cnt) {
		fprintf (stderr, "%s: mb read error\n", arg0);
		return (1);
	}

	return (0);
}

static
int macc_mb_read_block_done (macc_inp_t *minp, unsigned long size)
{
	size = (128 - (size & 127)) & 127;

	while (size > 0) {
		fgetc (minp->fp);
		size -= 1;
	}

	return (0);
}

static
int macc_mb_read_data_done (macc_inp_t *minp)
{
	macc_inp_mb_t *mb;

	mb = minp->ext;

	return (macc_mb_read_block_done (minp, mb->data_size));
}

static
int macc_mb_read_rsrc_done (macc_inp_t *minp)
{
	macc_inp_mb_t *mb;

	mb = minp->ext;

	return (macc_mb_read_block_done (minp, mb->rsrc_size));
}

static
int macc_mb_read_comm_done (macc_inp_t *minp)
{
	macc_inp_mb_t *mb;

	mb = minp->ext;

	return (macc_mb_read_block_done (minp, mb->comm_size));
}

macc_inp_t *macc_inp_mb_open (const char *fname)
{
	macc_inp_t    *minp;
	macc_inp_mb_t *mb;

	mb = malloc (sizeof (macc_inp_mb_t));

	if (mb == NULL) {
		return (NULL);
	}

	minp = &mb->inp;

	minp->ext = mb;

	minp->close = macc_mb_close;

	minp->read_header = macc_mb_read_header;

	minp->read_data_start = macc_mb_read_block_start;
	minp->read_data = macc_mb_read_block;
	minp->read_data_done = macc_mb_read_data_done;

	minp->read_rsrc_start = macc_mb_read_block_start;
	minp->read_rsrc = macc_mb_read_block;
	minp->read_rsrc_done = macc_mb_read_rsrc_done;

	minp->read_comm_start = macc_mb_read_block_start;
	minp->read_comm = macc_mb_read_block;
	minp->read_comm_done = macc_mb_read_comm_done;

	minp->fp = mac_open_inp (fname, "rb", 1);

	if (minp->fp == NULL) {
		free (mb);
		return (NULL);
	}

	return (minp);
}
