/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_out_as.c                                       *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/applesingle.h>
#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/file.h>
#include <lib/macfile.h>


typedef struct {
	macc_out_t    out;

	applesingle_t as;

	char          *data_name;
	FILE          *data_fp;

	unsigned long data_offset;
	unsigned long rsrc_offset;
} macc_out_as_t;


static
void macc_as_close (macc_out_t *mout)
{
	macc_out_as_t *as;

	as = mout->ext;

	mac_close (as->data_fp);
	free (as->data_name);

	mac_close (mout->fp);
	free (mout->ext);
}

static
int macc_as_write_header (macc_out_t *mout, macfile_t *mf)
{
	unsigned long       datn, rscn;
	const unsigned char *name;
	unsigned char       buf[64];
	macc_out_as_t       *as;

	as = mout->ext;

	as_init (&as->as, mout->fp);

	name = mf_get_name_pstr (mf);
	datn = mf_get_data_size (mf);
	rscn = mf_get_rsrc_size (mf);

	as_add_entry (&as->as, AS_ENTRY_NAME, name[0], 0);
	as_add_entry (&as->as, AS_ENTRY_TIME, 16, 0);
	as_add_entry (&as->as, AS_ENTRY_FINDER_INFO, 32, 0);
	as_add_entry (&as->as, AS_ENTRY_FILE_INFO, 4, 0);

	if ((datn > 0) && (as->data_fp == NULL)) {
		as_add_entry (&as->as, AS_ENTRY_DATA, datn, 0);
	}

	if (rscn > 0) {
		as_add_entry (&as->as, AS_ENTRY_RSRC, rscn, 0);
	}

	as_fix_offsets (&as->as);

	if (as_write (&as->as, name + 1, AS_ENTRY_NAME, 0, name[0])) {
		return (1);
	}

	set_uint32_be (buf, 0, as_time_from_mac (mf_get_ctime (mf)));
	set_uint32_be (buf, 4, as_time_from_mac (mf_get_mtime (mf)));
	set_uint32_be (buf, 8, 0);
	set_uint32_be (buf, 12, 0);

	if (as_write (&as->as, buf, AS_ENTRY_TIME, 0, 16)) {
		return (1);
	}

	set_uint32_be (buf, 0, mf_get_type (mf));
	set_uint32_be (buf, 4, mf_get_creator (mf));
	set_uint16_be (buf, 8, mf_get_finder_flags (mf));
	set_uint16_be (buf, 10, mf_get_position_y (mf));
	set_uint16_be (buf, 12, mf_get_position_x (mf));
	set_uint16_be (buf, 14, mf_get_folder_id (mf));
	set_uint16_be (buf, 16, mf_get_scroll_y (mf));
	set_uint16_be (buf, 18, mf_get_scroll_x (mf));

	memset (buf + 20, 0, 12);

	if (as_write (&as->as, buf, AS_ENTRY_FINDER_INFO, 0, 32)) {
		return (1);
	}

	set_uint32_be (buf, 0, mf_get_file_flags (mf));

	if (as_write (&as->as, buf, AS_ENTRY_FILE_INFO, 0, 4)) {
		return (1);
	}

	if (as_write_header (&as->as, (as->data_fp != 0))) {
		return (1);
	}

	return (0);
}

static
int macc_as_write_data_start (macc_out_t *mout)
{
	macc_out_as_t *as;

	as = mout->ext;

	as->data_offset = 0;

	return (0);
}

static
int macc_as_write_data (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	macc_out_as_t *as;

	as = mout->ext;

	if (as->data_fp != NULL) {
		if (fwrite (buf, 1, cnt, as->data_fp) != cnt) {
			fprintf (stderr, "%s: as data write error\n", arg0);
			return (1);
		}
	}
	else {
		if (as_write (&as->as, buf, AS_ENTRY_DATA, as->data_offset, cnt)) {
			fprintf (stderr, "%s: as data write error\n", arg0);
			return (1);
		}
	}

	as->data_offset += cnt;

	return (0);
}

static
int macc_as_write_data_done (macc_out_t *mout)
{
	return (0);
}

static
int macc_as_write_rsrc_start (macc_out_t *mout)
{
	macc_out_as_t *as;

	as = mout->ext;

	as->rsrc_offset = 0;

	return (0);
}

static
int macc_as_write_rsrc (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	macc_out_as_t *as;

	as = mout->ext;

	if (as_write (&as->as, buf, AS_ENTRY_RSRC, as->rsrc_offset, cnt)) {
		fprintf (stderr, "%s: as resource write error\n", arg0);
		return (1);
	}

	as->rsrc_offset += cnt;

	return (0);
}

static
int macc_as_write_rsrc_done (macc_out_t *mout)
{
	return (0);
}

static
int macc_as_write_comm_start (macc_out_t *mout)
{
	return (0);
}

static
int macc_as_write_comm (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	return (0);
}

static
int macc_as_write_comm_done (macc_out_t *mout)
{
	return (0);
}

macc_out_t *macc_out_as_open (const char *fname, const char *dname)
{
	macc_out_t    *mout;
	macc_out_as_t *as;

	as = malloc (sizeof (macc_out_as_t));

	if (as == NULL) {
		return (NULL);
	}

	mout = &as->out;

	mout->ext = as;

	mout->close = macc_as_close;

	mout->write_header = macc_as_write_header;

	mout->write_data_start = macc_as_write_data_start;
	mout->write_data = macc_as_write_data;
	mout->write_data_done = macc_as_write_data_done;

	mout->write_rsrc_start = macc_as_write_rsrc_start;
	mout->write_rsrc = macc_as_write_rsrc;
	mout->write_rsrc_done = macc_as_write_rsrc_done;

	mout->write_comm_start = macc_as_write_comm_start;
	mout->write_comm = macc_as_write_comm;
	mout->write_comm_done = macc_as_write_comm_done;

	mout->fp = mac_open_out (fname, "wb", 1);

	if (mout->fp == NULL) {
		free (as);
		return (NULL);
	}

	if (dname == NULL) {
		as->data_name = NULL;
		as->data_fp = NULL;
	}
	else {
		as->data_fp = mac_open_out (dname, "wb", 1);

		if (as->data_fp == NULL) {
			mac_close (mout->fp);
			free (as);
			return (NULL);
		}

		as->data_name = strdup (dname);
	}

	return (mout);
}
