/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_out_hqx.c                                      *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/file.h>
#include <lib/hqx.h>
#include <lib/macfile.h>


typedef struct {
	macc_out_t    out;

	hqx_t         hqx;

	int           rle;

	crc16_t       crc;
} macc_out_hqx_t;


static
void macc_hqx_close (macc_out_t *mout)
{
	mac_close (mout->fp);
	free (mout->ext);
}

static
int macc_hqx_write_header (macc_out_t *mout, macfile_t *mf)
{
	macc_out_hqx_t *hqx;

	hqx = mout->ext;

	hqx_init_write (&hqx->hqx, mout->fp, hqx->rle);

	hqx_set_macfile (&hqx->hqx, mf);

	hqx_set_finder_flags (&hqx->hqx, hqx_get_finder_flags (&hqx->hqx) & ~MF_FINDER_INITED);

	hqx_set_crc (&hqx->hqx, hqx_calc_crc (&hqx->hqx));

	hqx_write_header (&hqx->hqx);

	return (0);
}

static
int macc_hqx_write_block_start (macc_out_t *mout)
{
	macc_out_hqx_t *hqx;

	hqx = mout->ext;

	crc16_reset (&hqx->crc, 0);

	return (0);
}

static
int macc_hqx_write_block (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	macc_out_hqx_t *hqx;

	hqx = mout->ext;

	hqx_write (&hqx->hqx, buf, cnt);

	crc16_calc (&hqx->crc, buf, cnt);

	return (0);
}

static
int macc_hqx_write_block_done (macc_out_t *mout)
{
	unsigned char  crc_buf[2];
	macc_out_hqx_t *hqx;

	hqx = mout->ext;

	crc16_done (&hqx->crc);

	set_uint16_be (crc_buf, 0, crc16_get_crc (&hqx->crc));

	hqx_write (&hqx->hqx, crc_buf, 2);

	return (0);
}

static
int macc_hqx_write_rsrc_done (macc_out_t *mout)
{
	macc_out_hqx_t *hqx;

	if (macc_hqx_write_block_done (mout)) {
		return (1);
	}

	hqx = mout->ext;

	hqx_flush (&hqx->hqx);

	return (0);
}

static
int macc_hqx_write_comm_start (macc_out_t *mout)
{
	return (0);
}

static
int macc_hqx_write_comm (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	return (0);
}

static
int macc_hqx_write_comm_done (macc_out_t *mout)
{
	return (0);
}

macc_out_t *macc_out_hqx_open (const char *fname, int rle)
{
	macc_out_t     *mout;
	macc_out_hqx_t *hqx;

	hqx = malloc (sizeof (macc_out_hqx_t));

	if (hqx == NULL) {
		return (NULL);
	}

	mout = &hqx->out;

	mout->ext = hqx;

	mout->close = macc_hqx_close;

	mout->write_header = macc_hqx_write_header;

	mout->write_data_start = macc_hqx_write_block_start;
	mout->write_data = macc_hqx_write_block;
	mout->write_data_done = macc_hqx_write_block_done;

	mout->write_rsrc_start = macc_hqx_write_block_start;
	mout->write_rsrc = macc_hqx_write_block;
	mout->write_rsrc_done = macc_hqx_write_rsrc_done;

	mout->write_comm_start = macc_hqx_write_comm_start;
	mout->write_comm = macc_hqx_write_comm;
	mout->write_comm_done = macc_hqx_write_comm_done;

	mout->fp = mac_open_out (fname, "w", 1);

	if (mout->fp == NULL) {
		free (hqx);
		return (NULL);
	}

	hqx->rle = rle;

	crc16_init (&hqx->crc, CRC_CRC16_CCITT, 0);

	return (mout);
}
