/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/macc/macc_out_mar.c                                      *
 * Created:     2011-06-30 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "macc.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/file.h>
#include <lib/macfile.h>
#include <lib/mar.h>


typedef struct {
	macc_out_t    out;

	unsigned char mar[128];

	unsigned long data_size;
	unsigned long rsrc_size;
	unsigned long comm_size;

	crc32_t       crc;
} macc_out_mar_t;


static
void macc_mar_close (macc_out_t *mout)
{
	mac_close (mout->fp);
	free (mout->ext);
}

static
int macc_mar_write_header (macc_out_t *mout, macfile_t *mf)
{
	macc_out_mar_t *mar;

	mar = mout->ext;

	mar_init (mar->mar);

	if (mf_get_file_flags (mf) & MF_FILE_ISDIR) {
		mar_set_id (mar->mar, MAR_FOLDER_START);
	}
	else {
		mar_set_id (mar->mar, MAR_FILE);
	}

	mar_set_macfile (mar->mar, mf);

	mar_set_crc (mar->mar);

	if (mar_write (mar->mar, mout->fp)) {
		fprintf (stderr, "%s: mar header write error\n", arg0);
		return (1);
	}

	mar->data_size = mf_get_data_size (mf);
	mar->rsrc_size = mf_get_rsrc_size (mf);
	mar->comm_size = mf_get_comment_size (mf);

	return (0);
}

static
int macc_mar_write_block_start (macc_out_t *mout)
{
	macc_out_mar_t *mar;

	mar = mout->ext;

	crc32_reset (&mar->crc, 0);

	return (0);
}

static
int macc_mar_write_block (macc_out_t *mout, const void *buf, unsigned long cnt)
{
	macc_out_mar_t *mar;

	mar = mout->ext;

	if (fwrite (buf, 1, cnt, mout->fp) != cnt) {
		fprintf (stderr, "%s: mar write error\n", arg0);
		return (1);
	}

	crc32_calc (&mar->crc, buf, cnt);

	return (0);
}

static
int macc_mar_write_block_done (macc_out_t *mout, unsigned long size)
{
	unsigned char  crc_buf[4];
	macc_out_mar_t *mar;

	mar = mout->ext;

	if (size == 0) {
		return (0);
	}

	crc32_done (&mar->crc);

	set_uint32_be (crc_buf, 0, crc32_get_crc (&mar->crc));

	if (fwrite (crc_buf, 1, 4, mout->fp) != 4) {
		fprintf (stderr, "%s: mar write error\n", arg0);
		return (1);
	}

	size = (128 - ((size + 4) & 127)) & 127;

	while (size > 0) {
		fputc (0, mout->fp);
		size -= 1;
	}

	return (0);
}

static
int macc_mar_write_data_done (macc_out_t *mout)
{
	macc_out_mar_t *mar;

	mar = mout->ext;

	return (macc_mar_write_block_done (mout, mar->data_size));
}

static
int macc_mar_write_rsrc_done (macc_out_t *mout)
{
	macc_out_mar_t *mar;

	mar = mout->ext;

	return (macc_mar_write_block_done (mout, mar->rsrc_size));
}

static
int macc_mar_write_comm_done (macc_out_t *mout)
{
	macc_out_mar_t *mar;

	mar = mout->ext;

	return (macc_mar_write_block_done (mout, mar->comm_size));
}

macc_out_t *macc_out_mar_open (const char *fname)
{
	macc_out_t     *mout;
	macc_out_mar_t *mar;

	mar = malloc (sizeof (macc_out_mar_t));

	if (mar == NULL) {
		return (NULL);
	}

	mout = &mar->out;

	mout->ext = mar;

	mout->close = macc_mar_close;

	mout->write_header = macc_mar_write_header;

	mout->write_data_start = macc_mar_write_block_start;
	mout->write_data = macc_mar_write_block;
	mout->write_data_done = macc_mar_write_data_done;

	mout->write_rsrc_start = macc_mar_write_block_start;
	mout->write_rsrc = macc_mar_write_block;
	mout->write_rsrc_done = macc_mar_write_rsrc_done;

	mout->write_comm_start = macc_mar_write_block_start;
	mout->write_comm = macc_mar_write_block;
	mout->write_comm_done = macc_mar_write_comm_done;

	mout->fp = mac_open_out (fname, "wb", 1);

	if (mout->fp == NULL) {
		free (mar);
		return (NULL);
	}

	crc32_init (&mar->crc, MAR_CRC_POLY, 0);

	return (mout);
}
