/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mar/mar_create.c                                         *
 * Created:     2011-06-26 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mar.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/crc.h>
#include <lib/dir.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/mactime.h>
#include <lib/mar.h>
#include <lib/strings.h>


static int mar_encode_entry (void *ext, unsigned op, const mac_stat_t *ms);


static const char    *arg0 = NULL;

static int           par_verbose = 0;

static int           par_reset = 0;

static const char    *par_folder_file = ".folder";


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'f', 1, "mar-file", "string", "Set the output file name" },
	{ 'F', 1, "folder-file", "string", "Set the folder file name [.folder]" },
	{ 'i', 1, "input", "string", "Set an input file name" },
	{ 'r', 0, "reset", NULL, "Reset file and finder flags [no]" },
	{ 's', 2, "set", "2 * string", "Set a property" },
	{ 'v', 0, "verbose", NULL, "Verbose operation [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mar -c: create mar archives",
		"usage: mar -c [options] [archive] [file_or_dir]",
		opts
	);

	fflush (stdout);
}

static
void mar_encode_reset (unsigned char *mar, int folder)
{
	mar_set_position_x (mar, 0);
	mar_set_position_y (mar, 0);

	mar_set_scroll_x (mar, 0);
	mar_set_scroll_y (mar, 0);

	mar_set_folder_id (mar, 0);

	if (folder) {
		mar_set_data_size (mar, 0);
		mar_set_rsrc_size (mar, 0);
	}

	mar_set_finder_flags (mar, mar_get_finder_flags (mar) & ~MF_FINDER_INITED);
	mar_set_file_flags (mar, 0);

	mar_set_crc (mar);
}

static
int mar_create_block_raw (FILE *inp, FILE *out, unsigned long cnt)
{
	unsigned      n;
	unsigned long crc;

	if (cnt == 0) {
		return (0);
	}

	if (mar_copy_crc (inp, out, cnt, &crc)) {
		return (1);
	}

	set_uint32_be (par_buf, 0, crc);

	if (fwrite (par_buf, 1, 4, out) != 4) {
		return (1);
	}

	n = (128 - ((cnt + 4) & 127)) & 127;

	if (n > 0) {
		memset (par_buf, 0, n);

		if (fwrite (par_buf, 1, n, out) != n) {
			return (1);
		}
	}

	return (0);
}

static
int mar_create_block_mar (FILE *inp, FILE *out, unsigned long cnt)
{
	unsigned      n;
	unsigned long crc;

	if (cnt == 0) {
		return (0);
	}

	if (mar_copy_crc (inp, out, cnt, &crc)) {
		return (1);
	}

	if (fread (par_buf, 1, 4, inp) != 4) {
		return (1);
	}

	if (get_uint32_be (par_buf, 0) != crc) {
		fprintf (stderr, "%s: crc error\n", arg0);
		return (1);
	}

	if (fwrite (par_buf, 1, 4, out) != 4) {
		return (1);
	}

	n = (128 - ((cnt + 4) & 127)) & 127;

	if (n > 0) {
		if (fread (par_buf, 1, n, inp) != n) {
			return (1);
		}

		memset (par_buf, 0, n);

		if (fwrite (par_buf, 1, n, out) != n) {
			return (1);
		}
	}

	return (0);
}

static
int mar_create_block_mb (FILE *inp, FILE *out, unsigned long cnt)
{
	unsigned      n;
	unsigned long crc;

	if (cnt == 0) {
		return (0);
	}

	if (mar_copy_crc (inp, out, cnt, &crc)) {
		return (1);
	}

	set_uint32_be (par_buf, 0, crc);

	if (fwrite (par_buf, 1, 4, out) != 4) {
		return (1);
	}

	n = (128 - (cnt & 127)) & 127;

	if (n > 0) {
		if (fread (par_buf, 1, n, inp) != n) {
			return (1);
		}
	}

	n = (128 - ((cnt + 4) & 127)) & 127;

	if (n > 0) {
		memset (par_buf, 0, n);

		if (fwrite (par_buf, 1, n, out) != n) {
			return (1);
		}
	}

	return (0);
}

static
int mar_encode_file_raw (FILE *out, const mac_stat_t *ms, FILE *fp)
{
	unsigned char mar[128];
	unsigned long data;
	unsigned long mtime;

	fseek (fp, 0, SEEK_SET);

	mar_init (mar);
	mar_set_id (mar, MAR_FILE);

	if (mar_set_name_cstr (mar, mac_get_last_component (ms->name, '/'))) {
		return (1);
	}

	mar_set_type (mar, par_type_ok ? par_type : MAC_FOURCC_UNKN);
	mar_set_creator (mar, par_creator_ok ? par_creator : MAC_FOURCC_UNKN);

	mar_set_file_flags (mar, par_file_flags_val);
	mar_set_finder_flags (mar, par_finder_flags_val);

	mtime = mac_time_from_unix (ms->mtime);
	mar_set_mtime (mar, mtime);
	mar_set_ctime (mar, mtime);

	data = ms->size;

	mar_set_data_size (mar, data);
	mar_set_rsrc_size (mar, 0);
	mar_set_comment_size (mar, 0);

	mar_set_crc (mar);

	if (fwrite (mar, 1, 128, out) != 128) {
		return (1);
	}

	if (mar_create_block_raw (fp, out, data)) {
		return (1);
	}

	return (0);
}

static
int mar_encode_file_mar (FILE *out, const mac_stat_t *ms, FILE *fp, unsigned char *mar)
{
	if (par_reset) {
		mar_encode_reset (mar, 0);
	}

	if (fwrite (mar, 1, 128, out) != 128) {
		return (1);
	}

	if (mar_create_block_mar (fp, out, mar_get_data_size (mar))) {
		return (1);
	}

	if (mar_create_block_mar (fp, out, mar_get_rsrc_size (mar))) {
		return (1);
	}

	if (mar_create_block_mar (fp, out, mar_get_comment_size (mar))) {
		return (1);
	}

	return (0);
}

static
int mar_encode_file_mb (FILE *out, const mac_stat_t *ms, FILE *fp, unsigned char *mb)
{
	macfile_t     mf;
	unsigned char mar[128];

	mar_init (mar);
	mar_set_id (mar, MAR_FILE);

	mf_init (&mf);
	mb_get_macfile (mb, &mf);
	mar_set_macfile (mar, &mf);

	if (par_reset) {
		mar_encode_reset (mar, 0);
	}

	mar_set_crc (mar);

	if (fwrite (mar, 1, 128, out) != 128) {
		return (1);
	}

	if (mar_create_block_mb (fp, out, mar_get_data_size (mar))) {
		return (1);
	}

	if (mar_create_block_mb (fp, out, mar_get_rsrc_size (mar))) {
		return (1);
	}

	if (mar_create_block_mb (fp, out, mar_get_comment_size (mar))) {
		return (1);
	}

	return (0);
}

static
int mar_encode_file (FILE *out, const mac_stat_t *ms)
{
	int           r;
	FILE          *fp;
	unsigned char mar[128];

	if (par_folder_file != NULL) {
		if (strcmp (mac_get_last_component (ms->name, '/'), par_folder_file) == 0) {
			return (0);
		}
	}

	if (par_verbose) {
		fprintf (stderr, "%s\n", ms->name);
	}

	fp = fopen (ms->name, "rb");

	if (fp == NULL) {
		return (1);
	}

	if (fread (mar, 1, 128, fp) != 128) {
		r = mar_encode_file_raw (out, ms, fp);
	}
	else {
		if (mar_is_mar (mar)) {
			if (mar_get_id (mar) == MAR_FILE) {
				r = mar_encode_file_mar (out, ms, fp, mar);
			}
			else {
				r = mar_encode_file_raw (out, ms, fp);
			}
		}
		else if (mb_get_version (mar) == 2) {
			r = mar_encode_file_mb (out, ms, fp, mar);
		}
		else {
			r = mar_encode_file_raw (out, ms, fp);
		}
	}

	fclose (fp);

	if (r) {
		fprintf (stderr, "*** error (%s)\n", ms->name);
	}

	return (r);
}

static
int mar_encode_folder_start_inf (unsigned char *mar, const char *fname)
{
	FILE          *fp;
	macfile_t     mf;
	unsigned char buf[128];

	fp = fopen (fname, "rb");

	if (fp == NULL) {
		return (1);
	}

	if (fread (buf, 1, 128, fp) != 128) {
		fclose (fp);
		return (1);
	}

	if (mar_is_mar (buf)) {
		memcpy (mar, buf, 128);
	}
	else if (mb_get_version (buf) == 2) {
		if ((mb_get_id (buf) != 1) || (mb_get_type (buf) != MAC_FOURCC_FOLD)) {
			fprintf (stderr,
				"%s: folder file error (macbinary id)\n",
				arg0
			);
			return (1);
		}

		mf_init (&mf);
		mb_get_macfile (buf, &mf);

		mar_init (mar);
		mar_set_macfile (mar, &mf);

		mar_set_file_flags (mar, mar_get_file_flags (mar) | MF_FILE_ISDIR);
		mar_set_id (mar, MAR_FOLDER_START);
	}
	else {
		fseek (fp, 0, SEEK_SET);

		if (mf_load (&mf, fp)) {
			fprintf (stderr,
				"%s: folder file error\n",
				arg0
			);

			fclose (fp);

			return (1);
		}

		mar_init (mar);
		mar_set_macfile (mar, &mf);
	}

	fclose (fp);

	if (mar_get_id (mar) != MAR_FOLDER_START) {
		fprintf (stderr,
			"%s: folder file error (type)\n",
			arg0
		);
		return (1);
	}

	return (0);
}

static
int mar_encode_folder_start (FILE *out, const mac_stat_t *ms, unsigned char *mar)
{
	int           r;
	unsigned long mtime;
	char          *fold;

	if (par_folder_file != NULL) {
		fold = mac_combine_path (ms->name, par_folder_file);
		r = mar_encode_folder_start_inf (mar, fold);
		free (fold);
	}
	else {
		r = 1;
	}

	if (r) {
		mar_init (mar);

		mar_set_id (mar, MAR_FOLDER_START);

		if (mar_set_name_cstr (mar, mac_get_last_component (ms->name, '/'))) {
			return (1);
		}

		mtime = mac_time_from_unix (ms->mtime);
		mar_set_mtime (mar, mtime);
		mar_set_ctime (mar, mtime);

	}

	if (par_reset) {
		mar_encode_reset (mar, 1);
	}

	mar_set_file_flags (mar, mar_get_file_flags (mar) | MF_FILE_ISDIR);

	mar_set_crc (mar);

	if (mar_write (mar, out)) {
		return (1);
	}

	return (0);
}

static
int mar_encode_folder_done (FILE *out, const mac_stat_t *ms, unsigned char *mar)
{
	mar_init (mar);
	mar_set_id (mar, MAR_FOLDER_END);
	mar_set_name_cstr (mar, mac_get_last_component (ms->name, '/'));
	mar_set_crc (mar);

	if (mar_write (mar, out)) {
		return (1);
	}

	return (0);
}

static
int mar_encode_folder (FILE *out, const mac_stat_t *ms)
{
	unsigned char mar[128];

	if (par_verbose) {
		fprintf (stderr, "%s/\n", ms->name);
	}

	if (mar_encode_folder_start (out, ms, mar)) {
		fprintf (stderr, "*** error: folder start (%s)\n", ms->name);
		return (1);
	}

	if (mac_dir_list_dir (ms->name, 0, out, mar_encode_entry)) {
		return (1);
	}

	if (mar_encode_folder_done (out, ms, mar)) {
		fprintf (stderr, "*** error: folder done (%s)\n", ms->name);
		return (1);
	}

	return (0);
}

static
int mar_encode_entry (void *ext, unsigned op, const mac_stat_t *ms)
{
	FILE *out;

	out = ext;

	if (op == MAC_LIST_ENTER) {
		if (mar_encode_folder (out, ms)) {
			return (1);
		}

		return (0);
	}

	if (op == MAC_LIST_LEAVE) {
		return (0);
	}

	if (op == MAC_LIST_LIST) {
		if (ms->is_file) {
			if (mar_encode_file (out, ms)) {
				return (1);
			}
		}
		else if (ms->is_dir) {
			return (0);
		}
		else {
			fprintf (stderr, "*** error: bad file type (%s)\n",
				ms->name
			);
			return (1);
		}
	}

	return (0);
}

static
int mar_create_fp (FILE *out, const char *src)
{
	while ((src[0] == '.') && (src[1] == '/')) {
		src += 2;

		while (*src == '/') {
			src += 1;
		}
	}

	if (mac_dir_list (src, 0, out, mar_encode_entry)) {
		return (1);
	}

	return (0);
}

static
int mar_create (const char *out, const char *inp)
{
	int  r;
	FILE *fout;

	fout = mac_open_out (out, "wb", 1);

	if (fout == NULL) {
		return (1);
	}

	r = mar_create_fp (fout, inp);

	mac_close (fout);

	return (r);
}

int mar_create_main (int argc, char **argv)
{
	int        r;
	char       **optarg;
	const char *inp, *out;

	arg0 = argv[0];

	inp = NULL;
	out = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'f':
			out = optarg[0];
			break;

		case 'F':
			par_folder_file = optarg[0];

			if (par_folder_file[0] == 0) {
				par_folder_file = NULL;
			}
			break;

		case 'i':
			inp = optarg[0];
			break;

		case 'r':
			par_reset = 1;
			break;

		case 's':
			if (mar_set_property (optarg[0], optarg[1])) {
				fprintf (stderr, "%s: bad property (%s=%s)\n",
					arg0, optarg[0], optarg[1]
				);
				return (1);
			}
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (out == NULL) {
				out = optarg[0];
			}
			else if (inp == NULL) {
				inp = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n", arg0, optarg[0]);
				return (1);
			}
			break;

		default:
			return (1);
		}
	}

	if (inp == NULL) {
		return (1);
	}

	if (mar_create (out, inp)) {
		return (1);
	}

	return (0);
}
