/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mar/mar_encode.c                                         *
 * Created:     2011-06-25 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mar.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/mactime.h>
#include <lib/mar.h>
#include <lib/strings.h>
#include <lib/unixfile.h>



static const char *arg0 = NULL;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'd', 1, "data-file", "string", "Set the data fork file name [none]" },
	{ 'f', 1, "mar-file", "string", "Set the mar file name [none]" },
	{ 'i', 1, "info-file", "string", "Set the info file name [none]" },
	{ 'l', 1, "finder-flags", "int", "Set the finder flags [0]" },
	{ 'L', 1, "file-flags", "int", "Set the file flags [0]" },
	{ 'm', 1, "mac-name", "string", "Set the mac file name [same as data file]" },
	{ 'r', 1, "rsrc", "string", "Set the resource fork file name [none]" },
	{ 's', 2, "set", "2 * string", "Set a property" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mar -e: encode mar files",
		"usage: mar -e [options] [mar [data [resource [finf]]]]",
		opts
	);

	print_help_properties();

	fflush (stdout);
}

static
int mar_encode_block (FILE *inp, FILE *out, unsigned long cnt)
{
	unsigned long n;
	unsigned long crc;

	if (cnt == 0) {
		return (0);
	}

	if (mar_copy_crc (inp, out, cnt, &crc)) {
		return (1);
	}

	set_uint32_be (par_buf, 0, crc);

	if (fwrite (par_buf, 1, 4, out) != 4) {
		return (1);
	}

	n = (128 - ((cnt + 4) & 127)) & 127;

	if (n > 0) {
		memset (par_buf, 0, n);

		if (fwrite (par_buf, 1, n, out) != n) {
			return (1);
		}
	}

	return (0);
}

static
int mar_encode_fp (unsigned char *mar, FILE *fmar, FILE *fdat, FILE *frsc)
{
	unsigned long datn, rscn;

	datn = mar_get_data_size (mar);
	rscn = mar_get_rsrc_size (mar);

	if ((datn > 0) && (fdat == NULL)) {
		fprintf (stderr, "%s: can't open data file\n", arg0);
		return (1);
	}

	if ((rscn > 0) && (frsc == NULL)) {
		fprintf (stderr, "%s: can't open resource file\n", arg0);
		return (1);
	}

	mar_set_crc (mar);

	if (mar_write (mar, fmar)) {
		fprintf (stderr, "%s: error writing mar header\n", arg0);
		return (1);
	}

	if (fdat != NULL) {
		if (mar_encode_block (fdat, fmar, datn)) {
			return (1);
		}
	}

	if (frsc != NULL) {
		if (mar_encode_block (frsc, fmar, rscn)) {
			return (1);
		}
	}

	return (0);
}

static
int mar_encode_header (unsigned char *mar, const char *dat, const char *rsc, const char *inf)
{
	FILE          *finf;
	unsigned long datn, rscn;
	unsigned long rsct, datt;
	unsigned long val;

	datn = 0;
	rscn = 0;

	datt = 0;
	rsct = 0;

	if (uf_stat (dat, &datn, &datt)) {
		datn = 0;
		datt = 0;
	}

	if (uf_stat (rsc, &rscn, &rsct)) {
		rscn = 0;
		rsct = 0;
	}

	if (datt == 0) {
		datt = rsct;
	}

	mar_init (mar);
	mar_set_id (mar, MAR_FILE);

	finf = mac_open_inp (inf, "r", 0);

	if ((inf != NULL) && (finf == NULL)) {
		fprintf (stderr, "%s:error opening info file (%s)\n", arg0, inf);
		return (1);
	}

	if (finf == NULL) {
		const char    *cstr;
		unsigned char *pstr;

		if (dat == NULL) {
			cstr = "Noname";
		}
		else {
			cstr = mac_get_last_component (dat, '/');
		}

		pstr = mac_pstring_encode (cstr);

		if (pstr != NULL) {
			mar_set_name_pstr (mar, pstr);
			free (pstr);
		}

		if (par_mar_id_ok && (par_mar_id != MAR_FILE)) {
			mar_set_type (mar, 0);
			mar_set_creator (mar, 0);
		}
		else {
			mar_set_type (mar, MAC_FOURCC_UNKN);
			mar_set_creator (mar, MAC_FOURCC_UNKN);
		}

		mar_set_ctime (mar, datt);
		mar_set_mtime (mar, datt);
	}
	else {
		macfile_t mf;

		if (mf_load (&mf, finf)) {
			fprintf (stderr, "%s: error loading info file\n", arg0);
			return (1);
		}

		mar_set_macfile (mar, &mf);
	}

	mac_close (finf);

	if (par_mac_name_ok) {
		if (mar_set_name_cstr (mar, par_mac_name)) {
			fprintf (stderr, "%s: bad mac file name (%s)\n",
				arg0, par_mac_name
			);
			return (1);
		}
	}

	if (par_mar_id_ok) {
		mar_set_id (mar, par_mar_id);
	}

	if (par_type_ok) {
		mar_set_type (mar, par_type);
	}

	if (par_creator_ok) {
		mar_set_creator (mar, par_creator);
	}

	if (par_file_flags_ok) {
		val = mar_get_file_flags (mar);

		val &= ~par_file_flags_msk;
		val |= par_file_flags_val & par_file_flags_msk;

		mar_set_file_flags (mar, val);
	}

	if (par_finder_flags_ok) {
		val = mar_get_finder_flags (mar);

		val &= ~par_finder_flags_msk;
		val |= par_finder_flags_val & par_finder_flags_msk;

		mar_set_finder_flags (mar, val);
	}

	if (par_ctime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mar_set_ctime (mar, datt);
		}
		else if (strcmp (par_mtime, "mtime") == 0) {
			mar_set_ctime (mar, mar_get_mtime (mar));
		}
		else {
			mar_set_ctime (mar, mac_time_from_string (par_ctime));
		}
	}

	if (par_mtime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mar_set_mtime (mar, datt);
		}
		else if (strcmp (par_mtime, "ctime") == 0) {
			mar_set_mtime (mar, mar_get_ctime (mar));
		}
		else {
			mar_set_mtime (mar, mac_time_from_string (par_mtime));
		}
	}

	if (par_position_x_ok) {
		mar_set_position_x (mar, par_position_x);
	}

	if (par_position_y_ok) {
		mar_set_position_y (mar, par_position_y);
	}

	mar_set_data_size (mar, datn);
	mar_set_rsrc_size (mar, rscn);

	mar_set_comment_size (mar, 0);

	if (par_window_x1_ok) {
		mar_set_window_x1 (mar, par_window_x1);
	}

	if (par_window_y1_ok) {
		mar_set_window_y1 (mar, par_window_y1);
	}

	if (par_window_x2_ok) {
		mar_set_window_x2 (mar, par_window_x2);
	}

	if (par_window_y2_ok) {
		mar_set_window_y2 (mar, par_window_y2);
	}

	if (par_window_view_ok) {
		mar_set_window_view (mar, par_window_view);
	}

	if (par_scroll_x_ok) {
		mar_set_scroll_x (mar, par_scroll_x);
	}

	if (par_scroll_y_ok) {
		mar_set_scroll_y (mar, par_scroll_y);
	}

	if (mar_get_id (mar) == MAR_FOLDER_START) {
		mar_set_file_flags (mar, mar_get_file_flags (mar) | MF_FILE_ISDIR);
	}
	else {
		mar_set_file_flags (mar, mar_get_file_flags (mar) & ~MF_FILE_ISDIR);
	}

	return (0);
}

static
int mar_encode (const char *out, const char *dat, const char *rsc, const char *inf)
{
	int           r;
	FILE          *fout, *fdat, *frsc;
	unsigned char mar[128];

	fout = mac_open_out (out, "wb", 0);

	if (fout == NULL) {
		fprintf (stderr, "%s: error opening mar file (%s)\n", arg0, out);
		return (1);
	}

	fdat = mac_open_inp (dat, "rb", 0);

	if ((dat != NULL) && (fdat == NULL)) {
		fprintf (stderr, "%s: error opening data fork file (%s)\n", arg0, dat);
		return (1);
	}

	frsc = mac_open_inp (rsc, "rb", 0);

	if ((rsc != NULL) && (frsc == NULL)) {
		fprintf (stderr, "%s: error opening resource fork file (%s)\n", arg0, rsc);
		return (1);
	}

	r = mar_encode_header (mar, dat, rsc, inf);

	if (r == 0) {
		r = mar_encode_fp (mar, fout, fdat, frsc);
	}

	mac_close (frsc);
	mac_close (fdat);
	mac_close (fout);

	uf_set_mtime (out, mar_get_mtime (mar));

	return (r);
}

int mar_encode_main (int argc, char **argv)
{
	int        r;
	const char *out, *dat, *rsc, *inf;
	char       **optarg;

	arg0 = argv[0];

	out = NULL;
	dat = NULL;
	rsc = NULL;
	inf = NULL;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'd':
			dat = optarg[0];
			break;

		case 'f':
			out = optarg[0];
			break;

		case 'i':
			inf = optarg[0];
			break;

		case 'l':
			if (mar_set_property ("finder-flags", optarg[0])) {
				fprintf (stderr, "%s: bad finder flags (%s)\n",
					arg0, optarg[0]
				);
				return (1);
			}
			break;

		case 'L':
			if (mar_set_property ("file-flags", optarg[0])) {
				fprintf (stderr, "%s: bad file flags (%s)\n",
					arg0, optarg[0]
				);
				return (1);
			}
			break;

		case 'm':
			par_mac_name = optarg[0];
			par_mac_name_ok = 1;
			break;

		case 'r':
			rsc = optarg[0];
			break;

		case 's':
			if (mar_set_property (optarg[0], optarg[1])) {
				fprintf (stderr, "%s: bad property (%s=%s)\n",
					arg0, optarg[0], optarg[1]
				);
				return (1);
			}
			break;

		case 0:
			if (out == NULL) {
				out = optarg[0];
			}
			else if (dat == NULL) {
				dat = optarg[0];
			}
			else if (rsc == NULL) {
				rsc = optarg[0];
			}
			else if (inf == NULL) {
				inf = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv[0], optarg[0]
				);
				return (1);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (mar_encode (out, dat, rsc, inf)) {
		return (1);
	}

	return (0);
}
