/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mar/mar_fix.c                                            *
 * Created:     2011-06-26 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mar.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>

#include <lib/buffer.h>
#include <lib/getopt.h>
#include <lib/macfile.h>
#include <lib/mactime.h>
#include <lib/mar.h>
#include <lib/unixfile.h>


static const char *arg0 = NULL;

static char       par_fix_name = 0;

static char       par_force = 0;

static char       par_zero_mar = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'c', 1, "creator", "fourcc", "Set the creator" },
	{ 'C', 1, "ctime", "string", "Set creation time" },
	{ 'f', 1, "mar-file", "string", "Set an archive file name" },
	{ 'F', 0, "force", NULL, "Don't check header integrity [no]" },
	{ 'I', 1, "mar-id", "fourcc", "Set the mar id" },
	{ 'l', 1, "finder-flags", "int", "Set the finder flags" },
	{ 'L', 1, "file-flags", "int", "Set the file flags" },
	{ 'm', 1, "mac-name", "string", "Set the mac file name" },
	{ 'M', 1, "mtime", "string", "Set modification time" },
	{ 'n', 0, "fix-name", NULL, "Fix the mac file name" },
	{ 's', 2, "set", "2 * string", "Set a property" },
	{ 't', 1, "type", "fourcc", "Set the type" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'x', 1, "position-x", "int", "Set x position" },
	{ 'y', 1, "position-y", "int", "Set y position" },
	{ 'z', 0, "zero-mar", NULL, "Clear unused header fields [no]" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mar -f: set mar header fields",
		"usage: mar -f [options] [file...]",
		opts
	);

	print_help_properties();

	fflush (stdout);
}

static
char *str_base (const char *fname, const char *ext)
{
	char     *ret;
	unsigned n, m;

	n = strlen (fname);
	m = strlen (ext);

	ret = malloc (n + 1);

	if (ret == NULL) {
		return (NULL);
	}

	strcpy (ret, fname);

	if (n <= m) {
		return (ret);
	}

	if (strcmp (ret + n - m, ext) == 0) {
		ret[n - m] = 0;
	}

	return (ret);
}

static
void mar_zero_unused (unsigned char *mar)
{
	unsigned i;

	i = mar[8] + 8 + 1;
	while (i < 72) {
		mar[i++] = 0;
	}

	for (i = 116; i < 124; i++) {
		mar[i] = 0;
	}

	if (mar_get_id (mar) == MAR_FILE) {
		mar_set_folder_id (mar, 0);
		mar_set_scroll_y (mar, 0);
		mar_set_scroll_x (mar, 0);
	}
}

static
int mar_fix_mar (unsigned char *mar, FILE *fp, const char *fname)
{
	unsigned long ftime;
	unsigned long v;

	if (mar_is_mar (mar) == 0) {
		if (par_force == 0) {
			fprintf (stderr, "%s: not a mar file\n", fname);
			return (1);
		}

		mar_init (mar);
		mar_set_id (mar, MAR_FILE);
	}

	if (par_fix_name) {
		char *name;

		name = str_base (fname, ".mac");

		if (name != NULL) {
			if (mar_set_name_cstr (mar, name)) {
				fprintf (stderr, "%s: bad file name\n", fname);
				free (name);
				return (1);
			}

			free (name);
		}
	}
	else if (par_mac_name_ok) {
		if (mar_set_name_cstr (mar, par_mac_name)) {
			fprintf (stderr, "%s: bad mac file name\n", par_mac_name);
			return (1);
		}
	}

	if (par_type_ok) {
		mar_set_type (mar, par_type);
	}

	if (par_creator_ok) {
		mar_set_creator (mar, par_creator);
	}

	ftime = uf_get_mtime (fname);

	if (par_ctime != NULL) {
		if (strcmp (par_ctime, "file") == 0) {
			mar_set_ctime (mar, ftime);
		}
		else if (strcmp (par_ctime, "mtime") == 0) {
			mar_set_ctime (mar, mar_get_mtime (mar));
		}
		else {
			mar_set_ctime (mar, mac_time_from_string (par_ctime));
		}
	}

	if (par_mtime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mar_set_mtime (mar, ftime);
		}
		else if (strcmp (par_mtime, "ctime") == 0) {
			mar_set_mtime (mar, mar_get_ctime (mar));
		}
		else {
			mar_set_mtime (mar, mac_time_from_string (par_mtime));
		}
	}

	if (par_finder_flags_ok) {
		v = mar_get_finder_flags (mar);

		v &= ~par_finder_flags_msk;
		v |= par_finder_flags_val & par_finder_flags_msk;

		mar_set_finder_flags (mar, v);
	}

	if (par_file_flags_ok) {
		v = mar_get_file_flags (mar);

		v &= ~par_file_flags_msk;
		v |= par_file_flags_val & par_file_flags_msk;

		mar_set_file_flags (mar, v);
	}

	if (par_position_x_ok) {
		mar_set_position_x (mar, par_position_x);
	}

	if (par_position_y_ok) {
		mar_set_position_y (mar, par_position_y);
	}

	if (par_scroll_x_ok) {
		mar_set_scroll_x (mar, par_scroll_x);
	}

	if (par_scroll_y_ok) {
		mar_set_scroll_y (mar, par_scroll_y);
	}

	if (par_window_x1_ok) {
		mar_set_window_x1 (mar, par_window_x1);
	}

	if (par_window_y1_ok) {
		mar_set_window_y1 (mar, par_window_y1);
	}

	if (par_window_x2_ok) {
		mar_set_window_x2 (mar, par_window_x2);
	}

	if (par_window_y2_ok) {
		mar_set_window_y2 (mar, par_window_y2);
	}

	if (par_window_view_ok) {
		mar_set_window_view (mar, par_window_view);
	}

	if (par_mar_id_ok) {
		mar_set_id (mar, par_mar_id);
	}

	if (par_zero_mar) {
		mar_zero_unused (mar);
	}

	return (0);
}

static
int mar_fix_fp (unsigned char *mar, FILE *fp, const char *fname)
{
	if (mar_read (mar, fp)) {
		if (par_force) {
			mar_init (mar);
			mar_set_id (mar, MAR_FILE);
			mar_set_name_cstr (mar, "Noname");
		}
		else {
			fprintf (stderr,
				"%s: error reading mar header\n", fname
			);
			return (1);
		}
	}

	if (mar_fix_mar (mar, fp, fname)) {
		return (1);
	}

	if (fseek (fp, 0, SEEK_SET)) {
		fprintf (stderr, "%s: seek error\n", fname);
		return (1);
	}

	mar_set_crc (mar);

	if (mar_write (mar, fp)) {
		fprintf (stderr, "%s: write error\n", fname);
		return (1);
	}

	return (0);
}

static
int mar_fix (const char *fname)
{
	int           r;
	FILE          *fp;
	unsigned char mar[128];
	unsigned long mtime;

	fp = fopen (fname, "r+b");

	if ((fp == NULL) && par_force) {
		fp = fopen (fname, "w+b");
	}

	if (fp == NULL) {
		fprintf (stderr, "%s: open failed\n", fname);
		return (1);
	}

	r = mar_fix_fp (mar, fp, fname);

	fclose (fp);

	if (r) {
		return (1);
	}

	mtime = mar_get_mtime (mar);

	if (uf_set_mtime (fname, mtime)) {
		fprintf (stderr, "%s: can't set file time\n", fname);
		return (1);
	}

	return (0);
}

int mar_fix_main (int argc, char **argv)
{
	int  r;
	int  ret;
	char **optarg;

	arg0 = argv[0];

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'c':
			if (mar_set_property ("creator", optarg[0])) {
				fprintf (stderr,
					"%s: bad creator fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (1);
			}
			break;

		case 'C':
			par_ctime = optarg[0];
			break;

		case 'f':
			ret |= mar_fix (optarg[0]);
			break;

		case 'F':
			par_force = 1;
			break;

		case 'I':
			if (mar_set_property ("mar-id", optarg[0])) {
				fprintf (stderr,
					"%s: bad mar id fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (1);
			}
			break;

		case 'l':
			if (mar_set_property ("finder-flags", optarg[0])) {
				fprintf (stderr, "%s: bad finder flags (%s)\n",
					argv[0], optarg[0]
				);
				return (1);
			}
			break;

		case 'L':
			if (mar_set_property ("file-flags", optarg[0])) {
				fprintf (stderr, "%s: bad file flags (%s)\n",
					argv[0], optarg[0]
				);
				return (1);
			}
			break;

		case 'm':
			par_mac_name = optarg[0];
			par_mac_name_ok = 1;
			break;

		case 'M':
			par_mtime = optarg[0];
			break;

		case 'n':
			par_fix_name = 1;
			break;

		case 's':
			if (mar_set_property (optarg[0], optarg[1])) {
				fprintf (stderr, "%s: bad property (%s = %s)\n",
					argv[0], optarg[0], optarg[1]
				);
				return (1);
			}
			break;

		case 't':
			if (mar_set_property ("type", optarg[0])) {
				fprintf (stderr, "%s: bad type fourcc (%s)\n",
					argv[0], optarg[0]
				);
				return (1);
			}
			break;

		case 'x':
			par_position_x = (unsigned) strtoul (optarg[0], NULL, 0);
			par_position_x_ok = 1;
			break;

		case 'y':
			par_position_y = (unsigned) strtoul (optarg[0], NULL, 0);
			par_position_y_ok = 1;
			break;

		case 'z':
			par_zero_mar = 1;
			break;

		case 0:
			ret |= mar_fix (optarg[0]);
			break;

		default:
			return (1);
		}
	}

	if (ret) {
		return (1);
	}

	return (0);
}
