/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mar/mar_info.c                                           *
 * Created:     2011-06-25 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2011 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "mar.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macfile.h>
#include <lib/mactime.h>
#include <lib/mar.h>


static const char *arg0 = NULL;

static int par_inf_all = 1;
static int par_inf_mf = 0;

static char par_short = 0;
static char par_inf_fname = 0;
static char par_inf_mar_id = 0;
static char par_inf_type = 0;
static char par_inf_creator = 0;
static char par_inf_dsize = 0;
static char par_inf_rsize = 0;
static char par_inf_finder_flags = 0;
static char par_inf_file_flags = 0;
static char par_inf_position = 0;
static char par_inf_window_size = 0;
static char par_inf_scroll = 0;
static char par_inf_window_view = 0;
static char par_inf_folder_id = 0;
static char par_inf_ctime = 0;
static char par_inf_mtime = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "all", NULL, "Print all information [default]" },
	{ 'c', 0, "creator", NULL, "Print the creator" },
	{ 'C', 0, "ctime", NULL, "Print the creation time stamp" },
	{ 'f', 0, "mar-file", NULL, "Print the file name" },
	{ 'i', 1, "input", "string", "Set an input file name" },
	{ 'I', 0, "info", NULL, "Dump the mac file info [no]" },
	{ 'l', 0, "finder-flags", NULL, "Print the finder flags" },
	{ 'L', 0, "file-flags", NULL, "Print the file flags" },
	{ 'M', 0, "mtime", NULL, "Print the modification time stamp" },
	{ 's', 0, "short", NULL, "Short output format [no]" },
	{ 'S', 0, "scrollbar", NULL, "Print the scroll bar position [no]" },
	{ 't', 0, "type", NULL, "Print the type" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'w', 0, "window-size", NULL, "Print the window size" },
	{ 'x', 0, "position", NULL, "Print the position" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mar -i: print mar header information",
		"usage: mar -i [options] [file...]",
		opts
	);

	fflush (stdout);
}


static
int mar_get_sint16 (unsigned val)
{
	if (val & 0x8000) {
		return (-((~val + 1) & 0xffff));
	}
	else {
		return (val & 0xffff);
	}
}

static
void mar_print_finder_flags_sym (FILE *fp, unsigned ff)
{
	int        first;
	unsigned   mask;
	const char *flag;

	first = 1;

	mask = 0x8000;

	while (ff != 0) {
		if (ff & mask) {
			flag = mf_finder_flag_to_string (mask);
			ff &= ~mask;

			if (first == 0) {
				fputs (", ", fp);
			}
			else {
				first = 0;
			}

			fputs (flag, fp);
		}

		mask = mask >> 1;
	}
}

static
void mar_print_file_flags_sym (FILE *fp, unsigned long ff)
{
	int           first;
	unsigned long mask;
	const char    *flag;

	first = 1;

	mask = 0x80000000;

	while (ff != 0) {
		if (ff & mask) {
			flag = mf_file_flag_to_string (mask);
			ff &= ~mask;

			if (first == 0) {
				fputs (", ", fp);
			}
			else {
				first = 0;
			}

			fputs (flag, fp);
		}

		mask = mask >> 1;
	}
}


static
void mar_print_file_name (const unsigned char *mar, FILE *fp, const char *fname)
{
	if (par_inf_all == 0) {
		return;
	}

	fprintf (fp, "file name:      '%s'\n", fname);
}

static
void mar_print_mac_name (const unsigned char *mar, FILE *fp)
{
	char *name;

	if ((par_inf_fname == 0) && (par_inf_all == 0)) {
		return;
	}

	name = mar_get_name_cstr (mar);

	if (par_short) {
		fprintf (fp, "%s\n", name);
	}
	else {
		fprintf (fp, "mac name:       '%s'\n", name);
	}

	free (name);
}

static
void mar_print_mar_id (const unsigned char *mar, FILE *fp)
{
	unsigned long v;
	char          str[64];

	if ((par_inf_mar_id == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_id (mar);
	str_set_fourcc (str, 64, v);

	if (par_short) {
		fprintf (fp, "%s\n", str);
	}
	else {
		fprintf (fp,
			"entry id:       \"%s\" (0x%08lX)\n", str, v
		);
	}
}

static
void mar_print_type (const unsigned char *mar, FILE *fp)
{
	unsigned long v;
	char          str[64];

	if ((par_inf_type == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_type (mar);
	str_set_fourcc (str, 64, v);

	if (par_short) {
		fprintf (fp, "%s\n", str);
	}
	else {
		fprintf (fp,
			"file type:      \"%s\" (0x%08lX)\n", str, v
		);
	}
}

static
void mar_print_creator (const unsigned char *mar, FILE *fp)
{
	unsigned long v;
	char          str[64];

	if ((par_inf_creator == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_creator (mar);
	str_set_fourcc (str, 64, v);

	if (par_short) {
		fprintf (fp, "%s\n", str);
	}
	else {
		fprintf (fp,
			"file creator:   \"%s\" (0x%08lX)\n",
			str, v
		);
	}
}

static
void mar_print_data_size (const unsigned char *mar, FILE *fp)
{
	unsigned long v;

	if ((par_inf_dsize == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_data_size (mar);

	if (par_short) {
		fprintf (fp, "%lu\n", v);
	}
	else {
		fprintf (fp, "data fork size: %lu\n", v);
	}
}

static
void mar_print_rsrc_size (const unsigned char *mar, FILE *fp)
{
	unsigned long v;

	if ((par_inf_rsize == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_rsrc_size (mar);

	if (par_short) {
		fprintf (fp, "%lu\n", v);
	}
	else {
		fprintf (fp, "rsrc fork size: %lu\n", v);
	}
}

static
void mar_print_finder_flags (const unsigned char *mar, FILE *fp)
{
	unsigned long v;

	if ((par_inf_finder_flags == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_finder_flags (mar);

	if (par_short) {
		fprintf (fp, "%04lx\n", v);
	}
	else {
		fprintf (fp, "finder flags:   0x%04lx", v);

		if (v != 0) {
			fputs (" (", fp);
			mar_print_finder_flags_sym (fp, v);
			fputs (")", fp);
		}

		fputs ("\n", fp);
	}
}

static
void mar_print_file_flags (const unsigned char *mar, FILE *fp)
{
	unsigned long v;

	if ((par_inf_file_flags == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_file_flags (mar);

	if (par_short) {
		fprintf (fp, "%02lx\n", v);
	}
	else {
		fprintf (fp, "file flags:     0x%02lx", v);

		if (v != 0) {
			fputs (" (", fp);
			mar_print_file_flags_sym (fp, v);
			fputs (")", fp);
		}

		fputs ("\n", fp);
	}
}

static
void mar_print_position (const unsigned char *mar, FILE *fp)
{
	const char *fmt;

	if ((par_inf_position == 0) && (par_inf_all == 0)) {
		return;
	}

	if (par_short) {
		fmt = "%d %d\n";
	}
	else {
		fmt = "position:       (%d, %d)\n";
	}

	fprintf (fp, fmt,
		mar_get_sint16 (mar_get_position_x (mar)),
		mar_get_sint16 (mar_get_position_y (mar))
	);
}

static
void mar_print_window_size (const unsigned char *mar, FILE *fp)
{
	const char *fmt;

	if ((par_inf_window_size == 0) && (par_inf_all == 0)) {
		return;
	}

	if (par_short) {
		fmt = "%d %d %d %d\n";
	}
	else {
		fmt = "window size:    (%d, %d) - (%d, %d)\n";
	}

	fprintf (fp, fmt,
		mar_get_sint16 (mar_get_window_x1 (mar)),
		mar_get_sint16 (mar_get_window_y1 (mar)),
		mar_get_sint16 (mar_get_window_x2 (mar)),
		mar_get_sint16 (mar_get_window_y2 (mar))
	);
}

static
void mar_print_scroll (const unsigned char *mar, FILE *fp)
{
	const char *fmt;

	if ((par_inf_scroll == 0) && (par_inf_all == 0)) {
		return;
	}

	if (par_short) {
		fmt = "%d %d\n";
	}
	else {
		fmt = "scroll:         (%d, %d)\n";
	}

	fprintf (fp, fmt,
		mar_get_sint16 (mar_get_scroll_x (mar)),
		mar_get_sint16 (mar_get_scroll_y (mar))
	);
}

static
void mar_print_window_view (const unsigned char *mar, FILE *fp)
{
	const char *fmt;

	if ((par_inf_window_view == 0) && (par_inf_all == 0)) {
		return;
	}

	if (par_short) {
		fmt = "0x%04x\n";
	}
	else {
		fmt = "window view:    0x%04x\n";
	}

	fprintf (fp, fmt, (unsigned) mar_get_window_view (mar));
}

static
void mar_print_folder_id (const unsigned char *mar, FILE *fp)
{
	unsigned v;

	if ((par_inf_folder_id == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_folder_id (mar);

	if (par_short) {
		fprintf (fp, "%04x\n", v);
	}
	else {
		fprintf (fp, "folder id:      0x%04x\n", v);
	}
}

static
void mar_print_ctime (const unsigned char *mar, FILE *fp)
{
	unsigned      y, m, d, hrs, min, sec;
	unsigned long v;

	if ((par_inf_ctime == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_ctime (mar);

	mac_time_to_ymd (v, &y, &m, &d);
	mac_time_to_hms (v, &hrs, &min, &sec);

	if (par_short) {
		fprintf (fp, "%04u-%02u-%02u %02u:%02u:%02u\n",
			y, m + 1, d + 1, hrs, min, sec
		);
	}
	else {
		fprintf (fp,
			"ctime:          "
			"%04u-%02u-%02u %02u:%02u:%02u (%lu)\n",
			y, m + 1, d + 1, hrs, min, sec, v
		);
	}
}

static
void mar_print_mtime (const unsigned char *mar, FILE *fp)
{
	unsigned      y, m, d, hrs, min, sec;
	unsigned long v;

	if ((par_inf_mtime == 0) && (par_inf_all == 0)) {
		return;
	}

	v = mar_get_mtime (mar);

	mac_time_to_ymd (v, &y, &m, &d);
	mac_time_to_hms (v, &hrs, &min, &sec);

	if (par_short) {
		fprintf (fp, "%04u-%02u-%02u %02u:%02u:%02u\n",
			y, m + 1, d + 1, hrs, min, sec
		);
	}
	else {
		fprintf (fp,
			"mtime:          "
			"%04u-%02u-%02u %02u:%02u:%02u (%lu)\n",
			y, m + 1, d + 1, hrs, min, sec, v
		);
	}
}

static
int mar_info_fp (FILE *fp, FILE *out, const char *fname)
{
	unsigned char   mar[128];
	int             isdir;
	static unsigned cnt = 0;

	if (cnt > 0) {
		fputs ("\n", out);
	}

	cnt += 1;

	if (mar_read (mar, fp)) {
		fprintf (stderr, "%s: error reading mar header (%s)\n", arg0, fname);
		return (1);
	}

	if (mar_is_mar (mar) == 0) {
		fprintf (stderr, "%s: not a mar file (%s)\n", arg0, fname);
		return (1);
	}

	if (par_inf_mf) {
		macfile_t mf;

		mf_init (&mf);
		mar_get_macfile (mar, &mf);
		mf_save (&mf, out);
	}

	isdir = (mar_get_file_flags (mar) & MF_FILE_ISDIR) != 0;

	mar_print_file_name (mar, out, fname);
	mar_print_mac_name (mar, out);
	mar_print_mar_id (mar, out);

	if (isdir == 0) {
		mar_print_type (mar, out);
		mar_print_creator (mar, out);
	}

	mar_print_data_size (mar, out);
	mar_print_rsrc_size (mar, out);

	mar_print_finder_flags (mar, out);
	mar_print_file_flags (mar, out);

	mar_print_position (mar, out);

	if (isdir) {
		mar_print_window_size (mar, out);
		mar_print_scroll (mar, out);
		mar_print_window_view (mar, out);
	}
	else {
		mar_print_folder_id (mar, out);
	}

	mar_print_ctime (mar, out);
	mar_print_mtime (mar, out);

	fflush (out);

	return (0);
}

static
int mar_info (const char *fname, FILE *out)
{
	int  r;
	FILE *fp;

	fp = fopen (fname, "rb");

	if (fp == NULL) {
		fprintf (stderr, "error opening file (%s)\n", fname);
		return (1);
	}

	r = mar_info_fp (fp, out, fname);

	fclose (fp);

	return (r);
}

int mar_info_main (int argc, char **argv)
{
	int  r;
	int  ret;
	char **optarg;

	ret = 0;

	arg0 = argv[0];

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'a':
			par_inf_all = 1;
			break;

		case 'c':
			par_inf_creator = 1;
			par_inf_all = 0;
			break;

		case 'C':
			par_inf_ctime = 1;
			par_inf_all = 0;
			break;

		case 'f':
			par_inf_fname = 1;
			par_inf_all = 0;
			break;

		case 'F':
			par_inf_folder_id = 1;
			par_inf_all = 0;
			break;

		case 'i':
			ret |= mar_info (optarg[0], stdout);
			break;

		case 'I':
			par_inf_mf = 1;
			par_inf_all = 0;
			break;

		case 'l':
			par_inf_finder_flags = 1;
			par_inf_all = 0;
			break;

		case 'L':
			par_inf_file_flags = 1;
			par_inf_all = 0;
			break;

		case 'M':
			par_inf_mtime = 1;
			par_inf_all = 0;
			break;

		case 's':
			par_short = 1;
			break;

		case 'S':
			par_inf_scroll = 1;
			par_inf_all = 0;
			break;

		case 't':
			par_inf_type = 1;
			par_inf_all = 0;
			break;

		case 'w':
			par_inf_window_size = 1;
			par_inf_all = 0;
			break;

		case 'x':
			par_inf_position = 1;
			par_inf_all = 0;
			break;

		case 0:
			ret |= mar_info (optarg[0], stdout);
			break;

		default:
			return (1);
		}
	}

	if (ret) {
		return (1);
	}

	return (0);
}

int mar_info_mf_main (int argc, char **argv)
{
	par_inf_mf = 1;
	par_inf_all = 0;

	return (mar_info_main (argc, argv));
}
