/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mbp/mbplist.c                                            *
 * Created:     2007-12-02 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2007-2011 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"

#include "mbplist.h"

#include <lib/file.h>
#include <lib/fourcc.h>
#include <lib/getopt.h>
#include <lib/macbinary.h>
#include <lib/mactime.h>
#include <lib/strings.h>


static int mbp_list_folder (FILE *inp, FILE *out, unsigned long *ofs,
	unsigned char *mb, const char *prefix);


static const char *arg0 = NULL;


static int        par_short = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'i', 1, "input", "string", "Set a file name" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mbplist: list the contents of MacBinary II+ files",
		"usage: mbplist [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mbplist version " MACUTILS_VERSION_STR
		"\n\n"
		"Copyright (C) 2007-2009 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}

static
int mbp_list_mb (FILE *out, const unsigned char *mb,
	const char *name, const char *pre)
{
	unsigned      yy, mm, dd;
	unsigned      hrs, min, sec;
	unsigned long mtime;
	unsigned long data, rsrc, info;
	unsigned long val;
	unsigned      fflags, attrib;
	char          type[16];
	char          creator[16];

	data = mb_get_data_size (mb);
	rsrc = mb_get_rsrc_size (mb);
	info = mb_get_getinfo_size (mb);

	mtime = mb_get_mtime (mb);
	mac_time_to_ymd (mtime, &yy, &mm, &dd);
	mac_time_to_hms (mtime, &hrs, &min, &sec);

	str_set_fourcc (type, 16, mb_get_type (mb));

	val = mb_get_creator (mb);
	if ((val == 0xffffffff) || (val == 0xfffffffe)) {
		strcpy (creator, "----");
	}
	else {
		str_set_fourcc (creator, 16, val);
	}

	fflags = mb_get_finder_flags (mb);
	attrib = mb_get_file_flags (mb);

	fprintf (out,
		"%s %s %04u-%02u-%02u %02u:%02u:%02u %04X %02X %8lu %8lu %2lu %s%s\n",
		type, creator,
		yy, mm, dd, hrs, min, sec,
		fflags, attrib,
		data, rsrc, info,
		pre,
		name
	);

	return (0);
}

static
int mbp_list_file (FILE *inp, FILE *out, unsigned long *ofs,
	const unsigned char *mb, const char *prefix)
{
	char          *name;
	char          *path;
	unsigned long data, rsrc, info;

	if (mb_get_version (mb) == 0) {
		fprintf (stderr, "%s: macbinary header error\n", arg0);
		return (1);
	}

	name = mac_pstring_decode (mb + 1, "/");
	path = mac_combine_path (prefix, name);

	if (par_short) {
		mbp_list_mb (out, mb, name, " ");
	}
	else {
		mbp_list_mb (out, mb, path, " ");
	}

	free (path);
	free (name);

	data = mb_get_data_size (mb);
	rsrc = mb_get_rsrc_size (mb);
	info = mb_get_getinfo_size (mb);

	data = (data + 127) & ~0x7fUL;
	rsrc = (rsrc + 127) & ~0x7fUL;
	info = (info + 127) & ~0x7fUL;

	*ofs += 128 + data + rsrc + info;

	return (0);
}

static
int mbp_list_files (FILE *inp, FILE *out, unsigned long *ofs,
	unsigned char *mb, const char *prefix, int folder)
{
	unsigned long creator;

	while (1) {
		if (mac_read (inp, mb, *ofs, 128)) {
			memset (mb, 0, 128);

			if (folder) {
				return (1);
			}

			if (ferror (inp)) {
				return (1);
			}

			return (0);
		}

		if (mb[0] == 0) {
			if (mbp_list_file (inp, out, ofs, mb, prefix)) {
				return (1);
			}
		}
		else if (mb[0] == 1) {
			creator = mb_get_creator (mb);

			if (creator == 0xffffffff) {
				if (mbp_list_folder (inp, out, ofs, mb, prefix)) {
					return (1);
				}
			}
			else if (creator == 0xfffffffe) {
				return (0);
			}
			else {
				fprintf (stderr,
					"%s: macbinary header error\n",
					arg0
				);
				return (1);
			}
		}
		else {
			fprintf (stderr, "%s: macbinary header error\n",
				arg0
			);
			return (1);
		}
	}

	return (1);
}

static
int mbp_list_folder (FILE *inp, FILE *out, unsigned long *ofs,
	unsigned char *mb, const char *prefix)
{
	char *name;
	char *path;

	name = mac_pstring_decode (mb + 1, "/");
	path = mac_combine_path (prefix, name);

	mbp_list_mb (out, mb, path, "+");

	*ofs += 128;

	if (mbp_list_files (inp, out, ofs, mb, path, 1)) {
		free (path);
		free (name);
		return (1);
	}

	mbp_list_mb (out, mb, path, "-");

	*ofs += 128;

	free (path);
	free (name);

	return (0);
}

static
int mbp_list_fp (FILE *inp, FILE *out)
{
	unsigned char mb[128];
	unsigned long ofs;

	ofs = 0;

	if (mbp_list_files (inp, out, &ofs, mb, "", 0)) {
		return (1);
	}

	return (0);
}

static
int mbp_list (const char *inp, FILE *out)
{
	int  r;
	FILE *fp;

	fp = mac_open_inp (inp, "rb", 1);

	if (fp == NULL) {
		return (1);
	}

	r = mbp_list_fp (fp, out);

	mac_close (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;

	arg0 = argv[0];

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			print_version();
			return (EXIT_SUCCESS);

		case 'i':
			if (mbp_list (optarg[0], stdout)) {
				return (1);
			}
			break;

		case 0:
			if (mbp_list (optarg[0], stdout)) {
				return (1);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	return (0);
}
