/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mfs/mfsfix.c                                             *
 * Created:     2012-04-25 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2012 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include "mfsfix.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/mactime.h>
#include <lib/mfs.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


static const char     *arg0 = NULL;

static unsigned long  par_offset = 0;
static unsigned long  par_blocks = 0;

static char           par_update_alt_vinfo = 0;
static char           par_restore_alt_vinfo = 0;

static const char     *par_vol_name = NULL;

static const char     *par_ctime = NULL;
static const char     *par_mtime = NULL;

static char           par_vol_attr_ok = 0;
static unsigned short par_vol_attr = 0;

static char           par_zero_boot = 0;
static char           par_zero_unused = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "update-alternate", NULL, "Update the alternate volume information [no]" },
	{ 'A', 0, "restore-alternate", NULL, "Restore the alternate volume information [no]" },
	{ 'C', 1, "ctime", "string", "Set the creation time" },
	{ 'f', 0, "file", NULL, "Set an HFS file name" },
	{ 'M', 1, "mtime", "string", "Set the modification time" },
	{ 'n', 1, "blocks", "int", "Set the number of blocks in the volume [auto]" },
	{ 'o', 1, "offset", "int", "Set the starting offset [0]" },
	{ 't', 1, "volume-attributes", "int", "Set the volume attributes" },
	{ 'v', 1, "volume-name", "string", "Set the volume name" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'z', 0, "zero-unused", NULL, "Clear unused space" },
	{ 'Z', 0, "zero-boot", NULL, "Clear the boot block" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mfsfix: fix MFS data",
		"usage: mfsfix [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mfsfix version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2012 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}


static
void mfs_zero_name (mfs_vol_t *mfs)
{
	unsigned i;

	i = mfs->volinfo[36];

	while (i < 27) {
		mfs->volinfo[37 + i++] = 0;
	}

	mfs->volinfo_mod = 1;
}

static
int mfs_zero_blocks (mfs_vol_t *mfs)
{
	unsigned      i, n;
	unsigned long j;
	unsigned      free_cnt, used_cnt;
	unsigned long idx, cnt;
	unsigned char buf[512];

	n = mfs_get_ablk_count (mfs);

	if (n > 4096) {
		n = 4096;
	}

	memset (buf, 0, 512);

	free_cnt = 0;
	used_cnt = 0;

	cnt = mfs_get_ablk_size (mfs) / 512;
	idx = mfs_get_ablk_start (mfs);

	for (i = 0; i < n; i++) {
		if (mfs->bmap[i] != 0) {
			used_cnt += 1;
		}
		else {
			for (j = 0; j < cnt; j++) {
				if (mfs_write_block (mfs, buf, idx + j, 1)) {
					fprintf (stderr, "%s: write error block %lu\n",
						arg0, idx + j
					);
					return (1);
				}
			}

			free_cnt += 1;
		}

		idx += cnt;
	}

	if (mfs->blocks > 2) {
		while (idx < (mfs->blocks - 2)) {
			if (mfs_write_block (mfs, buf, idx, 1)) {
				fprintf (stderr, "%s: write error block %lu\n",
					arg0, idx
				);
				return (1);
			}

			idx += 1;
		}

		if (mfs_write_block (mfs, buf, mfs->blocks - 1, 1)) {
			fprintf (stderr, "%s: write error block %lu\n",
				arg0, mfs->blocks - 1
			);
			return (1);
		}
	}

	return (0);
}

static
int mfs_zero_unused (mfs_vol_t *mfs)
{
	mfs_zero_name (mfs);

	if (mfs_zero_blocks (mfs)) {
		return (1);
	}

	return (0);
}

static
int mfs_fix_mfs (mfs_vol_t *mfs, unsigned long mtime)
{
	if (par_restore_alt_vinfo) {
		mfs_restore_alt_volinfo (mfs);
	}

	if (par_vol_name != NULL) {
		if (mfs_set_vol_name (mfs, par_vol_name)) {
			fprintf (stderr, "%s: bad volume name (%s)\n",
				arg0, par_vol_name
			);
			return (1);
		}
	}

	if (par_ctime != NULL) {
		if (strcmp (par_ctime, "file") == 0) {
			mfs_set_ctime (mfs, mtime);
		}
		else if (strcmp (par_ctime, "mtime") == 0) {
			mfs_set_ctime (mfs, mfs_get_mtime (mfs));
		}
		else {
			mfs_set_ctime (mfs, mac_time_from_string (par_ctime));
		}
	}

	if (par_mtime != NULL) {
		if (strcmp (par_mtime, "file") == 0) {
			mfs_set_mtime (mfs, mtime);
		}
		else if (strcmp (par_mtime, "ctime") == 0) {
			mfs_set_mtime (mfs, mfs_get_ctime (mfs));
		}
		else {
			mfs_set_mtime (mfs, mac_time_from_string (par_mtime));
		}
	}

	if (par_vol_attr_ok) {
		mfs_set_volume_attributes (mfs, par_vol_attr);
	}

	if (par_zero_boot) {
		mfs->boot_mod = 1;
		memset (mfs->boot, 0, 1024);
	}

	if (par_zero_unused) {
		mfs_zero_unused (mfs);
	}

	if (par_update_alt_vinfo) {
		mfs_update_alt_volinfo (mfs);
	}

	return (0);
}

static
int mfs_fix_fp (FILE *fp, unsigned long offset, unsigned long blocks, unsigned long *mtime)
{
	int       r;
	mfs_vol_t mfs;

	if (mfs_open_auto (&mfs, fp, offset, blocks)) {
		fprintf (stderr, "%s: opening MFS volume failed\n", arg0);
		return (1);
	}

	r = mfs_fix_mfs (&mfs, *mtime);

	*mtime = mfs_get_mtime (&mfs);

	if (mfs_close (&mfs, 1)) {
		fprintf (stderr, "%s: write error\n", arg0);
		r = 1;
	}

	return (r);
}

static
int mfs_fix (const char *fname, unsigned long offset)
{
	int           r;
	unsigned long blocks, mtime;
	FILE          *fp;

	if (uf_stat (fname, &blocks, &mtime)) {
		fprintf (stderr, "%s: stat failed (%s)\n", arg0, fname);
		return (1);
	}

	if (par_blocks > 0) {
		blocks = par_blocks;
	}
	else {
		if (offset >= blocks) {
			fprintf (stderr, "%s: bad offset (%lu > %lu)\n",
				arg0, offset, blocks
			);
			return (1);
		}

		blocks = (blocks - offset) / 512;
	}

	fp = mac_open_inp (fname, "r+b", 0);

	if (fp == NULL) {
		fprintf (stderr, "%s: can't open file (%s)\n", arg0, fname);
		return (1);
	}

	r = mfs_fix_fp (fp, offset, blocks, &mtime);

	mac_close (fp);

	if (r == 0) {
		uf_set_mtime (fname, mtime);
	}

	return (r);
}

int main (int argc, char *argv[])
{
	int  r, ret;
	char **optarg;

	arg0 = argv[0];

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'a':
			par_update_alt_vinfo = 1;
			break;

		case 'A':
			par_restore_alt_vinfo = 1;
			break;

		case 'C':
			par_ctime = optarg[0];
			break;

		case 'f':
			ret |= mfs_fix (optarg[0], par_offset);
			break;

		case 'M':
			par_mtime = optarg[0];
			break;

		case 'n':
			par_blocks = strtoul (optarg[0], NULL, 0);
			break;

		case 'o':
			par_offset = strtoul (optarg[0], NULL, 0);
			break;

		case 't':
			par_vol_attr = strtoul (optarg[0], NULL, 0);
			par_vol_attr_ok = 1;
			break;

		case 'v':
			par_vol_name = optarg[0];
			break;

		case 'z':
			par_zero_unused = 1;
			break;

		case 'Z':
			par_zero_boot = 1;
			break;

		case 0:
			ret |= mfs_fix (optarg[0], par_offset);
			break;

		default:
			return (1);
		}
	}

	if (ret) {
		return (1);
	}

	return (0);
}
