/*****************************************************************************
 * mac-utils                                                                 *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mfs/mfsinfo.c                                            *
 * Created:     2012-04-21 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2012 Hampa Hug <hampa@hampa.ch>                          *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include "config.h"

#include "mfsinfo.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <lib/buffer.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/mfs.h>
#include <lib/mactime.h>
#include <lib/strings.h>
#include <lib/unixfile.h>


#define TAG "%-24s "


static const char    *arg0 = NULL;

static unsigned long par_offset = 0;
static unsigned long par_blocks = 0;

static char          par_info_short = 0;
static char          par_info_vinf = 0;
static char          par_info_boot = 0;

static char          par_info_name = 0;
static char          par_info_ctime = 0;
static char          par_info_mtime = 0;

static char          par_cat = 0;

static char          par_first = 1;

static char          par_alt_vinf = 0;


static mac_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "alternate", NULL, "Use the alternate volume info [no]" },
	{ 'b', 0, "info-boot", NULL, "Print boot block info [no]" },
	{ 'c', 0, "cat", NULL, "Dump the volume to standard output [no]" },
	{ 'C', 0, "ctime", NULL, "Print the volume creation time [no]" },
	{ 'i', 1, "input", "string", "Set an input file name [none]" },
	{ 'm', 0, "info-volume", NULL, "Print volume info [no]" },
	{ 'M', 0, "mtime", NULL, "Print the volume modification time [no]" },
	{ 'n', 1, "blocks", "int", "Set the number of blocks in the volume [auto]" },
	{ 'o', 1, "offset", "int", "Set the starting offset [0]" },
	{ 's', 0, "info-short", NULL, "Print short info [yes]" },
	{ 'v', 0, "volume-name", NULL, "Print the volume name [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	mac_getopt_help (
		"mfsinfo: print information about MFS volumes",
		"usage: mfsinfo [options] [file...]",
		opts
	);

	fflush (stdout);
}

static
void print_version (void)
{
	fputs (
		"mfsinfo version " MACUTILS_VERSION_STR "\n\n"
		"Copyright (C) 2012 Hampa Hug <hampa@hampa.ch>\n",
		stdout
	);

	fflush (stdout);
}


static
int mfs_cat (mfs_vol_t *mfs, FILE *fp)
{
	unsigned long i, n, max;
	unsigned char buf[8192];

	if (mfs->blocks == 0) {
		return (1);
	}

	max = 16;

	i = 0;
	while (i < mfs->blocks) {
		n = mfs->blocks - i;

		if (n > max) {
			n = max;
		}

		if (mfs_read_block (mfs, buf, i, n)) {
			if (max > 1) {
				max = 1;
				continue;
			}

			fprintf (stderr, "%s: read error at block %lu\n", arg0, i);

			memset (buf, 0, 512 * n);
		}

		if (fwrite (buf, 512, n, fp) != n) {
			return (1);
		}

		i += n;
	}

	return (0);
}

static
void mfs_print_hex16 (FILE *fp, const char *tag, unsigned val)
{
	fprintf (fp, TAG "0x%04x\n", tag, val);
}

static
void mfs_print_uint (FILE *fp, const char *tag, unsigned long val)
{
	fprintf (fp, TAG "%lu\n", tag, val);
}

static
void mfs_print_time (FILE *fp, const char *tag, unsigned long val)
{
	unsigned y, m, d, hrs, min, sec;

	mac_time_to_ymd (val, &y, &m, &d);
	mac_time_to_hms (val, &hrs, &min, &sec);

	fprintf (fp, TAG "%04u-%02u-%02u %02u:%02u:%02u (%lu)\n",
		tag, y, m + 1, d + 1, hrs, min, sec, val
	);
}

static
void mfs_print_time_short (FILE *fp, unsigned long val)
{
	unsigned y, m, d, hrs, min, sec;

	mac_time_to_ymd (val, &y, &m, &d);
	mac_time_to_hms (val, &hrs, &min, &sec);

	fprintf (fp, "%04u-%02u-%02u %02u:%02u:%02u\n",
		y, m + 1, d + 1, hrs, min, sec
	);
}

static
void mfs_print_pstr (FILE *fp, const char *tag, const unsigned char *str, unsigned max)
{
	char *tmp;

	if (str[0] >= max) {
		return;
	}

	tmp = mac_pstring_decode (str, "");

	if (tmp == NULL) {
		return;
	}

	fprintf (fp, TAG "\"%s\"\n", tag, tmp);

	free (tmp);
}

static
void mfs_print_cstr (FILE *fp, const char *tag, char *str, int free_str)
{
	if (str == NULL) {
		return;
	}

	fprintf (fp, TAG "\"%s\"\n", tag, str);

	if (free_str) {
		free (str);
	}
}


static
int mfs_print_info_boot (FILE *fp, mfs_vol_t *mfs, const unsigned char *boot)
{
	if (get_uint16_be (mfs->boot, 0) != MFS_BOOT_ID) {
		return (1);
	}

	mfs_print_hex16 (fp, "Boot block version:", get_uint16_be (boot, 6));
	mfs_print_pstr (fp, "System:", boot + 10, 16);
	mfs_print_pstr (fp, "Shell:", boot + 26, 16);
	mfs_print_pstr (fp, "Debugger:", boot + 42, 16);
	mfs_print_pstr (fp, "Debugger 2:", boot + 58, 16);
	mfs_print_pstr (fp, "Startup screen:", boot + 74, 16);
	mfs_print_pstr (fp, "Startup program:", boot + 90, 16);
	mfs_print_pstr (fp, "Scrap file:", boot + 106, 16);

	return (0);
}

static
int mfs_print_volinfo (FILE *fp, mfs_vol_t *mfs)
{
	if (mfs_get_id (mfs) != MFS_ID) {
		return (1);
	}

	mfs_print_cstr (fp, "Volume name:", mfs_get_vol_name (mfs), 1);
	fprintf (fp, TAG "%lu K (%lu)\n", "Volume size:", mfs->blocks / 2, 512 * mfs->blocks);
	mfs_print_time (fp, "Volume ctime:", mfs_get_ctime (mfs));
	mfs_print_time (fp, "Volume mtime:", mfs_get_mtime (mfs));
	mfs_print_hex16 (fp, "Signature:", mfs_get_id (mfs));
	mfs_print_hex16 (fp, "Volume attributes:", mfs_get_volume_attributes (mfs));
	mfs_print_uint (fp, "Files:", mfs_get_file_count (mfs));
	mfs_print_uint (fp, "First directory block:", mfs_get_directory_start (mfs));
	mfs_print_uint (fp, "Directory blocks:", mfs_get_directory_blocks (mfs));
	mfs_print_uint (fp, "Allocation blocks:", mfs_get_ablk_count (mfs));
	mfs_print_uint (fp, "Allocation block size:", mfs_get_ablk_size (mfs));
	mfs_print_uint (fp, "Clump size:", mfs_get_clump_size (mfs));
	mfs_print_uint (fp, "First allocation block:", mfs_get_ablk_start (mfs));
	mfs_print_uint (fp, "Free allocation blocks:", mfs_get_free_ablk (mfs));

	return (0);
}

static
int mfs_print_volinfo_short (FILE *fp, mfs_vol_t *mfs)
{
	if (mfs_get_id (mfs) != MFS_ID) {
		return (1);
	}

	mfs_print_cstr (fp, "Volume name:", mfs_get_vol_name (mfs), 1);
	fprintf (fp, TAG "%lu K (%lu)\n", "Volume size:", mfs->blocks / 2, 512 * mfs->blocks);
	mfs_print_time (fp, "Volume ctime:", mfs_get_ctime (mfs));
	mfs_print_time (fp, "Volume mtime:", mfs_get_mtime (mfs));
	mfs_print_hex16 (fp, "Volume attributes:", mfs_get_volume_attributes (mfs));
	fprintf (fp, TAG "%u * %lu (%u free)\n", "Allocation blocks:",
		mfs_get_ablk_count (mfs), mfs_get_ablk_size (mfs), mfs_get_free_ablk (mfs)
	);
	fprintf (fp, TAG "%u files\n", "Entries:", mfs_get_file_count (mfs));

	return (0);
}

static
int mfs_print_info_name (FILE *fp, mfs_vol_t *mfs)
{
	char *str;

	str = mfs_get_vol_name (mfs);

	if (str == NULL) {
		return (1);
	}

	fprintf (fp, "%s\n", str);

	free (str);

	return (0);
}

static
int mfs_info_fp (FILE *inp, FILE *out, const char *fname, mfs_vol_t *mfs)
{
	int done;

	mfs_set_volinfo (mfs, par_alt_vinf);

	if (par_info_name) {
		return (mfs_print_info_name (out, mfs));
	}
	else if (par_info_ctime) {
		mfs_print_time_short (out, mfs_get_ctime (mfs));
		return (0);
	}
	else if (par_info_mtime) {
		mfs_print_time_short (out, mfs_get_mtime (mfs));
		return (0);
	}

	done = 0;

	if (par_first) {
		par_first = 0;
	}
	else {
		fputs ("\n", out);
	}

	fprintf (out, TAG "%s\n", "File name:", fname);

	if (par_info_boot) {
		mfs_print_info_boot (out, mfs, mfs->boot);
		done = 1;
	}

	if (par_info_vinf) {
		mfs_print_volinfo (out, mfs);
		done = 1;
	}

	if (done == 0) {
		mfs_print_volinfo_short (out, mfs);
	}

	return (0);
}

static
int mfs_info (const char *fname, unsigned long offset)
{
	int           r;
	unsigned long blocks, mtime;
	FILE          *fp;
	mfs_vol_t     mfs;

	if (uf_stat (fname, &blocks, &mtime)) {
		fprintf (stderr, "%s: stat failed (%s)\n", arg0, fname);
		return (1);
	}

	if (par_blocks > 0) {
		blocks = par_blocks;
	}
	else {
		if (offset >= blocks) {
			fprintf (stderr, "%s: bad offset (%lu > %lu)\n",
				arg0, offset, blocks
			);
			return (1);
		}

		blocks = (blocks - offset) / 512;
	}

	fp = mac_open_inp (fname, "rb", 0);

	if (fp == NULL) {
		fprintf (stderr, "%s: can't open input file (%s)\n",
			arg0, fname
		);
		return (1);
	}

	if (mfs_open_auto (&mfs, fp, offset, blocks)) {
		fprintf (stderr, "%s: opening MFS volume failed (%s)\n",
			arg0, fname
		);
		mac_close (fp);
		return (1);
	}

	if (par_cat) {
		r = mfs_cat (&mfs, stdout);
	}
	else {
		r = mfs_info_fp (fp, stdout, fname, &mfs);
	}

	mfs_close (&mfs, 0);

	mac_close (fp);

	return (r);
}

int main (int argc, char *argv[])
{
	int  r, ret;
	char **optarg;

	arg0 = argv[0];

	ret = 0;

	while (1) {
		r = mac_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (1);
		}

		switch (r) {
		case '?':
			print_help();
			return (0);

		case 'V':
			print_version();
			return (0);

		case 'a':
			par_alt_vinf = 1;
			break;

		case 'b':
			par_info_boot = 1;
			break;

		case 'c':
			par_cat = 1;
			break;

		case 'C':
			par_info_ctime = 1;
			break;

		case 'i':
			ret |= mfs_info (optarg[0], par_offset);
			break;

		case 'm':
			par_info_vinf = 1;
			break;

		case 'M':
			par_info_mtime = 1;
			break;

		case 'n':
			par_blocks = strtoul (optarg[0], NULL, 0);
			break;

		case 'o':
			par_offset = strtoul (optarg[0], NULL, 0);
			break;

		case 's':
			par_info_short = 1;
			break;

		case 'v':
			par_info_name = 1;
			break;

		case 0:
			ret |= mfs_info (optarg[0], par_offset);
			break;

		default:
			return (1);
		}
	}

	if (ret) {
		return (1);
	}

	return (0);
}
